#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <fcntl.h>
#include <wait.h>

#include <string>
#include <sstream>

#include "powerlib.h"
#include "cpufreq.h"

#include "powersaved.h"
#include "acpi.h"
#include "fcntl.h"
#include "stringutil.h"

//using namespace std;

string ACPI_Interface::ACPI_EVENTFILE = "/proc/acpi/event";

ACPI_Interface::ACPI_Interface(GeneralConfig *config_obj):PM_Interface(config_obj){
	pDebug(DBG_DEBUG, "Constructor ACPI_Interface\n");

	// enable temperature control for acpi

	last_button_occurence = getTimestamp();

	if (!config->current_scheme->FORCE_BATTERY_POLLING 
	    && setBatteryAlarm_() < 0){
		pDebug(DBG_WARN, "could not set battery alarm,"
		       " maybe you should force polling mode.");
// 		config->current_scheme->FORCE_BATTERY_POLLING=1;
	}
};

ACPI_Interface::~ACPI_Interface(){
	/* reset thermal trip points to BIOS settings (general_conf object stored BIOS settings) */
	if (thermal_trip_points_supported && config->ENABLE_THERMAL_MANAGEMENT == KERNEL){
		for (int x=0; x < MAX_THERMAL_ZONES && config->thermal_zones[x].present; x++){
			setThermalTrippoints(x, config->thermal_zones[x]);
		}
	}
	close(hwEvent_fd);	
}

int ACPI_Interface::openHWEventFD(){
	char discard[MAX_LINE_SIZE];
	struct stat file_stats;

	if (stat(ACPI_EVENTFILE.c_str(), &file_stats)){
		pDebug(DBG_ERR, "Cannot stat %s\n", ACPI_EVENTFILE.c_str());
		return -1;
	}
	if (S_ISSOCK(file_stats.st_mode)){
		if ((hwEvent_fd = openAcpidSocket()) < 0){
			pDebug (DBG_ERR, "Can't connect to acpid socket %s. Is the acpid running?\n",
				ACPI_EVENTFILE.c_str());
			return -1;
		}
	}
	else if (S_ISREG(file_stats.st_mode)){
		hwEvent_fd = open(ACPI_EVENTFILE.c_str(), O_RDONLY| O_NONBLOCK);
		if (hwEvent_fd < 0) {
			pDebug (DBG_ERR, "Can't open %s, is another process accessing it?\n",
				ACPI_EVENTFILE.c_str());
			return -1;
		}
	}
	else{
		pDebug(DBG_ERR, "No valid acpi event file: %s", ACPI_EVENTFILE.c_str());
		return -1;
	}
	// discard any events happened before
	while (read(hwEvent_fd, discard, sizeof(discard)) > 0){
		pDebug(DBG_INFO, "Event %s discarded\n", discard);
	}
	// set errno back: could be EAGAIN if no data is available
	errno = 0;
	return hwEvent_fd;
}

int ACPI_Interface::handleHWEventRequest(){

	char buf[2] = "";
	int x = 0, y;
        std::string line = "";
	std::string type;
	std::string dev_name;
	std::string port;
	std::string count;
	long time_in_millisecs = 0;

	/* this may come from acpid socket or from /proc/acpi/event */
	while ((y = read(hwEvent_fd, buf, 1)) > 0) {
                line += buf;
                x++;
                if (*buf == '\n')
                        break;
        }
	// end of file ? - maybe acpid and/or socket missing ?
	if ( x == 0 ){
		pDebug(DBG_DIAG, "hwEvent is empty or not present, fd: %d",
		       hwEvent_fd);
		return -2;
	}
	// error	
	else if ( y < 0){
		/* TODO: what if acpid died ? */
		int tmp = DBG_WARN;
		if (errno == EAGAIN)
			tmp = DBG_DIAG;
		pDebug(tmp, "Could not read from hwEvent fd: %d, error: %s",
		       hwEvent_fd, strerror(errno));
                if (line.empty())   // this seems to be a real error; we can also get EAGAIN which
                        return -2;  // does not have to be an error (but should be handled, TODO!)
	}

        // empty event?
        line = stripLeadingWS(line);  // just to make sure
        line = stripTrailingWS(line); // this also removes the trailing '\n'
        if (line.empty())
		return 0;

        pDebug (DBG_DIAG, "ACPI Event: '%s' ignore buttons: %s",
                line.c_str(), (sleep_triggered ? "yes" : "no"));

	// notify the clients about an acpi event
        DBus_Server::emitSignal("AcpiEvent", line.c_str());

        std::istringstream eStream(line.c_str());

	eStream >> type >> dev_name  >> port >> count;
	if (!type.length() || !dev_name.length() || !port.length() || !count.length()){
                pDebug (DBG_WARN, "Empty or incomplete event. Socket problem? ignoring...");
                return 1; // we should not get here since we already tested for empty events above.
	}
	pDebug (DBG_DEBUG, "type: %s, dev_name: %s, port: %s, count: %s\n", 
		type.c_str(), dev_name.c_str(), port.c_str(), count.c_str());

	if (type == "button/sleep"){
		if (sleep_triggered){
			pDebug(DBG_DIAG, "Sleep button ignored while resuming.");
			return 0;
		}
		// workaround for double button occurencies
		pDebug(DBG_DEBUG, "1) last_button_occurence: %lu\n", 
		       last_button_occurence);
		time_in_millisecs = getTimestamp();
		pDebug(DBG_DEBUG, "2) last_button_occurence: %lu\n", 
		       last_button_occurence);

		if ((time_in_millisecs - last_button_occurence) 
		    < button_time_window){
			pDebug (DBG_DIAG, "button sleep event occured twice"
				"(or together with a power event) in the last %d"
				" milliseconds and will therefore be ignored\n"
				, button_time_window);
			last_button_occurence = time_in_millisecs;
			return 0;
		}
		else{
			last_button_occurence = time_in_millisecs;
			pDebug (DBG_INFO, "button sleep event occured\n");
			executeEvent("button.sleep", line.c_str());
			return 1;
		}
	}
	else if (type == "button/power"){
		if (sleep_triggered){
			pDebug(DBG_DIAG, "Power button ignored while resuming.");
			return 0;
		}
		// workaround for double button occurencies
		time_in_millisecs = getTimestamp();
		if ((time_in_millisecs - last_button_occurence) 
		    < button_time_window){
			pDebug (DBG_DIAG, "button power event occured twice"
				"(or together with a power event) in the last"
				"%d milliseconds and will therefore be ignored\n"
				, button_time_window);
			last_button_occurence = time_in_millisecs;		
			return 0;
		}
		else{
			// wait a bit, maybe the machine should be cut off power
			if (config->current_scheme->POWER_BUTTON_DELAY > 0)
				sleep(config->current_scheme->POWER_BUTTON_DELAY);
			// get new timestamp because of sleep.
			last_button_occurence = getTimestamp();		
			pDebug (DBG_INFO, "button power event occured\n");
			executeEvent("button.power", line.c_str());
			return 1;
		}
	}
	else if (type == "button/lid"){
		if (sleep_triggered){
			pDebug(DBG_DIAG, "Lid button ignored while resuming.");
			return 0;
		}
		switch(getLidState()){
		case LID_OPEN:
			pDebug (DBG_INFO, "lid open event occured\n");
			executeEvent("button.lid.open", line.c_str());
			return 1;
		case LID_CLOSED:
			pDebug (DBG_INFO, "lid closed event occured\n");
			executeEvent("button.lid.closed", line.c_str());
			return 1;
		default:
			pDebug (DBG_ERR, "Could not determine lid state\n");
			return -1;
		}
	}
	else if (type == "battery"){
		// Is ac event thrown before battery event? Expect yes, 
		// if not I need a workaround!

		checkACStateChanges();		
		checkBatteryStateChanges();

		// WTF is this good for? -> This is to avoid polling of battery
		// set /proc/acpi/battery/*/alarm and wait for events in /proc/acpi/event

		// adjust battery alarm here, maybe we have a false alarm!
		if (!config->current_scheme->FORCE_BATTERY_POLLING 
		    && setBatteryAlarm_() < 0){
			pDebug(DBG_DIAG, "could not set battery alarm,"
				" maybe you should force polling mode.");
// 			config->current_scheme->FORCE_BATTERY_POLLING=1;
		}
		return 1;
	}
	// break if else to execute a battery event if ac_offline 
	// and battery state is not normal
	else if (type == "ac_adapter"){
		if (checkACStateChanges() < 0){
			pDebug(DBG_WARN, "Could not read out AC_Adapter state.");
			return -1;
		}
		return 1;
	}
	else if (type == "thermal_zone"){
		pDebug (DBG_INFO, "Thermal event happend.");
		checkTemperatureStateChanges();
		return 1;
	}
	else{
                pDebug (DBG_DIAG, "unassigned HW event, using [other]."
                        " type '%s' dev_name '%s' port '%s' count '%s'",
 			type.c_str(), dev_name.c_str(), port.c_str(), count.c_str());
		executeEvent("other", line.c_str());
		return 1;
	}
}

int ACPI_Interface::setBatteryAlarm_(){
	
	switch(cur_state.BAT_STATE){
	case BAT_NORM_STATE:
		if (setBatteryAlarm(config->current_scheme->BAT_WARN_LIMIT) < 0)
			return -1;
		pDebug(DBG_INFO, "Alarm set to %d percent\n", 
		       config->current_scheme->BAT_WARN_LIMIT); 
		break;
	case BAT_WARN_STATE:
		if (setBatteryAlarm(config->current_scheme->BAT_LOW_LIMIT) < 0)
			return -1;
		pDebug(DBG_INFO, "Alarm set to %d percent\n", 
		       config->current_scheme->BAT_LOW_LIMIT);
		break;
	case BAT_LOW_STATE:
		if (setBatteryAlarm(config->current_scheme->BAT_CRIT_LIMIT) < 0)
			return -1;
		pDebug(DBG_INFO, "Alarm set to %d percent\n", 
		       config->current_scheme->BAT_CRIT_LIMIT);
		break;
	case BAT_CRIT_STATE:
		pDebug(DBG_INFO, "State is already critical: "
		       "alarm will not be set\n");
		// normally not needed but checked anyway ...
		break;
	case BAT_NONE_STATE:
		// should never happen ...
		return -1;
		break;
	}	
	return 1;
}

/* throttling and thermal control only in ACPI */
void ACPI_Interface::activateSettings(){
	PM_Interface::activateSettings();
	int x;
	/* override BIOS trip points in /proc/acpi/thermal_zone/../trip_points */
	if (thermal_trip_points_supported && config->ENABLE_THERMAL_MANAGEMENT == KERNEL){
		for (x=0; x < MAX_THERMAL_ZONES && config->current_scheme->thermal_zones[x].present; x++){
			setThermalTrippoints(x, config->current_scheme->thermal_zones[x]);
		}
		/* trip points could not be set -> deactivating */
		if (x == 0){
			thermal_trip_points_supported = 0;
			pDebug(DBG_DIAG, "No trip point support");
		}
	}
	if (cooling_mode_supported > 0){
		for (x = 0; x < MAX_THERMAL_ZONES && setCoolingMode(x, config->current_scheme->COOLING_POLICY) > 0; x++);
		if (x == 0){
			cooling_mode_supported = 0;
			pDebug(DBG_DIAG, "Cooling mode is not supported by BIOS or the "
			       "thermal module is not loaded, nag your vendor to implement"
			       " all ACPI specfications.");
		}
	}
	/* throttle CPU if configured  */
	if (config->current_scheme->ALLOW_THROTTLING && config->current_scheme->ALWAYS_THROTTLE){
		throttle(config->current_scheme->MAX_CPU_THROTTLING);
	}
	else
		dethrottle();
}

int ACPI_Interface::suspend_to_ram(){
	if (!((supported_sleeping_states & ACPI_S3)
	      || (supported_sleeping_states & ACPI_S3_BIOS))){
		pDebug(DBG_WARN, "ACPI S3 state not supported by system.");
		return -1;
	}

	pDebug(DBG_INFO, "Set machine into suspend2ram (S3) mode\n");

	executeScript( this->SLEEP_SCRIPT, "suspend2ram" );

	return 1;
}

int ACPI_Interface::standby(){
	if (!(supported_sleeping_states & ACPI_S1)){
		pDebug(DBG_WARN, "ACPI S1 state not supported by system.");
		return -1;
	}

	pDebug(DBG_INFO, "Set machine into standby (S1) mode\n");

	executeScript( this->SLEEP_SCRIPT, "standby");

	return 1;
}

unsigned long ACPI_Interface::getTimestamp(){
	
	timeval t;
	gettimeofday(&t, NULL);
	return (t.tv_sec*1000 + t.tv_usec/1000);
}


int ACPI_Interface::openAcpidSocket(){
	int fd;
	int r;
	struct sockaddr_un addr;
	int fl;
	
	fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (fd < 0) {
		return fd;
	}
	
	memset(&addr, 0, sizeof(addr));
	addr.sun_family = AF_UNIX;
	sprintf(addr.sun_path, "%s", ACPI_EVENTFILE.c_str());
	
	r = connect(fd, (struct sockaddr *)&addr, sizeof(addr));
	if (r < 0) {
		close(fd);
		return r;
	}
	fl = fcntl(fd, F_GETFL);
	fcntl(fd, F_SETFL, fl | O_NONBLOCK);	
	return fd;
}
