#include <errno.h>
#include <signal.h>
#include <sys/sysmacros.h>
#include <linux/apm_bios.h>
#include <sys/ioctl.h>


#include "powerlib.h"
#include "powersaved.h"
#include "apm.h"
#include "fcntl.h"
#include "cpufreq.h"

APM_Interface::APM_Interface(GeneralConfig *config_obj):PM_Interface(config_obj){	
	/* disable temperature control for apm
	   only acpi is supported */
	config->current_scheme->ENABLE_THERMAL_MANAGEMENT = OFF;
	pDebug(DBG_INFO, "APM constructor invoked!\n");
	// in apm the battery is always polled!
	poll_battery = 1;
	/* nothing of this exists in apm */
	cooling_mode_supported = 0;
	throttling_supported = 0;
	thermal_trip_points_supported = 0;
}

APM_Interface::~APM_Interface(){
	close(hwEvent_fd);	
}

int APM_Interface::openHWEventFD(){
	// from apmd ...
	/* Return a file descriptor for the apm_bios device, or -1 if there is an
	 * error. 	 */
	if ((hwEvent_fd = open(APM_BIOS_DEV, O_RDWR)) < 0){
		/* Try to create it. 
		 * Expect major and minor number 10/134
		 * This should work with driver version 1 and above 
		 * (cannot access version number here)
		 * power lib does not provide it -> parse /proc/apm for version
		 * number. Old bios versions may need other major/minor numbers
		 * These can be read out from /proc/devices.
		 * do it yourself and send patch if you have such an old system ...
		 */
		if (mknod(APM_BIOS_DEV, S_IFCHR | S_IRUSR | S_IWUSR, makedev(10, 134))){
			unlink(APM_BIOS_DEV);
			return -1;
		}
		if ((hwEvent_fd = open(APM_BIOS_DEV, O_RDWR)) < 0){
			/* this should never happen */
			pDebug(DBG_ERR, "Could not open APM device "
			       APM_BIOS_DEV ", exit.");
			kill(getpid(), SIGTERM);
		}
	}
	return hwEvent_fd;
}

int APM_Interface::openAcpidSocket(){
	return 1;
}
	
int APM_Interface::handleHWEventRequest(){

	int nr;

	// from /usr/include/linux/apm_bios.h
	/* Be careful this event structure has nothing to do
	   with the powersave events. */
	apm_event_t events[MAX_EVENTS];
	
	nr = read(hwEvent_fd, events, 
		  MAX_EVENTS * sizeof(apm_event_t)) / sizeof(apm_event_t);
	// process events
	for (int x = 0; x < nr; x++){
		// code partly pasted from apmd...
		switch(events[x]){
			case APM_SYS_STANDBY:
			case APM_USER_STANDBY:
				if (sleep_triggered){
					pDebug(DBG_DIAG, "Be careful, endless loop?");
					break;
				}
				pDebug(DBG_INFO, "APM - Standby requested!\n");
				/* may happen if sleep is initiated by HW, then 
				   we use standby events at this point 
				*/
				if (requested_sleep_state == 0){
					pDebug (DBG_DIAG, "Standby requested from BIOS");
					executeEvent("global.standby"); 
				}
				break;
			case APM_SYS_SUSPEND:
			case APM_USER_SUSPEND:
			case APM_CRITICAL_SUSPEND:
				if (sleep_triggered){
					pDebug(DBG_DIAG, "Be careful, endless loop?");
					break;
				}
				pDebug(DBG_INFO, "APM - Suspend2(ram/disk) requested.\n");
				/* may happen if sleep is initiated by HW, then 
				   we use suspend2ram events at this point 
				*/
				if (requested_sleep_state == 0){
					pDebug (DBG_DIAG, "Suspend requested from BIOS");
					executeEvent("global.suspend2ram"); 
				}
				break;
			case APM_STANDBY_RESUME:
				switch (requested_sleep_state){
					case PSD_SUSPEND2DISK:
						pDebug(DBG_DIAG, "Wake up from apm standby, but suspend2disk "
						       "has been triggered, this is no error");
						executeEvent("global.resume.suspend2disk");
						break;
					case PSD_SUSPEND2RAM:
						pDebug(DBG_DIAG, "Wake up from apm standby, but suspend2ram "
						       "has been triggered, this is no error");
						executeEvent("global.resume.suspend2ram");
						break;
					case PSD_STANDBY:
						pDebug(DBG_DIAG, "Wake up from apm standby");
						executeEvent("global.resume.standby");
						break;
					default:
						pDebug(DBG_DIAG, "Resuming from apm standby, no request has been made, "
						       "propably BIOS has initiated sleep state");
						executeEvent("global.resume.standby");
						break;
				}
				break;
				// does not happen, according to current kernel
				// implementation, driver sets time itself
			case APM_UPDATE_TIME:
				pDebug(DBG_INFO, "APM - Time updated "
				       "send from /dev/apm_bios\n");
				break;
				// normally suspend should generate NORMAL_RESUME
				// don't know when a critical resume could occur,
				// all are processed in the same way ...
			case APM_NORMAL_RESUME:
			case APM_CRITICAL_RESUME:
				switch (requested_sleep_state){
					case PSD_SUSPEND2DISK:
						pDebug(DBG_DIAG, "Wake up from apm suspend, suspend2disk "
						       "has been triggerd");
						executeEvent("global.resume.suspend2disk");
						break;
					case PSD_SUSPEND2RAM:
						pDebug(DBG_DIAG, "Wake up from apm suspend, suspend2ram "
						       "has been triggerd");
						executeEvent("global.resume.suspend2ram");
						break;
					case PSD_STANDBY:
						pDebug(DBG_DIAG, "Wake up from apm suspend, but standby "
						       "has been triggerd, this is no error");
						executeEvent("global.resume.standby");
						break;
					default:
						pDebug(DBG_DIAG, "Resuming from apm standby, no request has been made"
						       "propably BIOS has initiated sleep state");
						executeEvent("global.resume.suspend2ram");
						break;
				}
				requested_sleep_state = PSD_NO_SLEEP_REQUEST;
				sleep_triggered = 0;
				// reset frequency which may have changed
				// without noticing it.
				CPUFreq_Interface::reinitSpeeds();
				break;
			case APM_POWER_STATUS_CHANGE:
				pDebug(DBG_INFO, "APM - Power status changed occured,"
				       " checking if that is true...\n");
				checkACStateChanges();
				break;
			case APM_LOW_BATTERY:
				pDebug(DBG_INFO, "APM - Low Battery send from device,"
				       " checking battery.\n");
				checkBatteryStateChanges();
				break;
#ifdef        APM_CAPABILITY_CHANGE
			case APM_CAPABILITY_CHANGE:
				pDebug(DBG_INFO, "APM - Capability  change occured."
				       " Checking whether there is a battery.\n");
				checkACStateChanges();
				checkBatteryStateChanges();
				break;
#endif
			default:
				/* other events are not errors, 
				   see the APM BIOS 1.2 spec or apmd.
				*/
				pDebug(DBG_DIAG, "Received unknown APM event, this"
				       " may not be an error: 0x%04x.", events[x]);
		}
	}
	// never returns an error at the moment
	return 1;
}

void APM_Interface::activateSettings(){

	/* nothing special to do for apm */
	PM_Interface::activateSettings();
}


/* Be careful: this function are invoked in 
   child processes, do not set any variables */

int APM_Interface::suspend_to_ram(){
	sync();
	// resume comes asynchronous through /dev/apm_bios
	pDebug(DBG_INFO, "APM - Execute/Trigger suspend2ram!\n");
	if (ioctl(hwEvent_fd, APM_IOC_SUSPEND, NULL) < 0){
		pDebug(DBG_ERR, "Could not trigger suspend: %s", strerror(errno));
		return -1;
	}
	return 1;
}

/* Be careful: this function are invoked in 
   child processes, do not set any variables */

// In contrast to acpi the app is not sleeping at a
// specific point.
// I will get a notification through apm_bios (handleHWRequest, see above) 
// when resuming, then the global.resume event is thrown.

int APM_Interface::standby(){
	sync();
	pDebug(DBG_INFO, "APM - Execute/Trigger standby!\n");
	// resume comes asynchronous through /dev/apm_bios (see listenEvents)
	if (ioctl(hwEvent_fd, APM_IOC_STANDBY, NULL) < 0){
		pDebug(DBG_ERR, "Could not trigger standby: %s", strerror(errno));
		return -1;
	}
	return 1;
}
