/***************************************************************************
 *
 * brightness.cpp - Brightness Support for powersave daemon
 *
 * Copyright (C) 2005 Timo Hoenig <thoenig@suse.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/utsname.h>
#include <sstream>

/* merge #include "powerlib.h" */
/* merge #include "config_pm.h" */

#include "brightness.h"

/* Probe function to identify kernel interface */

Brightness* Brightness::Probe ()
{
    int fd = 0;

    if ( (fd = open (ACPI_ASUS, O_RDONLY)) > 0) {
        close (fd);
        return new BrightnessASUS ();
    }

    if ( (fd = open (ACPI_IBM, O_RDONLY)) > 0) {
        close (fd);
        return new BrightnessIBM ();
    }

    if ( (fd = open (ACPI_SONY, O_RDONLY)) > 0) {
        close (fd);
        return new BrightnessSony ();
    }

    if ( (fd = open (ACPI_TOSHIBA, O_RDONLY)) > 0) {
        close (fd);
        return new BrightnessToshiba ();
    }

    if (fd) {
        close (fd);
    }

    return new Brightness ();
}


Brightness::~Brightness()
{
}

/* Brightness (dummy object) */

void Brightness::Init ()
{
    return;
}

int Brightness::Get ()
{
    return -1;
}

void Brightness::Set (int level)
{
    return;
}

void Brightness::Min ()
{
    return;
}

void Brightness::Med ()
{
    return;
}

void Brightness::Max ()
{
    return;
}

void Brightness::Up ()
{
    return;
}

void Brightness::Down ()
{
    return;
}

int Brightness::GetLevels ()
{
    return -1;
}


/* Brightness (ASUS object) */

void BrightnessASUS::Init ()
{
    iface = ACPI_ASUS;
    return;
}

int BrightnessASUS::Get ()
{
    char value[3];
    int  level;

    if ((fd = open (iface, O_RDONLY)) < 0) {
        perror (iface);
        level = -1;
        goto out;
    }

    if (read (fd, &value, 2) == -1) {
        level = -1;
        goto out;
    }

    sscanf (value, "%2d", &level);

out:
    if (fd) {
        close (fd);
    }
    return level;
}

void BrightnessASUS::Set (int level)
{
    char value[3];

    if ((fd = open (iface, O_RDWR)) < 0) {
        perror (iface);
        goto out;
    }

    if (level > 15)
        level = 15;
    else if (level < 0)
        level = 0;

    sprintf (value, "%2d", level);
    write (fd, value, 2);

out:
    if (fd) {
        close (fd);
    }
    return;
}

void BrightnessASUS::Min ()
{
    Set (0);
    return;
}

void BrightnessASUS::Med ()
{
    Set (7); 
    return;
}

void BrightnessASUS::Max ()
{
    Set (15);
    return;
}

void BrightnessASUS::Up ()
{
    Set (Get () + 1);
    return;
}

void BrightnessASUS::Down ()
{
    Set (Get () - 1);
    return;
}

int BrightnessASUS::GetLevels ()
{
    return 16;
}


/* Brightness (IBM object) */

void BrightnessIBM::Init ()
{
    iface = ACPI_IBM;
    return;
}

int BrightnessIBM::Get ()
{
    char value[18];
    int  level;

    if ((fd = open (iface, O_RDONLY)) < 0) {
        perror (iface);
        level = -1;
        goto out;
    }

    if (read (fd, &value, 17) == -1) {
        level = -1;
        goto out;
    }

    sscanf (value, "level:%1d", &level);

out:
    if (fd) {
        close (fd);
    }
    return level;
}

void BrightnessIBM::Set (int level)
{
    char value[8];

    if ((fd = open (iface, O_RDWR)) < 0) {
        perror (iface);
        goto out;
    }

    if (level > 7)
        level = 7;
    else if (level < 0)
        level = 0;
    sprintf (value, "level %d", level);
    write (fd, value, 7);

out:
    if (fd) {
        close (fd);
    }
    return;
}

void BrightnessIBM::Min ()
{
    Set (0);
    return;
}

void BrightnessIBM::Med ()
{
    Set (4);
    return;
}

void BrightnessIBM::Max ()
{
    Set (7);
    return;
}

void BrightnessIBM::Up ()
{
    Set (Get () + 1);
    return;
}

void BrightnessIBM::Down ()
{
    Set (Get () - 1);
    return;
}

int BrightnessIBM::GetLevels ()
{
    return 8;
}


/* Brightness (Sony object) */

void BrightnessSony::Init ()
{
    iface = ACPI_SONY;
    return;
}

/* sony_acpi has brightness levels 1 - 8, all others start with Zero.
   So we normalize sony_acpi to 0 - 7 in Get() and Set() */
int BrightnessSony::Get ()
{
    char value[2];
    int  level;

    if ((fd = open (iface, O_RDONLY)) < 0) {
        perror (iface);
        level = -1;
        goto out;
    }

    if (read (fd, &value, 1) == -1) {
        level = -1;
        goto out;
    }

    sscanf (value, "%1d", &level);
    level--;

out:
    if (fd) {
        close (fd);
    }
    return level;
}

void BrightnessSony::Set (int level)
{
    char value[2];

    if ((fd = open (iface, O_RDWR)) < 0) {
        perror (iface);
        goto out;
    }

    if (level > 7)
        level = 7;
    else if (level < 0)
        level = 0;
    sprintf (value, "%1d", level + 1);
    write (fd, value, 1);

out:
    if (fd) {
        close (fd);
    }
    return;
}

void BrightnessSony::Min ()
{
    Set(0);
    return;
}

void BrightnessSony::Med ()
{
    Set(3);
    return;
}

void BrightnessSony::Max ()
{
    Set(7);
    return;
}

void BrightnessSony::Up ()
{
    Set(Get() + 1);
    return;
}

void BrightnessSony::Down ()
{
    Set(Get() - 1);
    return;
}

int BrightnessSony::GetLevels ()
{
    return 8;
}


/* Brightness (Toshiba object) */

void BrightnessToshiba::Init ()
{
    iface = ACPI_TOSHIBA;
    return;
}

int BrightnessToshiba::Get ()
{
    char value[27];
    int  level;

    if ((fd = open (iface, O_RDONLY)) < 0) {
        perror (iface);
        level = -1;
        goto out;
    }

    if (read (fd, &value, 26) == -1) {
        level = -1; 
        goto out;
    }

    sscanf (value, "brightness:%1d", &level);

out:
    if (fd) {
        close (fd);    
    }
    return level;
}

void BrightnessToshiba::Set (int level)
{
    char value[27];

    if ((fd = open (iface, O_RDWR)) < 1) {
        perror (iface);
        goto out;
    }

    if (level > 7)
        level = 7;
    else if (level < 0)
        level = 0;
    sprintf (value, "brightness:%1d", level);
    write (fd, value, 26);

out:
    if (fd) {
        close (fd);
    }
    return;
}

void BrightnessToshiba::Min ()
{
    Set (0);
    return;
}

void BrightnessToshiba::Med ()
{
    Set (4);
    return;
}

void BrightnessToshiba::Max ()
{
    Set (7);
    return;
}

void BrightnessToshiba::Up ()
{
    Set (Get () + 1);
    return;
}

void BrightnessToshiba::Down ()
{
    Set (Get () - 1);
    return;
}

int BrightnessToshiba::GetLevels ()
{
    return 8;
}
