#ifndef POWERSAVE_CONFIG_H
#define POWERSAVE_CONFIG_H

#include <map>
#include <string>
//#include "powersaved.h"
#include "powerlib.h"
#include "event.h"

#ifdef CREATE_DEBUG_OUTPUT
#include <sstream>
#include <iostream>
#endif

using namespace std;

#define MAX_SCHEMES 50
#define MAX_THERMAL_ZONES 10
// TODO: get this from kernel headers
#define ACPI_THERMAL_MAX_ACTIVE 10

#define DEFAULT_CONFIG_DIR "/etc/sysconfig/powersave/";

//#define CONFIG_FILES 

enum CPUFREQ_MODE{_DYNAMIC, _PERFORMANCE, _POWERSAVE};

enum CPUFREQ_CONTROL_MODE{CPUFREQ_USERSPACE, CPUFREQ_KERNEL};

enum THERMAL_MANAGEMENT_DELIGATION { OFF, KERNEL };

class Event;
class GeneralConfig;

// map that contains all EVENT objects. All EVENT objects are initialised
// in PM_Interface constructor.
typedef map<string, Event>::iterator EventIter;

/** @brief Powersave configuration class */
class PS_Config{

 private:

	/** @brief just to avoid thousand of same debug statements */
	void strtointErr(char *str, int line, const char *file);
	void PS_Config::valueError(const std::string what, std::string min, std::string max);
	int PS_Config::checkValue(const std::string what, int min, int max);
	int PS_Config::checkValue(const std::string what, int min);
	int PS_Config::checkValue(int def, const std::string what, int min, int max);
	int PS_Config::checkValue(int def, const std::string what, int min);
	int PS_Config::checkYes(const std::string what);
	int PS_Config::checkYes(int def, const std::string what);
	/** @brief an array that can be accessed by string name */
	typedef std::map<std::string, std::string> StringMap;
	/** @brief holds all configuration variables */
	StringMap data;

 public:

	/* CONFIGURATION VARIABLES FROM CONFIG FILES **********/

    /** @brief cpufreq should be controlled by userspace or kernel */
	CPUFREQ_CONTROL_MODE CPUFREQ_CONTROL;
	
    /** @brief cpu load percentage at which we increase the cpu speed */
	int CPU_HIGH_LIMIT;

	THERMAL_MANAGEMENT_DELIGATION ENABLE_THERMAL_MANAGEMENT;
	/** @brief how long our event loop sleeps (in miliseconds) */
	unsigned long POLL_INTERVAL;
	/** @brief how often the battery gets polled */
	unsigned int BATTERY_POLLING_INTERVAL;
	/** @brief timeout (seconds) after which an additional idle event is thrown */
	unsigned long CPU_IDLE_TIMEOUT;
	/** @brief cpu limit (in percent) under which after CPU_IDLE_TIMEOUT
	 * the idle event is thrown */
	int CPU_IDLE_LIMIT;
	
	/** @brief should niced processes count for cpufreq calculatiob, too? */
	int CONSIDER_NICE;
	
	int USER_SUSPEND2DISK_DISABLED;
	int USER_SUSPEND2RAM_DISABLED;
	int USER_STANDBY_DISABLED;

	int SUSPEND2DISK_DELAY;
	int SUSPEND2RAM_DELAY;
	int STANDBY_DELAY;

	/** @brief absolute brightness level (depends on driver) or:
		   -1 ignore -2 low -3 medium -4 high */
	int DISPLAY_BRIGHTNESS;

	int POWER_BUTTON_DELAY;

	/* Percent of battery limits */
	int BAT_WARN_LIMIT;
	int BAT_LOW_LIMIT;
	int BAT_CRIT_LIMIT;

	/** @brief do we need to poll the battery?
	 * is used in SIGHUP handler function -> public
	 */
	int FORCE_BATTERY_POLLING;

	int JUMP_CPU_FREQ_MAX_LIMIT;
	int CPUFREQ_HYSTERESIS;

	/** these contain the file name of the scheme without leading "scheme_*"
	 * Do not mix up with the SCHEME_NAME variable inside the scheme
	 */
	string BATTERY_SCHEME;
	string AC_SCHEME;
	
	// uses enum (see above)
	int ALLOW_THROTTLING;
	int MAX_CPU_THROTTLING;
	int ALWAYS_THROTTLE;

	CPUFREQ_MODE CPUFREQUENCY;

	int USER_INPUT_TIMEOUT1;
	int USER_INPUT_TIMEOUT2;

	/** @brief uses enum (see above) */
	_COOLING_POLICY COOLING_POLICY;
	
	/** Actions for events as specified in events config file
	   or maybe specialised in a scheme configuration */
	map<string, Event> eventMap;
	/******************************************/

	/** holds all thermal info comes from powerlib.h 
	   is initialised with kernel/BIOS values that could
	   be overridden by config files
	*/
	ThermalDev thermal_zones[MAX_THERMAL_ZONES];

	/** this will only be assigned with valid value for schemes */
	string SCHEME_NAME;

	/* CONFIGURATION VARIABLES FROM CONFIG FILES **********/

	/** @brief the directory where scheme files are found */
	static string scheme_dir;

	PS_Config();
	virtual ~PS_Config();
	
	virtual Event& getEvent( string event_name );	
	virtual int readConfigFile(const std::string &file = "");
	virtual void assignConfigEntries();

};

/** @brief Scheme configuration */
class SchemeConfig: public PS_Config {
	
 private:
	string file_path;
 public:      	
	string file_name;
	/** @brief this is a list... */
	SchemeConfig *next;

      	SchemeConfig(string file_name, const GeneralConfig&);
	virtual ~SchemeConfig();
	virtual int readConfigFile(string file = "");

#ifdef CREATE_DEBUG_OUTPUT
	friend ostream& operator<<(ostream&, SchemeConfig&);
#endif

};

class GeneralConfig: public PS_Config {

 private:
	/** @brief general configs are now split in several files */
	static const string config_files[6];
	/** @brief beginning of scheme list */
	SchemeConfig *scheme_configs;
	/** @brief current activated scheme */
	SchemeConfig *scheme_in_charge;

	/** @brief returns number of initialised schemes or < 0 on error */
	int initSchemes();

	int parseSchemes();

 public:

	/******************************************/

	SchemeConfig* current_scheme;
	SchemeConfig* AC_scheme;
	SchemeConfig* Battery_scheme;

	string config_dir;
	int disable_CPU_freq;
	int no_of_schemes;

	GeneralConfig();
       	virtual ~GeneralConfig();

	/** @brief read the config files
	 *
	 * @return 
	 *     - -1 if no schemes available, have to quit then or we
	 * may access NULL pointer in powersave.cpp:
	 * config->current_scheme, makes no sense to assign a default
	 * scheme
	 */
	int readConfFiles();

	SchemeConfig* getScheme(string scheme_name);

	/** @brief use for iteration to get all schemes... 
	 *
	 * @return pointer to a SchemeConfig
	 * @retval pointer to requested SchemeConfig
	 * @retval NULL if scheme at position no does not exist
	 */
	SchemeConfig* getSchemeByNumber(int no);

	/** @brief sets the active scheme
	 *
	 * pass the name of the scheme as it is defined in POWERSAVE_SCHEME_NAME
	 * 
	 * get the filename (without dir prefix) through the second parameter
	 * 
	 * will be used to inform powersave_proxy (needs file name, not scheme name)
	 *
	 * if something goes wrong file_name = ""
	 *
	 * @return integer representing result of set active scheme
	 * @retval -1 if the scheme does not exist
	 * @retval 0 if scheme is already active
         * @retval 1 if scheme is successfully activated
	 */
	int setActiveScheme(string scheme_name);

	/** @brief returns either Battery or AC scheme
	 *
	 * @return the name of the scheme
	 * @retval "" if no scheme is associated (this would be a bug)
	 */
	string getBatteryScheme();
	string getACScheme();


#ifdef CREATE_DEBUG_OUTPUT
	friend ostream& operator<<(ostream& os, const GeneralConfig& gc);
#endif

};
#endif

