#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/utsname.h>
#include <sstream>

#include "powerlib.h"
#include "cpufreq.h"
#include "config_pm.h"

int CPUFreq_Interface::cpu_max = 90;
int CPUFreq_Interface::cpu_hysteresis = 5;
int CPUFreq_Interface::high_cpu_limit = 0;
int CPUFreq_Interface::consider_nice = 1;
int CPUFreq_Interface::cpus = -1;
CPUFREQ_MODE CPUFreq_Interface::mode = _DYNAMIC;
CPUFREQ_CONTROL_MODE CPUFreq_Interface::control = CPUFREQ_USERSPACE;

map<int, CPUFreq_Kernel> CPUFreq_Interface::cpu_objects_kernel = map<int, CPUFreq_Kernel>();
map<int, CPUFreq_Userspace> CPUFreq_Interface::cpu_objects_userspace = map<int, CPUFreq_Userspace>();


#ifdef CPUFREQ_MEASURE
unsigned long CPUFreq_Interface::polling_interval = 333;
time_t *CPUFreq_Interface::start_time = NULL;
#define MEASURE_LOG_FILE "/tmp/cpufreq.log"
#endif

string CPUFreq_Interface::USERSPACE_STRING   = "userspace";
string CPUFreq_Interface::ON_DEMAND_STRING   = "ondemand";
string CPUFreq_Interface::POWERSAVE_STRING   = "powersave";
string CPUFreq_Interface::PERFORMANCE_STRING = "performance";

/***** H E L P E R   F U N C T I O N S *********************/
/** @brief read a line from a file */
inline 
int read_line(const char * filename, char * line, unsigned len)
{
	FILE * fp = fopen(filename, "r");
	if (!fp){
		pDebug(DBG_ERR, "Could not open file for reading: '%s'", filename);
		return -1;
	}
	if ( (!fgets(line, len, fp)) ){
		pDebug(DBG_ERR, "Could not read from file: '%s'", filename);
		fclose(fp);
		return -1;
	}
	fclose(fp);
	return 1;
}
//governor

/** @brief write a line to a file */
inline 
int write_line(const char * filename, const char *fmt, ...)
{
	FILE * fp = fopen(filename, "w+");
	if (!fp){
		//	    pDebug(DBG_ERR, "Could not open file for writing: %s; %s", filename, strerror(errno));
		return -1;
	}
	va_list ap;
	va_start(ap, fmt);         // get variable argument list passed
	if (vfprintf(fp, fmt, ap) < 0){
		//	    pDebug(DBG_ERR, "Could not write to file: %s", filename);
		fclose (fp);
		return -1;
	}
	va_end(ap);
	fclose(fp);
	return 1;
}

/** @brief read an integer value from a file, error is 0!!! */
inline
unsigned read_value(const char * filename)
{
	char line[MAX_LINE_SIZE];
	if (read_line(filename, line, sizeof line) < 0){
		pDebug(DBG_ERR, "Could not read value in file %s: error: %s!\n", filename, strerror(errno));
		return 0;
	}
	else {
		pDebug(DBG_INFO, "Value: %s read out from file: %s!\n", line, filename);
		return atoi(line);
	}
}
/***** H E L P E R   F U N C T I O N S *********************/


CPUFreq_Interface::CPUFreq_Interface(int sysfs, int cpu){
	
	ostringstream strstr;
	// initialise variables
	if (sysfs){
		strstr << SYSFS_FILES << "cpu" << cpu << "/cpufreq/scaling_min_freq";
		MIN_SPEED_FILE = strstr.str();
		strstr.str("");
		strstr << SYSFS_FILES << "cpu" << cpu << "/cpufreq/scaling_max_freq";
		MAX_SPEED_FILE = strstr.str();
		strstr.str("");
		strstr << SYSFS_FILES << "cpu" << cpu << "/cpufreq/scaling_governor";
		GOVERNOR_FILE = strstr.str();
		strstr.str("");
		strstr << SYSFS_FILES << "cpu" << cpu << "/cpufreq/scaling_available_frequencies";
		AVAILABLE_FREQS_FILE = strstr.str();
		strstr.str("");
	}
	// supported speeds
	
	this->cpu = cpu;
}
	


void CPUFreq_Interface::createCPUFreqObject(CPUFREQ_CONTROL_MODE ctrl, unsigned long sampling_rate) {
	// 2.4.x kernels need also set speed min/max
	// when userspace governor should be set
	// 20 chars should be enough for frequency value
	CPUFreq_Interface::control = ctrl;

	struct utsname uts_struct;
	int x;

	/* only invoke once */
	if (cpus != -1)
		return;
	// only goes inside once !!!
	uname(&uts_struct);

	// test whether necessery files 
	// and modules are loaded for specific kernel
	
	if ((cpus = getCPUCount()) < 0){
		cpus = -1;
		return;
	}

#ifdef OBSOLETE_2_4_KERNEL
	// use sysfs if kernel is higher than 2.4
	// I will not check older kernels explicitly, if files are not there
	// I can be sure that there is no cpufreq support.
        int use_sysfs = (memcmp(uts_struct.release, "2.4", 3) != 0);
#else
        const int use_sysfs = 1;
#endif
        for (x = 0; x < cpus; x++){
                switch(control) {
                        case CPUFREQ_USERSPACE:
                                cpu_objects_userspace[x] = CPUFreq_Userspace(use_sysfs, x);
                                if (cpu_objects_userspace[x].readFrequencies()){
                                        cpu_objects_userspace.clear();
                                        /* indicates no support ! */
                                        cpus = -1;
                                        return;
                                }
                                break;
                        case CPUFREQ_KERNEL:
                                cpu_objects_kernel[x] = CPUFreq_Kernel(use_sysfs, x, sampling_rate);
                                if (cpu_objects_kernel[x].readFrequencies()){
                                        cpu_objects_kernel.clear();
                                        /* indicates no support ! */
                                        cpus = -1;
                                        return;
                                }
                                break;
                }
        }
}
// SYSFS OR NOT ... ********************************************/



int CPUFreq_Interface::getSupportedSpeeds(unsigned *min, unsigned *max){
	// this code is a hack to get the various speed steps supported by the
	// CPU by looping from the maximum speed to the minimum speed and trying
	// to set every possible speed divisible by step! 
	
	*min = read_value((*this).MIN_SPEED_FILE.c_str());
	*max = read_value(MAX_SPEED_FILE.c_str());
	if (*min > 0 && *max > 0){
		pDebug(DBG_DIAG, "Speed min: %u kHz  - Speed max: %u kHz\n", *min, *max);
		return 1;
	}
	else{
		pDebug(DBG_WARN, "Speedstepping not supported! "
		       "Was the correct cpufreq module loaded?\n");
		return -1;
	}
}	

CPUFREQ_MODE CPUFreq_Interface::getMode(){
	return mode;
}

int CPUFreq_Interface::isSupported(){
	int ret = 0;
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			ret = cpu_objects_userspace.empty();
			break;
		case CPUFREQ_KERNEL:
			ret = cpu_objects_kernel.empty();
			break;
	}
	if (ret)
		/* supported */
		return 0;
	else
		return 1;
}

void CPUFreq_Interface::adjustSpeeds(){
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			for (userspace_iter cpu_core = cpu_objects_userspace.begin(); 
			     cpu_core != cpu_objects_userspace.end(); cpu_core++)
				(*cpu_core).second.adjustSpeed(
					(int)getCPULoadSMP(consider_nice, (*cpu_core).first));
			break;
		case CPUFREQ_KERNEL:
// adjustSpeed is a dummy function in cpufreq_kernel
// 			for (kernel_iter cpu_core = cpu_objects_kernel.begin(); 
// 			     cpu_core != cpu_objects_kernel.end(); cpu_core++)
// 				(*cpu_core).second.adjustSpeed(
// 					(int)getCPULoadSMP(consider_nice, (*cpu_core).first));
			break;
	}
}

/**
 * @brief changes the cpufreq policy (powersave, performance, dynamic)
 *
 * @return   0 on success, !0 on failure
 */
int CPUFreq_Interface::setModes(CPUFREQ_MODE mode){
	int ret = 0;
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			for (userspace_iter cpu_core = cpu_objects_userspace.begin(); 
			     cpu_core != cpu_objects_userspace.end(); cpu_core++)
				ret |= (*cpu_core).second.setMode(mode);
			break;
		case CPUFREQ_KERNEL:
			for (kernel_iter cpu_core = cpu_objects_kernel.begin();
			     cpu_core != cpu_objects_kernel.end(); cpu_core++)
				ret |= (*cpu_core).second.setMode(mode);
	}
	return ret;
}

CPUFREQ_CONTROL_MODE CPUFreq_Interface::getControlMode(){
	return CPUFreq_Interface::control;
}


void CPUFreq_Interface::reinitSpeeds(){
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			for (userspace_iter cpu_core = cpu_objects_userspace.begin();
			     cpu_core != cpu_objects_userspace.end(); cpu_core++)
				(*cpu_core).second.reinitSpeed();
			break;
		case CPUFREQ_KERNEL:
// reinitSpeed is a dummy function in coufreq_kernel
// 			for (kernel_iter cpu_core = cpu_objects_kernel.begin();
// 			     cpu_core != cpu_objects_kernel.end(); cpu_core++)
// 				(*cpu_core).second.reinitSpeed();
			break;
	}			
}

void CPUFreq_Interface::rereadFrequencies(){
	int ret = 0;
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			for (userspace_iter cpu_core = cpu_objects_userspace.begin(); 
			     cpu_core != cpu_objects_userspace.end(); cpu_core++)
				ret |= (*cpu_core).second.readFrequencies();
			break;
		case CPUFREQ_KERNEL:
			for (kernel_iter cpu_core = cpu_objects_kernel.begin(); 
			     cpu_core != cpu_objects_kernel.end(); cpu_core++)
				ret |= (*cpu_core).second.readFrequencies();
			break;
	}
	if (!ret)
		adjustSpeeds();
}
 
void CPUFreq_Interface::destroy(){
	freeCPULoadSMPCalls();
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			cpu_objects_userspace.clear();
			break;
		case CPUFREQ_KERNEL:
			cpu_objects_kernel.clear();
			break;
	}
}

int CPUFreq_Interface::setGovernor(const string new_governor){
	char governor[MAX_LINE_SIZE+1];	

	if (write_line(CPUFreq_Interface::GOVERNOR_FILE.c_str(), "%s\n", new_governor.c_str()) < 0)
		return -1;
	// check if governor has been set
	usleep(1000);
	read_line (CPUFreq_Interface::GOVERNOR_FILE.c_str(), governor, MAX_LINE_SIZE);
	if (strstr(governor, new_governor.c_str()))
		return 0;
	else
		return -1;
}

int CPUFreq_Interface::setGovernors(const string new_governor){
	int ret = 0;
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			for (userspace_iter cpu_core = cpu_objects_userspace.begin(); 
			     cpu_core != cpu_objects_userspace.end(); cpu_core++)
				ret |= (*cpu_core).second.setGovernor(new_governor);
			break;
		case CPUFREQ_KERNEL:
			for (kernel_iter cpu_core = cpu_objects_kernel.begin(); 
			     cpu_core != cpu_objects_kernel.end(); cpu_core++)
				ret |= (*cpu_core).second.setGovernor(new_governor);
			break;
	}			
	/* if (ret != 0) -> something wrong */
	return ret;
}

// get the speed limits supported by the CPU
// just reads the min max files, needs not to initialise speed steps
int CPUFreq_Interface::getMinMaxSpeeds(unsigned *min, unsigned *max){

	*min = read_value((*this).MIN_SPEED_FILE.c_str());
	*max = read_value(MAX_SPEED_FILE.c_str());
	if (*min > 0 && *max > 0){
		pDebug(DBG_DIAG, "Speed min: %u kHz  - Speed max: %u kHz\n", *min, *max);
		return 1;
	}
	else{
		pDebug(DBG_WARN, "Speedstepping not supported! "
		       "Was the correct cpufreq module loaded?\n");
		return -1;
	}
}	

void CPUFreq_Interface::setConfigs(int max, int high_cpu, int hyster, int consider){

	if (max < cpu_hysteresis)
		cpu_max = cpu_hysteresis;
	else
		cpu_max = max;

	if (high_cpu <= 0 || high_cpu > 100)
		high_cpu_limit = 0;
	else
		high_cpu_limit = high_cpu;

	if (hyster < 0)
		cpu_hysteresis = 5;
	else
		cpu_hysteresis = hyster;

	consider_nice = consider;
	switch (CPUFreq_Interface::control){
		case CPUFREQ_USERSPACE:
			for (userspace_iter cpu_core = cpu_objects_userspace.begin(); 
			     cpu_core != cpu_objects_userspace.end(); cpu_core++)
				(*cpu_core).second.setConfig();
			break;
		case CPUFREQ_KERNEL:
			for (kernel_iter cpu_core = cpu_objects_kernel.begin(); 
			     cpu_core != cpu_objects_kernel.end(); cpu_core++)
				(*cpu_core).second.setConfig();
			break;
	}			
}
