#ifndef POWERSAVE_CPUFREQ_H
#define POWERSAVE_CPUFREQ_H

#include <map>
#include "powersaved.h"
#ifdef CPUFREQ_MEASURE
#include <time.h>
#endif

/* code in to test available speeds (initFreqsViaTesting())
   is converted from cpuspeed from Carl Thompson
   cpufreq.cc -> getspeeds() */

#define MAX_SPEEDS  20

// SYSFS OR PROC ********************************************/
// cpufreq interfaced is located at /proc/sys/cpu/... on 2.4.x kernels
// and in /sys/devices/system/cpu/...    on 2.6.x kernels

#define SYSFS_FILES \
    "/sys/devices/system/cpu/"

#define PROC_FILES \
    "/proc/sys/cpu/"
#define PROC_USERSPACE \
    "0%0%100%userspace"
// SYSFS OR PROC ********************************************/


/***** H E L P E R   F U N C T I O N S *********************/
int read_line(const char * filename, char * line, unsigned len);
int write_line(const char * filename, const char *fmt, ...);
unsigned read_value(const char * filename);
/***** H E L P E R   F U N C T I O N S *********************/

class CPUFreq_Kernel;
class CPUFreq_Userspace;

class CPUFreq_Interface{
	
 protected:

	/** @brief map that holds cpufreq objects per CPU.
	 *  Key value is the number of the CPU.
	 *  Unfortunately stl containers cannot store abstract
	 *  classes, therefore we have two containers, kernel
	 *  and userspace (there shouldn't be any more subclasses
	 *  in the future ...
	 */
	static map<int, CPUFreq_Kernel> cpu_objects_kernel;
	static map<int, CPUFreq_Userspace> cpu_objects_userspace;
	typedef map<int, CPUFreq_Kernel>::iterator kernel_iter;
	typedef map<int, CPUFreq_Userspace>::iterator userspace_iter;


	/** @brief file where active governor is in */
	string GOVERNOR_FILE;

	/** @brief of the cpufreq governors as they have to be written
	 *  into the scaling_governors file (e.g. "ondemand")
	 */
	static string USERSPACE_STRING;
	static string ON_DEMAND_STRING;
	static string POWERSAVE_STRING;
	static string PERFORMANCE_STRING;

	/** @brief files in /sys/.../cpufreq/ the right string is
	 *  concatenated in the constructor
	 */
	string  MIN_SPEED_FILE;
	string  MAX_SPEED_FILE;
	string  AVAILABLE_FREQS_FILE;

	/** @brief configuration variables
	 * cpu_max: when to switch up
	 * cpu_hysteresis: avoids constant up/down switching
	 * high_cpu_limit: if the load difference is higher than this,
	 *                 go directly for full speed
	 * consider_nice: if niced processes should count, too
         * not all of those work with "kernel" methods.
	 */
	static int cpu_max, cpu_hysteresis;
	static int high_cpu_limit;
	static int consider_nice;

        /** @brief need to now whether we use userspace or
	 *  kernel governors to differ the cpu_objects maps
	 */
	static CPUFREQ_CONTROL_MODE control;

	/** @brief stores the current cpufreq policy 
	 * (powersave, performance, dynamic)
	 */
	static CPUFREQ_MODE mode;

	/* are not really globally needed -> outcommented
	 * if needed somewhere else put the getMinMaxSpeeds() call
	 * from CPUFreq_Userspace::initFreqsViaTesting into
	 * CPUFreq_Interface::CPUFreq_Interface()
	 
	unsigned min;
	unsigned max;
	 */

	/** @brief get the minimum and maximum available speeds
	 *  This is currently only used by the userspace class
	 *  to test frequencies if available_frequencies file
	 *  is not exported by the kernel's low-level driver
	 */
	int getMinMaxSpeeds(unsigned *min, unsigned *max);

	/** @brief reads out frequencies
	 *
	 * @return
	 *     0 if available cpufreq steps
	 *     !0 on error
	 */
	virtual int readFrequencies() = 0;

 public:
	/** @brief adjusts speeds
	 * 
	 * @return
	 *    - -1 if speed has been decreased
	 *    - 1 if speed has been increased
	 *    - 0 if speed has not been modified
	 */
	virtual int adjustSpeed(int CPULoad) = 0;

	/** @brief reinits speeds
	 *
	 * function for suspend workaround.  after resuming the frequency
	 * might have changed and we haven't recognized.  This could
	 * possibly break old PowerNow! machines, but as far as I know
	 * jumping more than one frequency at once is fixed in kernel
	 * module for that machines
	 */
	virtual void reinitSpeed() = 0;

	/** @brief will be invoked for each cpufreq object 
	 *  if setConfigs(...) is invoked. Currently the only the 
	 *  hysteresis array is renewed for the userspace class
	 */
	virtual void setConfig() = 0;

	/** @brief set a new kernel cpufreq governor	*/
	int setGovernor(const string new_governor);

	/** @brief set the cpufreq policy (powersave, performance, dynamic)
	 *  for this CPU
	 */
	virtual int setMode(CPUFREQ_MODE mode) = 0;
      
	/** @brief get the speed limits supported by the CPU
	 *
	 * just reads the min max files, needs not to initialise speed steps
	 * via Testing (not the available_speeds file)
	 */
	int getSupportedSpeeds(unsigned *min, unsigned *max);

	/** @brief set the governor on all CPUs */
	static int setGovernors(const string new_governor);

 public:

	
	CPUFreq_Interface(int sysfs = 1, int cpu = 0);
	virtual ~CPUFreq_Interface() {};

	/** @brief Stores to which CPU this object corresponds
	 *  This is the same as the maps are accessed object_cpus_xxx[cpu] */
	int cpu;

	/** @brief This can be called outside after calling @createCPUFreqObject()
	 *  
	 *  @return 0 if cpufreq is not supported
	 *          1 if cpufreq is working
	 */
	static int isSupported();
	
	/** @brief set cpufreq policy for all CPUs */
	static int setModes(CPUFREQ_MODE mode);
	/** @brief get current cpufreq policy */
	static CPUFREQ_MODE getMode();
	/** @brief no. of available CPUs beginning with 1 */
	static int cpus;

	/** @brief create object
	 *
	 * SYSFS OR NOT ... ********************************************
	 * this function checks which kernel version is used, whether
	 * cpufrequency is supported and creates the right objects
	 *
	 * The objects are only created when cpu freq is really
	 * supported, check with isSupported after calling this function
	 *
	 * @param ctrl -> should cpufreq controlled by userspace or kernel
	 */
	static void createCPUFreqObject(CPUFREQ_CONTROL_MODE ctrl, unsigned long sampling_rate);

	/** @brief Adjusts the speeds of all CPUs to their load (if in dynamic mode)
	 *  This funcion should only be called if the userspace class is used
	 */
	static void adjustSpeeds();

	/** @brief This function can be used if the userspace class thinks it
	 *  run out of sync, what means it has another frequency stored
	 *  that actually is active.
	 *  IIRC this has been used after suspend, but should be obsolute.
	 *  The same for the according reinitSpeed function
	 *
	 *  @ToDo: Delete this one if running stable
	*/
	static void reinitSpeeds();

	/** @brief Reinitialises speeds. This function should be called if the
	 *  frequencies could change unnoticably (e.g. after supsend, AC events, ...)
	 *  This funcion should only be called if the userspace class is used
	 */
	static void rereadFrequencies();

	/** @brief Destroy all objects created with createCPUFreqObject(..) */
	static void destroy();
	
	/** @brief Apply new configs
	 *
	 *  @param high_cpu       -> increase frequency if CPU load exceeds this value
	 *         max            -> directly jump to the highest available frequency
	 *                           if the difference between CPU Load and last
	 *                           measured CPU load exceeds this value (only valid for
	 *                           userspace governor
	 *	   hyster         -> hysteresis value to not oscillate between two
	 *                           frequencies if CPU Load does not change @createHysteresisArray()
	 *	   consider_nice  -> consider niced processes values as CPU load or not
	*/
	static void setConfigs(int high_cpu, int max_limit, int hysters, int consider_nice);

	static CPUFREQ_CONTROL_MODE getControlMode();
};


/** @brief Class to control one CPU's frequency by the powersave daemon itself.
 *         Outside functions don't need to know about this class at all,
 *         just make use of the CPUFreq_Interface() class and pass CPUFREQ_USERSPACE
 *         to its @createCPUFreqObject() function if you want the cpufreq controlled
 *         by this class.
 */
class CPUFreq_Userspace : public CPUFreq_Interface{

	string CURRENT_SPEED_FILE;

	unsigned speeds_kHz[MAX_SPEEDS + 1];
	unsigned demotion[MAX_SPEEDS + 1];

	unsigned current_speed;
	unsigned new_speed;
	
	unsigned last_step;

	int last_cpu_load;

	/** @brief get the current CPU speed -> Error is 0 */
	unsigned getSpeed() ;

	/** @brief set speed to the next higher supported value
	 *
	 * @return integer with result of increase speed
	 * @retval 0 if maximum is already reached
	 * @retval 1 if new speed could be set
	 * @retval -1 if mode is not userspace
	 */
	int increaseSpeed();

	/** @brief see @increaseSpeed() */
	int decreaseSpeed();

	/** @brief set the current CPU speed */
	int setSpeed(unsigned kHz);

	/** @brief sets the speed 
	 *
	 * This code smoothly transitions the CPU speed from 'current' to 'target'
	 * instead of jumping directly to the new speed because some AMD Mobile
	 * Athlon parts seem to choke on large differentials causing kernel panics.
	 */
	void setSpeed(unsigned current, unsigned target);

	void createHysteresisArray();

	/** @brief get the speed limits supported by the CPU
	 *
	 * just reads the min max files, needs not to initialise speed steps
	 */
	int getSupportedSpeeds(unsigned *min, unsigned *max);




	/** @brief reads out frequencies
	 *
	 * @return integer with status of read frequencies
	 * @retval 1 if available cpufreq steps changed
	 * @retval 0 if they are the same
	 * @retval -1 on error
	 */
	int readFrequencies();

	int adjustSpeed(int CPULoad);
        void reinitSpeed();

	int initFreqsViaFile();
	int initFreqsViaTesting();

	/**  
	 * @brief You can use protected functions of the father in children
	 * but not vice versa -> declare the father class as friend
	 * solves this
	 */
	friend class CPUFreq_Interface;

	void setConfig();

	/** @brief set the cpufreq policy to 
	 *
	 *  _PERFORMANCE, _DYNAMIC, _POWERSAVE
	 *
	 * @return
	 *     - 0 on success
	 *     - -1 on failure
	 *
	 * Always test the return value as the kernel
	 * class might loose the powersave or performance governor
	 *
	 * If the call fails the cpufreq classes are destroyed, isSupported
	 * will return false and you must not call any CPUFreq functions any
	 * more (in fact these calls should be just ignored, but better 
	 * be careful)
	 */
	int setMode(CPUFREQ_MODE mode);

 public:
#ifdef CPUFREQ_MEASURE
	unsigned long time_spent[MAX_SPEEDS + 1];
	unsigned int count;
	unsigned long cpu_load_sum;
	
	static time_t *start_time;
	
	static unsigned long polling_interval;

	/** @brief append measure output to /tmp/cpufreq.log statically for now */
	static void startMeasures();
	/** @brief append measure output to /tmp/cpufreq.log statically for now */
	static void stopMeasures();

        /** @brief need polling interval for time measures */
#endif

	CPUFreq_Userspace(int sysfs, int cpu = 0);
	CPUFreq_Userspace(){};
	~CPUFreq_Userspace();
};

/** @brief Class to control one CPU's frequency by a kernel governor.
 *         Outside functions don't need to know about this class at all,
 *         just make use of the CPUFreq_Interface() class and pass CPUFREQ_KERNEL
 *         to its @createCPUFreqObject() function if you want the cpufreq controlled
 *         by this class.
 */

class CPUFreq_Kernel : public CPUFreq_Interface{

	int down_threshold;
//         long sampling_rate;
// 	long sampling_rate_min;
// 	long sampling_rate_max;
	int sampling_down_factor;


	/* all virtual functions from CPUFreq_Interface *************/
	int readFrequencies();
	int adjustSpeed(int CPULoad);
	void reinitSpeed();
	void setConfig();
	int setMode(CPUFREQ_MODE mode);
	/* all virtual functions from CPUFreq_Interface *************/

 public:

	/* 
	 * @brief You can use protected functions of the father in children
	 * but not vice versa -> declare the father class as friend
	 * solves this.
	 */
	friend class CPUFreq_Interface;

	CPUFreq_Kernel(int sysfs, int cpu = 0, unsigned long sampling_rate = 333000);
	CPUFreq_Kernel(){};
	~CPUFreq_Kernel();

 private:

	int writeOndemand(string name, int value);
	unsigned long sampling_rate;
};


#endif /* POWERSAVE_CPUFREQ_H */
