#include <sstream>
#include <fstream>
#include <string>

#include "cpufreq.h"


CPUFreq_Kernel::CPUFreq_Kernel(int sysfs, int cpu, unsigned long rate):CPUFreq_Interface(sysfs, cpu){

	ostringstream strstr;
	string ondemand_files;

	sampling_rate = rate;
	POWERSAVE_STRING = "powersave";
	PERFORMANCE_STRING = "performance";

	if (sysfs){
		if (setGovernor(ON_DEMAND_STRING) < 0){
			/* no userspace governor set */
			pDebug(DBG_ERR, "Could not set ondemand governor,"
			       " seems as if the cpufreq_ondemand module is not loaded.");
			return;
		}
	}
	else {
		/* this is bad in constructor,
		   try to solve this in another way
		   objects must be created
		*/
	}

}

CPUFreq_Kernel::~CPUFreq_Kernel(){
}

int CPUFreq_Kernel::readFrequencies(){

    /* this just returns the maximum frequency step to get the value
     * for downstepping the cpu
     */
    int freq_hi = 0, freq_lo = 0;
    float ratio = 0, tmp_ratio = 0;
    string line;
    string::size_type pos;

    std::ifstream input(CPUFreq_Interface::AVAILABLE_FREQS_FILE.c_str());
    if (!input)
        return -1;

    getline(input, line);
    if (line == "")
        return -1;

    freq_hi = atoi(line.c_str());

    while ((pos = line.find(" ")) != string::npos){
        line = line.substr(pos + 1);
        freq_lo = atoi(line.c_str());
        if (freq_lo && (freq_lo != freq_hi)) {
            tmp_ratio = ((float)freq_hi / (float)freq_lo);
            pDebug(DBG_INFO, "freq_hi: %d freq_lo: %d, ratio: %4.2f", freq_hi, freq_lo, tmp_ratio);
            freq_hi = freq_lo;
            if (tmp_ratio > ratio)
                ratio = tmp_ratio;
        }
    }
    down_threshold = (int)((float)cpu_max / ratio) - cpu_hysteresis;
    pDebug(DBG_DIAG, "cpu_max: %d, ratio: %4.3f, hysteresis: %d => down_threshold ((m/r)-h): %d", cpu_max, ratio, cpu_hysteresis, down_threshold);
    if (down_threshold < 20){
        down_threshold = 20;
        pDebug(DBG_WARN, "down_threshold was too low, set to 20.");
    }
    return 0;
}

int CPUFreq_Kernel::writeOndemand(string name, int value){

    ostringstream strstr("");
    string ondemand_config_file;

    strstr << SYSFS_FILES << "cpu" << cpu << "/cpufreq/ondemand/" << name;
    ondemand_config_file = strstr.str();
    if (access(ondemand_config_file.c_str(), F_OK)) {
        pDebug(DBG_DIAG, "On demand config file '%s' does not exist.",
                          ondemand_config_file.c_str());
        return -1;
    }
    if (access(ondemand_config_file.c_str(), W_OK)){
        pDebug(DBG_WARN, "Could not write on demand config file %s - %s", 
                          ondemand_config_file.c_str(), strerror(errno));
        return 1;
    }
    write_line (ondemand_config_file.c_str(), "%d", value);
    pDebug(DBG_INFO, "wrote '%d' to file '%s'", value, ondemand_config_file.c_str());
    return 0;
}

void CPUFreq_Kernel::setConfig(){
	/* probably forgot some ondemand configs ?

	   Current powersaved configs:
	   int CPUFreq_Interface::cpu_max = 90;
	   int CPUFreq_Interface::cpu_hysteresis = 5;
	   int CPUFreq_Interface::high_cpu_limit = 0;
	   int CPUFreq_Interface::consider_nice = 1;

	   Current ondemand config files:
	   down_threshold
	   sampling_down_factor
	   sampling_rate
	   sampling_rate_max
	   sampling_rate_min
	   up_threshold

	   Current dynamic vs. ondemand config mapping:
           cpu_max       == up_threshold
           sampling_rate == poll_interval * 1000
	*/

        /* up_threshold is always there, even in "old" kernels (2.6.11)
           if not, there is something wrong and we return early. */
        if (writeOndemand("up_threshold", CPUFreq_Interface::cpu_max)) {
            pDebug(DBG_WARN, "Problems with ondemand governor, could not write up_threshold.");
            return;
        }
        /* later kernels don't have down_threshold, since it is auto-calculated.
           but they have ignore_nice which earlier kernels don't                 */
        if (writeOndemand("down_threshold", CPUFreq_Kernel::down_threshold) < 0) {
            writeOndemand("ignore_nice", consider_nice ? 0 : 1);
        }
        writeOndemand("sampling_down_factor", 1); // todo: use define or configure it.
        writeOndemand("sampling_rate", sampling_rate);
	return;
}

int CPUFreq_Kernel::setMode(CPUFREQ_MODE mode){
	
	switch(mode){
		case _DYNAMIC:
			if (setGovernor(ON_DEMAND_STRING) < 0){
				pDebug(DBG_WARN, "Could not set on_demand governor. Is"
				       "the module loaded?");
				return -1;
			}
			setConfig();     // after switching back to ondemand, the config needs to be re-set.
			break;
		case _PERFORMANCE:
			if (setGovernor("performance") < 0){
				pDebug(DBG_WARN, "Could not set performance governor. Is"
				       "the module loaded?");
				return -1;
			}
			break;
			
		case _POWERSAVE:
			if (setGovernor("powersave") < 0){
				pDebug(DBG_WARN, "Could not set powersave governor. Is"
				       "the module loaded?");
				return -1;
			}
			break;
		default:
			pDebug(DBG_WARN, "Unknown cpufreq kernel governor requested");
			return -1;
	}
	CPUFreq_Interface::mode = mode;
	return 0;
}

int CPUFreq_Kernel::adjustSpeed(int CPULoad){
	pDebug (DBG_WARN, "Kernel governor should not invoke %s:", __FUNCTION__);
	return 1;
}


 void CPUFreq_Kernel::reinitSpeed(){
	 pDebug (DBG_WARN, "Kernel governor should not invoke %s:", __FUNCTION__);
 }
