#include <sstream>

#include "cpufreq.h"

CPUFreq_Userspace::CPUFreq_Userspace(int sysfs, int cpu):CPUFreq_Interface(sysfs, cpu){

	ostringstream strstr;

	if (sysfs){
		strstr << SYSFS_FILES << "cpu" << cpu << "/cpufreq/scaling_setspeed";
		CURRENT_SPEED_FILE = strstr.str();
		strstr.str("");
		USERSPACE_STRING = "userspace";
	}
#ifdef OBSOLETE_2_4_KERNEL
	else{
		if (cpu != 0){
			pDebug(DBG_ERR, "SMP under proc fs under"
			       "kernel 2.4.x is not supported");
			//			kill (getpid(), SIGTERM);
		}
		strstr << PROC_FILES << cpu << "/speed-min";
		MIN_SPEED_FILE = strstr.str();
		strstr.str("");
		strstr << PROC_FILES << cpu << "/speed-max";
		MAX_SPEED_FILE = strstr.str();
		strstr.str("");
		strstr << PROC_FILES << cpu << "/speed";
		CURRENT_SPEED_FILE = strstr.str();
		strstr.str("");
		GOVERNOR_FILE = "/proc/cpufreq";
		USERSPACE_STRING = PROC_USERSPACE;
	}
#endif
	/* Cannot check for success here, this can be solved by using
	   an init() function for setting the governor, ..., maybe there is
	   a better solution than introducing a new function?
	*/
	if (setGovernor(USERSPACE_STRING) < 0){
		/* no userspace governor set */
		pDebug(DBG_ERR, "Could not set userspace governor,"
		       " seems as if the cpufreq_userspace module is not loaded.");
		return;
	}
	current_speed = 0;
	last_cpu_load  = 50;
	new_speed = 0;
	
#ifdef CPUFREQ_MEASURE
	for (int x = 0; x <= MAX_SPEEDS;x++)
	    time_spent[x] = 0;
	count = 0;
	cpu_load_sum = 0;
	start_time = new (time_t);
#endif

}

CPUFreq_Userspace::~CPUFreq_Userspace(){
	// uncomment to set to maximum speed on exit
	//	setSpeed(current_speed, 0);        
}

int CPUFreq_Userspace::readFrequencies(){

	if (initFreqsViaFile() < 0)
		if (initFreqsViaTesting() < 0)
			return -1;
	createHysteresisArray();
	pDebug(DBG_DIAG, "Available speeds:\n");
	for (int x = 0; speeds_kHz[x]; x++){
		pDebug(DBG_DIAG, " %2u: %9uKHz\n", x, speeds_kHz[x]);
	}
	return 0;
}


unsigned CPUFreq_Userspace::getSpeed() 
{ 
	return read_value(CURRENT_SPEED_FILE.c_str());
}

int CPUFreq_Userspace::setSpeed(unsigned kHz){
	// userspace already checked in setSpeed(unsigned, unsigned)
	if(write_line(CURRENT_SPEED_FILE.c_str(), "%u\n", kHz) < 0){
		pDebug(DBG_ERR, "Could not set speed to: %u kHz; %s", kHz, strerror(errno));
		return -1;
	}
	else {
		pDebug(DBG_INFO, "Speed set to: %uKHz\n in file %s", kHz, CURRENT_SPEED_FILE.c_str());
		return 1;
	}
}


void CPUFreq_Userspace::setSpeed(unsigned current, unsigned target){

	if (current == target)
		return;
	int delta = (current > target) ? -1 : 1;
	do {
		current += delta;
		setSpeed(speeds_kHz[current]);
	}
	while (current != target);
	current_speed = new_speed;
}

int CPUFreq_Userspace::increaseSpeed(){
	if (mode != _DYNAMIC)
		return -1;

	if (new_speed != 0){
		new_speed--;
	}
	else 
		return 0;
	if (current_speed != new_speed){
		pDebug(DBG_INFO, "current speed: %uKHz     new speed: %uKHz\n",
			  speeds_kHz[current_speed], speeds_kHz[new_speed]);
		setSpeed(current_speed, new_speed);
		current_speed = new_speed;
	}
	return 1;
}

int CPUFreq_Userspace::decreaseSpeed(){
	if (mode != _DYNAMIC)
		return -1;

	if (speeds_kHz[new_speed + 1] != 0){
		new_speed++;
	}
	else 
		return 0;
	if (current_speed != new_speed) {
		pDebug(DBG_INFO, "current speed: %uKHz     new speed: %uKHz\n",
			  speeds_kHz[current_speed], speeds_kHz[new_speed]);
		setSpeed(current_speed, new_speed);
		current_speed = new_speed;
	}
	return 1;
}

int CPUFreq_Userspace::setMode(CPUFREQ_MODE mode){

	/* throw these two lines out to force governor to be set */
	if (mode == CPUFreq_Interface::mode)
		return 0;

	CPUFreq_Interface::mode = mode;

	adjustSpeed((int)getCPULoadSMP(consider_nice, cpu));
	return 0;
}

int CPUFreq_Userspace::adjustSpeed(int CPULoad){
	
	int ret = 0;

	/* DEBUG */
	switch(mode){
	case _DYNAMIC:
		pDebug(DBG_DEBUG, "Dynamic mode\n");
		break;
	case _PERFORMANCE:
		pDebug(DBG_DEBUG, "Performance mode\n");
		break;
	case _POWERSAVE:
		pDebug(DBG_DEBUG, "Powersave mode\n");
		break;
	}

	/* Debug */
	pDebug(DBG_DEBUG, "cpu_max: %d\nhigh_cpu_limit: %d\nconsider_nice: %d\nhysteresis: %d, hysteresis value: %d",
	       cpu_max, high_cpu_limit, consider_nice, cpu_hysteresis, demotion[current_speed]); 
	pDebug(DBG_DEBUG, "Current: %u; current speed: %u MHz, New: %u, new_speed: %u MHz\n", 
	       current_speed, speeds_kHz[current_speed], new_speed, speeds_kHz[new_speed]);
	pDebug(DBG_DEBUG, "CPU load: %d, Last CPU load %d, CPULoad diff: %d, last_step: %d, cpu_max: %d, demotion: %u\n", 
	       CPULoad, last_cpu_load, CPULoad-last_cpu_load, last_step, cpu_max, demotion[current_speed]);
	
	/* Debug */

#ifdef CPUFREQ_MEASURE
	time_spent[current_speed]+=polling_interval;
	count++;
	cpu_load_sum+=CPULoad;
#endif
	if (mode == _DYNAMIC){
		/* directly increase speed to maximum if cpu load jumped */
		if (high_cpu_limit && (CPULoad - last_cpu_load) > high_cpu_limit){
			if (current_speed != 0){
				new_speed = 0;
				setSpeed(current_speed, new_speed);
				current_speed = new_speed;
				pDebug(DBG_INFO, "Speed directly set to max (%d) MHz\n", speeds_kHz[current_speed]);
				ret = 1;
			}
		}
		else if (CPULoad > cpu_max && current_speed > 0){
			increaseSpeed();
			pDebug(DBG_INFO, "Speed increased to (%d) MHz\n", speeds_kHz[new_speed]);
			ret = 1;
		}
		else if(CPULoad < (int)demotion[current_speed] && current_speed < last_step){
			decreaseSpeed();
			pDebug(DBG_INFO, "Speed decreased to (%d) MHz\n", speeds_kHz[new_speed]);
			ret = -1;
		}
		else{
			ret = 0;
			pDebug(DBG_DEBUG, "Speed not changed\n");
		}
	}
	else if (mode == _POWERSAVE && current_speed != last_step){
		new_speed=last_step;
		if (current_speed != new_speed) {
			pDebug(DBG_INFO, "current speed: %uKHz     new speed: %uKHz\n",
			       speeds_kHz[current_speed], speeds_kHz[new_speed]);
			setSpeed(current_speed, new_speed);
			current_speed = new_speed;
		}
		ret = -1;
	}
	else if (mode == _PERFORMANCE && current_speed != 0){
		new_speed = 0;
		if (current_speed != new_speed) {
			pDebug(DBG_INFO, "current speed: %uKHz     new speed: %uKHz\n",
			       speeds_kHz[current_speed], speeds_kHz[new_speed]);
			setSpeed(current_speed, new_speed);
			current_speed = new_speed;
		}
		ret = 1;
	}
	else
		ret = 0;
	last_cpu_load = CPULoad;
	return ret;
}

void CPUFreq_Userspace::createHysteresisArray(){

	if ( last_step > 0 ) {
            for (unsigned i = 0; i < last_step; i++) {
		    demotion[i] = (cpu_max-cpu_hysteresis)*speeds_kHz[i+1]/speeds_kHz[i];
		    pDebug(DBG_DIAG, "Speed: %2u, kHz: %9u, demotion: %3u %%\n", 
			   i, speeds_kHz[i], demotion[i]);
	    }
        }
	demotion[last_step] = 0;
}

void CPUFreq_Userspace::reinitSpeed(){
	setSpeed(speeds_kHz[current_speed]);
	pDebug(DBG_DIAG, "forced speed to %d MHz",speeds_kHz[current_speed]);
	new_speed = current_speed;
}

void CPUFreq_Userspace::setConfig(){

	createHysteresisArray();
	/* no need to set anything, just set each variable
	   it will automatically be active
	*/
}

#ifdef CPUFREQ_MEASURE

 void CPUFreq_Userspace::startMeasures(){
	 for (int x = 0; x < cpus; x++){
		 for (int y = 0; y <= MAX_SPEEDS;y++)
			 cpu_objects[x]->time_spent[y] = 0;
		 cpu_objects[x]->count = 0;
		 cpu_objects[x]->cpu_load_sum = 0;
	 }
	 time(start_time);
	 struct tm *tm = new (struct tm);
	 tm = localtime(start_time);

	 FILE *logFile = fopen(MEASURE_LOG_FILE, "a+");
	 if (!logFile){
		 pDebug(DBG_ERR, "Could not open log file %s", MEASURE_LOG_FILE);
		 exit(1);
	 }
	 fprintf(logFile, "\nStart cpufreq measures at: %d.%d.%d at %d:%d:%d\n",
		 tm->tm_mday, tm->tm_mon, tm->tm_year, tm->tm_hour, tm->tm_min, tm->tm_sec);
	 fclose(logFile);
	 free (tm);
 }
 
 void CPUFreq_Userspace::stopMeasures(){
	 FILE *logFile = fopen(MEASURE_LOG_FILE, "a+");
	 if (!logFile){
		 pDebug(DBG_ERR, "Could not open log file %s", MEASURE_LOG_FILE);
		 exit(1);
	 }
	 time_t *stop_time = new (time_t);
	 tm *tm = new (struct tm);
	 time(stop_time);
	 tm = localtime(stop_time);
	 unsigned long measure_time = *stop_time - *start_time;
	 for (int x = 0; x < cpus; x++){
		 for (int k=0;cpu_objects[x]->speeds_kHz[k] != 0;k++){
			 fprintf(logFile, "%d\t: %d minutes %d seconds\n", 
				 cpu_objects[x]->speeds_kHz[k], 
				 (cpu_objects[x]->time_spent[k]/60000),
				 (cpu_objects[x]->time_spent[k]/1000)%60);
		 }
		 if (cpu_objects[x]->count > 0)
			 fprintf(logFile, "Average CPU usage: %##f - Count: %d\n", 
				 (double)cpu_objects[x]->cpu_load_sum/cpu_objects[x]->count, cpu_objects[x]->count);
	 }
	 fprintf(logFile, "\nStop cpufreq measures at: %d.%d.%d at %d:%d:%d\n",
		 tm->tm_mday, tm->tm_mon, tm->tm_year, tm->tm_hour, tm->tm_min, tm->tm_sec);

	 fprintf(logFile, "Total measure time: %d minutes %d seconds - Check: %d minutes %d seconds\n", 
		 measure_time/60, (measure_time)%60,
		 (cpu_objects[0]->count*polling_interval/1000)/60, (cpu_objects[0]->count*polling_interval/1000)%60);
	 fclose(logFile);
	 free(tm);
	 free(stop_time);
 }
#endif

int CPUFreq_Userspace::initFreqsViaFile(){

	int speed_no = 1, offset = 0, sort_temp = 0, len;
	char speeds_line[MAX_LINE_SIZE+1];
	char *temp;

	if (read_line(AVAILABLE_FREQS_FILE.c_str(), speeds_line, MAX_LINE_SIZE) < 0)
		return -1;
	temp = speeds_line;
	len = strlen(speeds_line);
	if (sscanf (speeds_line, "%d", &speeds_kHz[0]) == 1){
		while ((temp = strstr(speeds_line + offset, " ")) != NULL ){
			offset = temp - speeds_line;
			if (sscanf (speeds_line + offset, "%d", &speeds_kHz[speed_no]) != 1
			    || speeds_kHz[speed_no] == 0)
				break;
			speed_no++;
			offset++;
			if (offset >= len)
				break;
		}
	}
	if (speed_no == 0){
		speeds_kHz[0] = 0;
		last_step = 0;
		return -1;
	}

	last_step = speed_no-1;

	speeds_kHz[speed_no] = 0;
	/* sort speeds in right order (fastest first) *****************/
	for (int x = 0; x < speed_no && speeds_kHz[x] > 0; x++){
		for (int y = x; y < speed_no && speeds_kHz[y]; y++){
			if (speeds_kHz[x] < speeds_kHz[y]){
				sort_temp = speeds_kHz[x];
				speeds_kHz[x] = speeds_kHz[y];
				speeds_kHz[y] = sort_temp;
			}
		}
	}
	/* sort speeds in right order (fastest first) *****************/
	setSpeed(speeds_kHz[0]);
	current_speed = 0;
	return speed_no;
}

int CPUFreq_Userspace::initFreqsViaTesting(){
	unsigned int min, max;

	if (getMinMaxSpeeds(&min, &max) < 0){
		/* Place this somewhere else where we could get out */
		pDebug(DBG_WARN, "Could not read min and max speed");
		return -1;
	}

	// initialise speed steps that can be set
	const unsigned step = 25000;
	/* Go to max speed if we are not already there  */
	for (unsigned current = getSpeed(); current < max; current += step)
		setSpeed(current);
	/* loop CPU from the maximum speed to the minimum speed and try
	   to set every possible speed divisible by step! */
	speeds_kHz[0] = max;
	unsigned current_init_speed = 0;
	for (unsigned current = max - step; current > min - step; current -= step) {
		setSpeed(current);
		unsigned real = getSpeed();
		if (real != speeds_kHz[current_init_speed]) {
			speeds_kHz[++current_init_speed] = real;
			if (current_init_speed + 1 == MAX_SPEEDS)
				break;
		}
	}
	speeds_kHz[current_init_speed + 1] = 0;
	// go back to maximum speed because program expects to start there 
	setSpeed(current_init_speed, 0);
	// the last step is the lowest found speed 
	last_step = current_init_speed;
	
	return 1;
}
