#ifndef POWERSAVE_EVENT_H
#define POWERSAVE_EVENT_H

#include <map>
#include <string>
#include <ext/slist>

#include "config_pm.h"


using namespace __gnu_cxx;
using namespace std;

#define PASS_EVENT_TYPE 1
#define PASS_CURRENT_SCHEME 2
#define PASS_ACPI_EVENT 4

class PM_Interface;

enum eventStatus{ EVENT_SUCCESS, EVENT_FAILED, EVENT_WAITING, EVENT_UNKNOWN };

class Event {
 private:
	/** scripts are executed with negativ nice factor (higher
	 * priority) to avoid race conditions and provide fast
	 * suspend/standby maybe these should be configurable somewhen ...
	 */
	static const int proxy_nice_factor = -20;

	slist<string> programs_2_execute;

	/** @brief execute an internal action
	 *
	 * @return integer with result of execute requested internal action
	 * @retval 1 if the action could be found
	 * @retval -1 if this was not an internal action
	 */
	int executeInternalAction(PM_Interface *pm_object, const string action);

	// the following is needed for managing events--------------------
	/** @brief a unique id, assigned by EventManagement */
	int id;
	/** @brief the timeout in seconds before the event is considered as failed */
	time_t timeout;
	/** @brief the filedescriptor of the client if a  client requested this event */
	int clientFD;
	/** @brief enum of states in which the event can be */
	eventStatus state;
	/** @brief the starttime */
	long starttime;
	/** @brief the time how long we slept while this event is running */
	unsigned long sleep_time;
	//end of managing part--------------------------------------------

 public:
	Event();
	Event(string n);
	Event(const Event&);
	Event& operator=(Event&);
	~Event();
	string name;

	/** @brief the directory where executables are that will be
	   executed on events */
	static string scripts_dir;

	/** @brief execute only one external program in the list of pending events 
	 * 
	 * I need a whole PM_Interface object here 
	 * to get the current scheme name as a parameter for the invoked script
	 *  and for internal actions, maybe this could be solved nicer ...
	 */
	int execute_program(PM_Interface *pm_object, string acpi_event_line="");

	/** @brief set the actions to be executed by this event
	 *
	 * @return the execute variable as it used in config files.
	 *
	 * rule: temperature.hot -> TEMPERATURE_HOT
	 *
	 * Add a POWERSAVE_EVENT_ in front of the return value -> this
	 * is the variable as used in the config file to set which programs
	 * should be executed on the event. 
	 */
	void setActions2Execute(const string actions_line);

	/** @brief checks whether the given action is internal
	 *
	 * @return integer value with result of the check
	 * @retval 1 if action is an internal action
	 * @retval -1 if not
	 */
	int isInternalAction(const string action);

	// the following is needed for managing events--------------------
	/** @brief returns the num state of the event */
	eventStatus getState();

	void setState(eventStatus);

	// all the following functions are for setting and getting the different
	// properties

	/** @brief sets the timeout of an event */
	void setTimeout(unsigned long);

	/** @brief gets the timeout of an event */
	unsigned long getTimeout();

	/** @brief starts the timer */
	void start();

	/** @brief adds time to sleep_time */
	void increaseSleeptime(unsigned long time);

	
//	void setClientFD(int fd);
//	int getClientFD();

	/** @brief sets the id of the event */
	void setID(int id);

	/** @brief returns the event id */
	int ID();

	/** @brief returns the amount of time an event is running already */
	unsigned long timeElapsed();

	slist<string>::iterator end();

	/** @brief an iterator which points to the current script in programs_2_execute */
	slist<string>::iterator current_program_2_execute;
	//end of managing part-------------------------------------------


#ifdef CREATE_DEBUG_OUTPUT
	friend ostream& operator<<(ostream&, const Event&);
#endif
};


#ifdef NOT_DEF
/*************  HASH TO FIND INTERNAL ACTIONS FAST *******************/

struct eqstr{
	bool operator()(const char* s1, const char* s2){
		return strcmp(s1, s2) == 0;
	}
};


/** @brief class to find internal actions fast over hash_set
   does not work yet */
class Internal_Actions_Hash{
 private:
	hash_set<const char *, hash<const char*>, eqstr> internal_actions;
 public:
	Internal_Actions_Hash(){
		internal_actions.insert("ignore");
		internal_actions.insert("throttle");
		internal_actions.insert("dethrottle");
		internal_actions.insert("do_suspend_to_disk");
		internal_actions.insert("do_suspend_to_ram");
		internal_actions.insert("do_standby");
	}
	int isInternalAction(const char* x);	
};

/** @brief checks if an action is internal
 *
 * @return integer value with result of the check
 * @retval 1 if param is an internal action
 * @retval -1 if not
 */
int Internal_Actions_Hash::isInternalAction(const char* x){

	const hash_set<const char*, hash<const char*>, eqstr>::const_iterator it 
		= internal_actions.find(x);
	if (it != internal_actions.end())
		return 1;
	else
		return -1;
}
/*************  HASH TO FIND INTERNAL ACTIONS FAST *******************/
#endif

#endif
