#include <sys/wait.h>
#include <sys/fcntl.h>

#include <getopt.h>
#include <errno.h>
#include <stdarg.h>

#include <fstream>
#include <sstream>

#include "powerlib.h"
#include "powersaved.h"
#include "acpi.h"
#include "apm.h"
#include "cpufreq.h"

#include <syslog.h>
#include <sys/time.h>

#include "config_pm.h"
#include "event_management.h"

// global object, on this the whole app runs
// must be declare here to use it in handler functions
PM_Interface *pm;

string PM_Interface::INT_SCRIPT_PATH = "/usr/lib/powersave/";
string PM_Interface::SLEEP_SCRIPT = "do_acpi_sleep";

/* apm workaround variables for sleep modes ******************/
PSD_SLEEP_STATE_REQUESTS PM_Interface::requested_sleep_state = PSD_NO_SLEEP_REQUEST;
int PM_Interface::sleep_triggered = 0;
/* apm workaround variables for sleep modes ******************/

/* send by child after resume from sleep is finished
   used to know when ACPI button events should be ignored
*/

#ifdef CPUFREQ_MEASURE

void RT_START_handler(int){
	FILE *logFile = fopen(MEASURE_LOG_FILE, "a+");
	if (!logFile){
		pDebug(DBG_ERR, "Could not open log file %s", MEASURE_LOG_FILE);
		exit(1);
	}
	if (CPUFreq_Interface::isSupported()){
		fprintf(logFile, "\nSTART ---------------------\n");
		if (pm){
			fprintf(logFile, "\nVARIABLES:\nPolling Interval: %lu ms, Consider Nice: %s\n"
				"Jump directly to highest speed on %d percent CPU load difference\n\n",
				pm->config->POLL_INTERVAL, (pm->config->CONSIDER_NICE ? "yes" : "no"),
				pm->config->JUMP_CPU_FREQ_MAX_LIMIT);
		}
		fclose(logFile);
		CPUFreq_Interface::startMeasures();
	}
	else{
		fprintf(logFile, "No CPU frequency support.\n");
		fclose(logFile);
	}
}

void RT_STOP_handler(int){

	FILE *logFile = fopen(MEASURE_LOG_FILE, "a+");	
	if (!logFile){
		pDebug(DBG_ERR, "Could not open log file %s", MEASURE_LOG_FILE);
		exit(1);
	}

	if (CPUFreq_Interface::isSupported()){
		CPUFreq_Interface::stopMeasures();
		fprintf(logFile, "STOP ---------------------\n");
	}
	else
		fprintf(logFile, "No CPU frequency support.\n");
	fclose(logFile);
}

#endif

/* we ignore this one for security reasons
void brokenpipe_handler(int) {
	pDebug( DBG_DEBUG, "Sigpipe received. Maybe one of the clients is not "
		"available anymore\n");
}
*/

void USR1_handler(int){
	if (pm){
		pm->resume();
	}
}

void hup_handler(int){
	if (!pm){
		pDebug(DBG_ERR, "Could not re-read configs, there "
		       "is no main object, this should not happen\n");
	}
	else{
		pDebug(DBG_INFO, "HUP signal occured, going to re-read configs\n");
		if (pm->rereadConfig() < 0){
			fprintf(stderr, "Not all config files could be read in HUP signal handler, exit\n");
		}
		// check whether force battery polling could be disabled in this function
		pm->checkBatteryStateChanges();
	}
}

int PM_Interface::rereadConfig(){
	int ret = config->readConfFiles();
	
	/* reset/reread everything that could have been changed in
	   configs ... */
	if (CPUFreq_Interface::isSupported()){
		CPUFreq_Interface::setConfigs(config->current_scheme->CPU_HIGH_LIMIT, 
					      config->current_scheme->JUMP_CPU_FREQ_MAX_LIMIT,
					      config->current_scheme->CPUFREQ_HYSTERESIS,
					      config->current_scheme->CONSIDER_NICE);
	}
	activateSettings();
	return ret;
}

// I currently destroy the object I am still working on.
// Don't know yet if this is proper but should free all resources
void exit_handler(int){
	// unlink files here, I do not know if there is already a pm object!
	unlink (POWERSAVED_PID_FILE_PATH);
	if (!pm){
		// this is not an error!
		pDebug(DBG_WARN, "Main object not created yet,"
		       "prog seems to be terminated shortly after starting\n");
		exit(0);
	}
	else{
		delete pm;
		exit(0);
	}
}

void usage(){
	fprintf(stderr,	"Usage: powersaved [OPTIONS] \n\
       where OPTIONS can be :\n\
            -f --acpi-file FILE\n\
                 sets the acpi file the daemon can receive ACPI events from.\n\
                 This could either be /proc/acpi/events or the acpi daemon's socket file\n\n\
            -d --daemonize\n\
                 run as a daemon\n\n\
            -x --scripts-dir\n\
                 sets scripts directory where executables that are triggered on \n\
                 events are found (default: /usr/lib/powersave/scripts/) \n\n\
            -c --config-dir dir\n\
                 sets the config directory (default is /etc/sysconfig/powersave/)\n\n\
            -h --help\n\
                 print this help\n\n\
            -v --verbosity value\n\
                 valid values are from 0-15. \n\
                 1-Error, 2-Warning, 4-Diag, 8-Info and combinations \n\n");
}

int daemonize() {
	int fdlimit;
	switch (fork())	{
	case 0:  break;
	case -1: return -1;
	default: _exit(0);          /* exit the original process */
	}

	if (setsid() < 0)               /* shoudn't fail */
		return -1;

	switch (fork()){
	case 0:  break;
	case -1: return -1;
	default: _exit(0);
	}
	chdir("/");
	// close all fds
	fdlimit = sysconf(_SC_OPEN_MAX);

	for (int fd = 0; fd < fdlimit; fd++)
		close(fd);
	open("/dev/null",O_RDWR);
	dup(0); dup(0);
	return 0;
}

PM_Interface::PM_Interface(GeneralConfig *config_obj){
       	pDebug(DBG_DEBUG, "Constructor PM_Interface\n");


	// initialise everything we have in the constructor !
	// acpi/apm specifc things are initialised in the
	// appropriate ACPI/APM_Interface constructor

	// **** I N I T   S E C T I O N *************************/
	
	config = config_obj;
	// this should never happen, hopefully the term signal cleans up memory and stops app */
	if (!config || !config->current_scheme || !config->AC_scheme || !config->Battery_scheme){
		pDebug (DBG_ERR, "Current, AC or Battery scheme is not set.\n");
		kill (getpid(), SIGTERM);
	}
	
	supported_sleeping_states = getSupportedSleepStates();
	
	processor_event_thrown = NO_PROCESSOR_EVENT_THROWN;
	msecs_last_battery_read = 0;
	msecs_cpu_idle = 0;
	cooling_mode_supported = 1;
	thermal_trip_points_supported = 1;
	throttling_supported = 1;
	
	daemon_pid = getpid();
	// this will be != NULL if cpufrequency scaling is supported
	// else the mode will be by default _DYNAMIC!
	//
	if (!config->disable_CPU_freq){
		CPUFreq_Interface::createCPUFreqObject(config->CPUFREQ_CONTROL, config->POLL_INTERVAL * 1000);
		if (CPUFreq_Interface::isSupported()){
			CPUFreq_Interface::setConfigs(
				config->current_scheme->CPU_HIGH_LIMIT, 
				config->current_scheme->JUMP_CPU_FREQ_MAX_LIMIT, 
				config->current_scheme->CPUFREQ_HYSTERESIS,
				config->current_scheme->CONSIDER_NICE);
		}
		else{
			pDebug (DBG_ERR, "Cannot create cpufreq objects");
		}
	}

	// allocate memory for event management object
	eM = new EventManagement();
	
	// create a new client connection object
	cC = new clientConnection();

	brightness = Brightness::Probe();
	brightness->Init ();

	/* this also sets current scheme, do not change order */
	updateACState();
	activateScheme(cur_state.AC_STATE);
	// intialise battery state. If no battery and no cpufreq do not poll at all!
	cur_state.BAT_STATE = BAT_NORM_STATE;
	if (updateBatteryState() < 0){
		pDebug(DBG_INFO, "No battery available.\n");
	}
	else{
		pDebug(DBG_INFO, "There is at least one battery available.\n");
		// reset battery state, so the first checkBatteryChanges throws a battery
		// event if necessary
		cur_state.BAT_STATE = BAT_NORM_STATE;
	}
	/*******************************************************************/
};


PM_Interface::~PM_Interface(){

	pDebug(DBG_DEBUG, "PM Destructor: delete allocated Objects\n");

	executeEvent("daemon.terminate");
	sleep(1);
	delete (cC);
	delete (eM);
	delete (brightness);
	CPUFreq_Interface::destroy();
	delete config;
	unlink (POWERSAVED_PID_FILE_PATH);
}

int PM_Interface::start(){

	fd_set fd_set_reading;
	fd_set temp_fd_set;
	struct timeval req, req_init;
	int ret;
	bool acpidDied = false;
	bool DBusRefetchFD = false;
	FD_ZERO(&fd_set_reading);

	// fd is member variable!
	if (openHWEventFD() < 0){
		return -1;
	}
	max_FD_value = max(hwEvent_fd, max_FD_value);
	FD_SET(hwEvent_fd, &fd_set_reading);

	//	dbus_server.openSystemConnection(PM_Interface::handleDBusRequest);
	dbus_server.openSystemConnection();
	/* wait for connection to be established
	   ToDo: make this asynchronous by invoke
	         global function in add_watch callback ?\
	*/
	sleep(1);
	//	dbus_server.update_fds(fd_set_reading);
	int dbus_fd = dbus_server.getFd();
	if (dbus_fd < 0){
		pDebug(DBG_ERR, "Could not establish DBus connection");
		exit(255);
	}
	else{
		max_FD_value = max(dbus_fd, max_FD_value);
		FD_SET(dbus_fd, &fd_set_reading);
	}


	// ********  trigger start event ****************/
	// I do it exactly here, because the socket is already
	// open and people can use powersave commands that need
	// a running powersaved(connection to socket)

	// ********  trigger start event ****************/
	executeEvent("daemon.start");
	// Warning: when we start, the scheme should change from "nothing" to "active
	// scheme", a daemon.scheme.change event should be thrown and activateSettings
	// should be called. In reality, it changes from "Performance" to "active scheme",
	// which on AC is often "Performance" => no change => no event => no activate.
	// Not fatal, but we should fix it. In the meantime, we work around it here:
	activateSettings();
	executeEvent("daemon.scheme.change");
	// if the battery is already low, throw an appropriate event.
	checkBatteryStateChanges();

	struct timeval tv_start;
	struct timeval tv_stop;
	struct timezone dummy;

	switch (config->CPUFREQ_CONTROL){
		case CPUFREQ_USERSPACE:
			req_init.tv_sec  = config->current_scheme->POLL_INTERVAL / 1000;
			req_init.tv_usec = config->current_scheme->POLL_INTERVAL * 1000 % 1000000;
			break;
                default:    // CPUFREQ_KERNEL
			req_init.tv_sec  = config->current_scheme->BATTERY_POLLING_INTERVAL;
			req_init.tv_usec = 0;
			break;
	}

	req.tv_sec  = req_init.tv_sec;
	req.tv_usec = req_init.tv_usec;
	gettimeofday(&tv_start, &dummy);
	for (;;){
		do{
			temp_fd_set=fd_set_reading;

			// if no cpufreq and no battery is supported do not poll at all!
			if ((cur_state.BAT_STATE == BAT_NONE_STATE && 
			    !CPUFreq_Interface::isSupported() && !acpidDied) ||
			    (!config->current_scheme->FORCE_BATTERY_POLLING &&
			    (config->CPUFREQ_CONTROL == CPUFREQ_KERNEL))) {
				pDebug(DBG_DEBUG, "Select waiting without timeout, polling disabled.");
				ret = select (max_FD_value+1, &temp_fd_set, NULL, NULL, NULL);
				pDebug(DBG_INFO, "Select came out");
			}
			else{
				pDebug(DBG_DEBUG, "Select waiting: %u secs"
				       " and %u millisecs\n",
				       (unsigned)req.tv_sec, 
				       (unsigned)req.tv_usec / 1000);
				ret = select (max_FD_value+1, 
					      &temp_fd_set, NULL, NULL, &req);
			}
			if (ret < 0){
				if (errno == EINTR){
					pDebug(DBG_DEBUG, 
					       "Select returned with EINTR, continue\n");
					continue;
				}
				// be careful after suspend select may return x < 0, but errno is not set!
				// for now I ignore the set file descriptor(for not waiting endlessly on socket
				// and continue. Maybe we should try a poll implementation here later...
				else{
					struct stat buf;
					if (fstat(dbus_fd, &buf) < 0){
						// We could also check errno -> remove it in every
						// case is probably the safest way...
						// DBus broke away, clear it out of fds to listen on
						DBus_Server::DBusDied = true;
						FD_CLR(dbus_fd, &fd_set_reading);
						pDebug(DBG_ERR, "DBus connection broke away,"
						       " no client communication possible\n");
						continue;
					}
					pDebug(DBG_WARN, "Select returned an Error, "
					       "but not interrupted through a signal: %s"
					       " This is not nice, but normal after a suspend2disk\n", strerror(errno));
					if (FD_ISSET(hwEvent_fd, &temp_fd_set)){
						pDebug(DBG_DEBUG, "hw FD %d is set, ignore it, see warning above.\n", hwEvent_fd);
					}
					else if (FD_ISSET(dbus_fd, &temp_fd_set)){
					        pDebug(DBG_DEBUG, "Dbus FD %d is set, ignore it, see warning above.\n", dbus_fd);
					}
					else
						pDebug(DBG_DEBUG, "No fd is set\n");
					continue;
				}
			}
			else
				break;
		}while(1);

		// check all events if one reached the timeout
		checkEventTimeouts();

		/* measure time that functions need to process to substract from polling interval */
		gettimeofday(&tv_start, NULL);

		if (!DBus_Server::DBusDied && FD_ISSET(dbus_fd, &temp_fd_set)) {
			pDebug(DBG_INFO, "Handling dbus request on fd: %d", dbus_fd);
			if (dbus_server.handleRequest() != 0){
				// This is not the normal way (not sure if this ever could happen
				// Normally we get an error above (invalid file descriptor in select)
				DBus_Server::DBusDied = true;
				FD_CLR(hwEvent_fd, &fd_set_reading);
				pDebug(DBG_ERR, "DBus connection broke away,"
				       " no client communication possible\n");
			}
			continue;
		}
		// This shouldn't be placed here.
		// However, we need proper time out handling in emergency case.
		// Need total rework of the select loop
		// Connection to DBus comes up at least after 2 ACPI events now.
		if ( DBus_Server::DBusDied && DBusRefetchFD) {
			dbus_fd = dbus_server.getFd();
			if (dbus_fd < 0){
				// better close connection again and reopen it below
				// something went badly wrong, maybe we came out of the
				// select too fast to let DBus initialise our connection,
				// try it again on next select cycle...
				pDebug(DBG_DIAG, "Could not initialise connection correctly,"
				       " will try again later.");
			}
			else{
				pDebug(DBG_ERR, "DBus connection established again"
				       "powersaved is fully functional");
				FD_SET(dbus_fd, &fd_set_reading);
				DBusRefetchFD = false;
				DBus_Server::DBusDied = false;
			}
		}
		if ( DBus_Server::DBusDied ) {
			// try to reopen dbus connection
			// fetch fd on next select cycle, we cannot sleep here.
			// this is bad style, need to reimplement select loop properly after 10.0
			if (dbus_server.openSystemConnection() >= 0){
				// succeeded
				// get file descriptor after next select, we need to wait
				pDebug (DBG_DIAG, "DBus connection established again");
				DBusRefetchFD = true;
			}
			else {
				pDebug(DBG_DIAG, "DBus still not responding");
			}
		}

		/* ignore ACPI events if we went to sleep or the machine will possibly shutdown
		   because of a pressed power button. Watch out the suspend/standby functions in 
		   acpi.cpp where the value for this variable is set to 1*/
		if ( !acpidDied && FD_ISSET(hwEvent_fd, &temp_fd_set)){
			pDebug(DBG_DEBUG, "hw FD %d is set, handling it ...\n", hwEvent_fd);
			// acpid socket died ?
			if (handleHWEventRequest() == -2) {
				acpidDied = true;
				close(hwEvent_fd);
				FD_CLR(hwEvent_fd, &fd_set_reading);
				pDebug (DBG_ERR, "acpid socket died\n");
			}
			continue;
		} 
		else
			pDebug(DBG_DEBUG, "Timeout.\n");


			
		if ( acpidDied ) {
			// try to reopen the acpid socket
			if ( (hwEvent_fd = openAcpidSocket())<0) {
				// error should already be written in openAcpidSocket
			}
			// succeeded
			else {
//				std::cerr << "acpid there again\n";
				FD_SET(hwEvent_fd, &fd_set_reading);
				acpidDied = false;
			}
		}

		pDebug(DBG_DEBUG, "Shall I poll Battery?: FORCE_BATTERY_POLLING: %d; msecs_last_battery_read:"
		       " %u, battery_poll_interval (sec): %u\n",
		       config->current_scheme->FORCE_BATTERY_POLLING, msecs_last_battery_read,
		       config->current_scheme->BATTERY_POLLING_INTERVAL);

		/* the check for CPUFREQ_KERNEL below is because with CPUFREQ_KERNEL we check
		   battery after every select timeout */
		if (config->current_scheme->FORCE_BATTERY_POLLING &&
		    cur_state.BAT_STATE != BAT_NONE_STATE &&
		    (config->CPUFREQ_CONTROL == CPUFREQ_KERNEL ||
		    msecs_last_battery_read > config->current_scheme->BATTERY_POLLING_INTERVAL * 1000)) {
			checkACStateChanges();
			checkBatteryStateChanges();
			msecs_last_battery_read = 0;
		}
		else if (config->current_scheme->FORCE_BATTERY_POLLING &&
			 cur_state.BAT_STATE != BAT_NONE_STATE){
			// sum up last battery read with cpu/general polling intervall
			msecs_last_battery_read += config->current_scheme->POLL_INTERVAL;
		}
		updateCPUState();

		/* adjust interval -> substract time need to process functions */
		gettimeofday(&tv_stop, &dummy);
		/* time needed to process funcs in usecs */
		unsigned long process_time = (tv_stop.tv_sec*(1000*1000) + tv_stop.tv_usec) - 
			(tv_start.tv_sec*(1000*1000) + tv_start.tv_usec);

		// POLL_INTERVAL is set in milliseconds
		if (process_time > 1000 * config->current_scheme->POLL_INTERVAL){
			/* reading battery could take longer than 333 ms -> this message appears
			("/proc/stats has not updated" also appears), but can be ignored */
			pDebug (DBG_DIAG, "Processing functions took longer than "
				"polling interval. Polling Interval: %lu ms, Needed time processing functions %lu ms",
				config->current_scheme->POLL_INTERVAL, process_time/1000);
			req.tv_sec =  req_init.tv_sec; 
			req.tv_usec = req_init.tv_usec;
		}
		else {
			req.tv_sec  = req_init.tv_sec  - process_time / (1000*1000);
			long tmp = req_init.tv_usec - process_time % (1000*1000);
			if (tmp < 0) {                  // usec underflow
				if (req.tv_sec > 0) {   // we have a second left to compensate
					req.tv_sec--;
					req.tv_usec = tmp + (1000*1000);
				} else {                // complete underflow => reinit
					req.tv_sec =  req_init.tv_sec; 
					req.tv_usec = req_init.tv_usec;
				}
			} else {                        // no underflow
				req.tv_usec = tmp;
			}
			pDebug (DBG_DEBUG, "Needed %lu sec and %lu usec time to process functions after select\n",
			process_time / (1000*1000), process_time % (1000*1000));
		}
	}
}

void PM_Interface::checkBatteryStateChanges(){

	pDebug(DBG_DEBUG, "Check battery state\n");
	// only switch if a lower battery level is reached
	if (updateBatteryState() == 1){
		switch(cur_state.BAT_STATE){
		case BAT_NORM_STATE:
			executeEvent("battery.normal");
			pDebug(DBG_INFO, "Battery state changed to normal\n");
			break;
		case BAT_WARN_STATE:
			executeEvent("battery.warning");
			pDebug(DBG_INFO, "Battery state changed to warning\n");
			break;
		case BAT_LOW_STATE:
			executeEvent("battery.low");
			pDebug(DBG_INFO, "Battery state changed to low\n");
			break;
		case BAT_CRIT_STATE:
			executeEvent("battery.critical");
			pDebug(DBG_INFO, "Battery state changed to critical\n");
			break;
		case BAT_NONE_STATE:
			pDebug(DBG_INFO, "Battery disappeared.\n");
			break;
		default:
			pDebug(DBG_ERR, "Undefined battery state ..., that's bad\n");
		}
	}
	else{
		pDebug(DBG_INFO, "Battery state unchanged\n");
	}
}

int PM_Interface::checkACStateChanges(){
	AC_ADAPTER_STATE ac_state;
	if (updateACState()){
		pDebug(DBG_DIAG, "AC state changed");
		ac_state = cur_state.AC_STATE;

		if (ac_state == AC_OFFLINE){
			// set battery state
			executeEvent("acadapter.offline");
			pDebug (DBG_INFO, "ac adapter offline event occured\n");
			checkBatteryStateChanges();
			return 1;
		}
		else if (ac_state == AC_ONLINE){
			executeEvent("acadapter.online");
			pDebug (DBG_INFO, "ac adapter online event occured\n");
			return 1;
		}
		else if (ac_state == AC_UNKNOWN){
			// assume online for now...
			pDebug (DBG_DIAG, "AC state changed to unknown, assume online");
			ac_state = AC_ONLINE;
			return 1;
		}
		else{
			pDebug(DBG_ERR, "Could not read out AC adapter state, FATAL ERROR");
			return -1;
		}
	}
	else
		return 0;
}

void PM_Interface::checkTemperatureStateChanges(){
	int ret = 0;
	pDebug(DBG_DEBUG, "Check temperature State\n");
	ret = updateTemperatureState();
	/* it is getting hotter */
	if ( ret == 1){
		switch(cur_state.TEMP_STATE){
		case CRITICAL:
			executeEvent("temperature.critical");
			pDebug(DBG_INFO, "Temperature state changed to critical.\n");
			break;
		case HOT:
			executeEvent("temperature.hot");
			pDebug(DBG_INFO, "Temperature state changed to hot.\n");
			break;
		case PASSIVE:
			executeEvent("temperature.passive");
			pDebug(DBG_INFO, "Temperature state changed to passive.\n");
			break;
			/* I do not differ between ACTIVE (fans on) and
			   OK mode (no cooling) at the moment */
		case ACTIVE:
			executeEvent("temperature.active");
			pDebug(DBG_INFO, "Temperature state changed to active.\n");
			break;
		case OK:
			pDebug(DBG_DIAG, "Temperature state changed to ok - this should"
			       "not happen at this point ->bug ->mail@renninger.de.\n");
			break;
		default:
			pDebug(DBG_DIAG, "Undefined temperature state ->bug ->mail@renninger.de.\n");
			break;
		}
	}
	/* it is getting cooler */
	else if (ret == 2){
		switch(cur_state.TEMP_STATE){
		case CRITICAL:
			executeEvent("temperature.critical");
			pDebug(DBG_INFO, "Temperature state changed to critical. This should"
			       "not happen at this point ->bug ->mail@renninger.de.\n");
			break;
		case HOT:
			executeEvent("temperature.hot");
			pDebug(DBG_INFO, "Temperature state changed to hot.\n");
			break;
		case PASSIVE:
			executeEvent("temperature.passive");
			pDebug(DBG_INFO, "Temperature state changed to passive.\n");
			break;
			/* I do not differ between ACTIVE (fans on) and
			   OK mode (no cooling) at the moment */
		case ACTIVE:
			executeEvent("temperature.active");
			pDebug(DBG_INFO, "Temperature state changed to active.\n");
			break;
		case OK:
			executeEvent("temperature.ok");
			pDebug(DBG_INFO, "Temperature state changed to ok.\n");
			break;
		default:
			pDebug(DBG_INFO, "Undefined temperature state ->bug ->mail@renninger.de.\n");
			break;
		}		
	}
}

void get_args(int argc, char** argv, GeneralConfig *config_obj, int &daemon, int acpi_apm) {
	int option_index=0;
	int verbose = 0;
	char *endptr[MAX_PARAM_LENGTH+1];
	daemon = 0;

	struct stat *fbuf = NULL;

	struct option opts[] =
		{
			{"daemonize", 0, 0, 'd'},
			{"acpi-file", 0, 0, 'f'},
			{"scheme-dir", 0, 0, 's'},
			{"scripts-dir", 0, 0, 'x'},
			{"verbosity", 0, 0, 'v'},
			{"config-dir", 0, 0, 'c'},
			{"help", 0, 0, 'h'},
			{NULL, 0, 0, 0},
		};
	while (1){
		int i = getopt_long(argc, argv, "v:c:s:x:f:ndh", opts, &option_index);
		if (i == -1){
			break;
		}
		if (optarg && strlen(optarg) >= MAX_PARAM_LENGTH){
			fprintf(stderr, "Parameter: '%s' is too long\n", optarg);
			goto exit_failure;
		}
		switch (i){
		case 0:
			break;
		case 'n':
			config_obj->disable_CPU_freq = 1;
			break;
		case 's':
			fbuf = (struct stat*) new struct stat();
			if (!fbuf)
				goto exit_failure;
			if (stat(optarg, fbuf)){
				printf("Cannot stat directory %s", optarg);
				delete fbuf;
				goto exit_failure;
			}
			if (!S_ISDIR(fbuf->st_rdev)){
				printf("%s is not a directory", optarg);
				delete fbuf;
				goto exit_failure;
			}
			delete fbuf;
			if (optarg[strlen(optarg) - 1] != '/')
				PS_Config::scheme_dir = string(optarg) + '/';
			else
				PS_Config::scheme_dir = string(optarg);
			printf ("Scheme directory set to %s", optarg);
			break;
		case 'c':
			if (!access(optarg, R_OK)){
				config_obj->config_dir = optarg;
				if (optarg[strlen(optarg)-1] != '/'){
					config_obj->config_dir += '/';
				}
			}
			else{
				fprintf(stderr, "Cannot access directory %s ", optarg);
				goto exit_failure;
			}
			break;
		case 'x':
			if (!access(optarg, X_OK)){
				Event::scripts_dir = optarg;
				if (optarg[strlen(optarg)-1] != '/'){
					Event::scripts_dir += '/';
				}
			}
			else{
				fprintf(stderr, "Cannot access directory %s ", optarg);
				goto exit_failure;
			}
			break;
		case 'f':
                        if (acpi_apm == ACPI) {
                                if (!access(optarg, W_OK)){
                                        ACPI_Interface::ACPI_EVENTFILE = optarg;
                                }
                                else{
                                        fprintf(stderr, "Cannot access '%s'\n", optarg);
                                        goto exit_failure;
                                }
			}
			break;
		case 'v':
			verbose = strtol(optarg, endptr, 10);
			if (**endptr !='\0' || verbose < 0 || verbose > 31){
				fprintf(stderr,"Wrong verbose paramater, must be between 0-31\n");
				goto exit_failure;
			}
			else {
				setDebugLevel(verbose);
			}
			break;
		case 'd':
			daemon = 1;
			break;
		case 'h':
		default:
			usage();
			goto exit_failure;
		}
	}
	return;
 exit_failure:
	delete config_obj;
	config_obj = NULL;
	exit(EXIT_FAILURE);	
}


int main(int argc, char** argv) {
	// vars for get_args:
	int daemon;
	GeneralConfig *config_obj;
	struct sigaction signal_action;
	int acpi_apm;	
	FILE *pid_file;

	/* check perms */
	if (geteuid() != 0) {
		fprintf(stderr, "Must be run as root\n");
		exit(EXIT_FAILURE);
	}

	// check whether there is already a pid file!
	if (!access(POWERSAVED_PID_FILE_PATH, F_OK)){
		ifstream is(POWERSAVED_PID_FILE_PATH);
		pid_t pid = 0;
		if (is.good())
			is >> pid;
		if (pid) {
			// finding out if the pid exists or not
			// if it's already taken by some other process we fall
			// back to the normal behaviour: assuming it's a powersaved
			if (kill(pid, 0) == -1 && errno == ESRCH)
				pid = 0;
		}
		if (pid) {
			fprintf(stderr, "powersave daemon is already running. "
				"If it is not running, delete pid file: %s\n", POWERSAVED_PID_FILE_PATH);
			exit(EXIT_FAILURE);
		}
	}
	config_obj = (GeneralConfig*) new GeneralConfig();
	if (config_obj == NULL)
	        goto exit_failure;

        // check here, we need to know if we can ignore "-f $acpid_socket" in get_args
	acpi_apm = checkACPI();

	get_args(argc, argv, config_obj, daemon, acpi_apm);

	if (daemon)
		daemonize();

	if (config_obj->readConfFiles() <= 0){
		pDebug(DBG_ERR, "Error while parsing config file.");
		printf ("Error while parsing config file. See syslog file.");
		goto exit_failure;
	}
	//printf("event: %s, general: %s", event_conf_file.c_str(), general_conf_file.c_str());
	// create pid file here, so I can call exit in get_args(don't need to unlink pid file)
	/*** write process ID file! *************************************/
	if ((pid_file = fopen(POWERSAVED_PID_FILE_PATH, "w")) == NULL){
		fprintf(stderr, "Could not create pid file: %s, error: %s\n", POWERSAVED_PID_FILE_PATH, strerror(errno));
		goto exit_failure;
	}
	fprintf(pid_file, "%d", getpid());
	fclose(pid_file);
	/*** write process ID file! *************************************/

	/***** S E T   S I G N A L   H A N D L I N G ********************/
	// set signal handlers for SIGHUP and exit signals
	// ignore childs, so that main process can go on with listening
	// on socket and hardware event file

	memset(&signal_action, 0, sizeof(signal_action));
	sigaddset(&signal_action.sa_mask, SIGHUP);
	sigaddset(&signal_action.sa_mask, SIGCHLD);
	sigaddset(&signal_action.sa_mask, SIGTERM);
	sigaddset(&signal_action.sa_mask, SIGINT);
	sigaddset(&signal_action.sa_mask, SIGQUIT);
	sigaddset(&signal_action.sa_mask, SIGUSR1);
	sigaddset(&signal_action.sa_mask, SIGPIPE);


	// ignore all signals during signal handling of these signals
	// restart system calls if signal happen
	// this is for restarting select
	signal_action.sa_flags = SA_RESTART || SA_NOCLDSTOP;

#ifdef CPUFREQ_MEASURE
	sigaddset(&signal_action.sa_mask, SIGRTMIN+1);
	sigaddset(&signal_action.sa_mask, SIGRTMIN+2);
	signal_action.sa_handler = RT_START_handler;
	sigaction(SIGRTMIN+1, &signal_action, 0);
	signal_action.sa_handler = RT_STOP_handler;
	sigaction(SIGRTMIN+2, &signal_action, 0);
#endif
	
	//do a broken pipe from clients
	signal_action.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &signal_action, 0);

	// re-read configs
	// set the HUP handler to our function
	signal_action.sa_handler = hup_handler;
	sigaction(SIGHUP, &signal_action, 0);

	// terminate properly ...
	// set the TERM handler to our function
	// ToDo: could the signal_action struct
	// be modified here? -> think no
	signal_action.sa_handler = exit_handler;
	sigaction(SIGTERM, &signal_action, 0);
	// set the QUIT handler to our function
	sigaction(SIGQUIT, &signal_action, 0);
	// set the INT handler to our function
	sigaction(SIGINT, &signal_action, 0);

	/* used to inform main process that resume of sleep state is finished */
	signal_action.sa_handler = USR1_handler;
	sigaction(SIGUSR1, &signal_action, 0);

	signal_action.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &signal_action, 0);
	/****************************************************************/

	if (acpi_apm == ACPI){
		pm = new ACPI_Interface(config_obj);
	}
	else if (acpi_apm == APM) {
		config_obj->FORCE_BATTERY_POLLING = 1;
		pm = new APM_Interface(config_obj);
	}
	else {
		fprintf(stderr, "Neither ACPI, nor APM is supported\n");
		goto exit_failure;
	}
	pm->start();

 exit_failure:
	delete pm;
	unlink (POWERSAVED_PID_FILE_PATH);
	exit(EXIT_FAILURE);
}

int PM_Interface::executeEvent(const string eventName, const string acpi_event_line){
	/* this is only for acpi events...*/
	Event execEvent = config->getEvent(eventName);
	int eventId;

	if ( execEvent.name == "global.suspend2disk" ||
	     execEvent.name == "global.suspend2ram"  ||
	     execEvent.name == "global.standby" ) {

		if ( eM->eventIsRunning( execEvent.name ) ) {
			pDebug(DBG_WARN,"Event %s already running. Ignoring this one", execEvent.name.c_str());
			return 1;
		}
	}

	if (execEvent.name == "default"){
		pDebug(DBG_ERR, "Tried to execute non-existing event.");
		return -1;
	}

	DBus_Server::emitSignal( "PowersaveEvent", eventName.c_str() );


	// append the event to the list of pending events. Timeout is set to 120 seconds.
	// are 2 minutes enough?
	eventId = eM->registerEvent(execEvent, 120);

	// do not ignore child if daemon terminates
	// this is the only event where main process has
	// to wait until proxy is finished
	if (execEvent.name == "daemon.terminate")
		signal(SIGCHLD, SIG_DFL);
	
	// execute the first script/internal action of the event
	execEvent.execute_program(this, acpi_event_line);

	return 1;
}


int PM_Interface::updateBatteryState(){
	int changed = 0;

	int cur_bat_state = cur_state.BAT_STATE;;
	int cur_bat_perc = bg.remPerc;
	int cur_bat_mins = bg.remMins;
	int cur_charging_state = bg.charging_state;
	int cur_charging_mins = bg.rem_charging_mins;

	BATTERY_STATES new_bat_state;

	msecs_last_battery_read = 0;

	if (getBatteriesInfo(&bg) < 0){
		pDebug(DBG_INFO, "Could not get battery info\n");
		// cannot read out battery, no cpufreq, stop polling
		// suppors AC online and set battery state to normal
		cur_state.BAT_STATE=BAT_NONE_STATE;
		return -1;
	}

	if (cur_bat_perc != bg.remPerc || cur_bat_mins != bg.remMins ||
		cur_charging_state != bg.charging_state || cur_charging_mins != bg.rem_charging_mins)
		executeEvent("battery.info");

	if ((cur_state.AC_STATE == AC_ONLINE) || ((config->current_scheme->BAT_WARN_LIMIT < bg.remPerc) && (bg.remPerc <= 100)))
		new_bat_state = BAT_NORM_STATE;
	else if (config->current_scheme->BAT_LOW_LIMIT < bg.remPerc && bg.remPerc <= config->current_scheme->BAT_WARN_LIMIT)
		new_bat_state = BAT_WARN_STATE;
	else if (config->current_scheme->BAT_CRIT_LIMIT < bg.remPerc && bg.remPerc <= config->current_scheme->BAT_LOW_LIMIT)
		new_bat_state = BAT_LOW_STATE;
	else if (0 < bg.remPerc && bg.remPerc <= config->current_scheme->BAT_CRIT_LIMIT)
		new_bat_state = BAT_CRIT_STATE;
	else if (bg.remPerc <= 0){
		/* if we cannot calc remaining percent at least set 
		   warning/low/crit battery state at 8/4/2 minutes of remaining battery */
		if (bg.remMins <= 8 && bg.remMins > 0)
			new_bat_state = BAT_WARN_STATE;
		else if (bg.remMins <= 4 && bg.remMins > 0)
			new_bat_state = BAT_LOW_STATE;
		else if (bg.remMins <= 2 && bg.remMins > 0)
			new_bat_state = BAT_CRIT_STATE;
		else
			new_bat_state = BAT_NORM_STATE;
	}
	else{
		pDebug(DBG_INFO, "Could not determine battery state, suppose normal\n");
		new_bat_state = BAT_NORM_STATE;
	}

	pDebug(DBG_INFO, "Get Battery state: %d\n", new_bat_state);


	if (cur_bat_state == new_bat_state)
		changed = 0;
	else if (new_bat_state > cur_bat_state){
		changed = 2;
		cur_state.BAT_STATE = new_bat_state;
	}
	else if (new_bat_state < cur_bat_state){
		changed = 1;
		cur_state.BAT_STATE = new_bat_state;
	}
	return changed;
}

int PM_Interface::updateCPUState(){
	int cpu_load;
	// get CPU load, just cast it, accuracy not important
	cpu_load = (int)getCPULoadConsiderNice(config->current_scheme->CONSIDER_NICE);
	if (cpu_load == -1){
		pDebug(DBG_ERR, "Could not evalute cpu load(%d)\n", cpu_load);
		return -1;
	}
	else if (cpu_load == -2){
		pDebug(DBG_DIAG, "/proc/stat not updated yet, CPU load could not be evaluated\n");
		return -1;
	}
	/* adjust CPU frequency directly here now **********/
	/* ToDo: Maybe the main loop (select on fds), specially the timeout needs
		 to be reimplemented to avoid waste of CPU performance in case
		 we have kernel controlled cpufreq
	*/
	if (config->CPUFREQ_CONTROL == CPUFREQ_USERSPACE)
		CPUFreq_Interface::adjustSpeeds();
	/*******************************************/
	/* this checks whether CPU should get throttled ... *******/
	/* CPUs get throttled if load of CPU stays below CPU_IDLE_LIMIT
	   longer than CPU_IDLE_TIMEOUT */
	if(config->current_scheme && config->current_scheme->ALLOW_THROTTLING && config->current_scheme->CPU_IDLE_TIMEOUT > 0){
		if (cpu_load < config->current_scheme->CPU_IDLE_LIMIT)
			cur_state.CPU_IS_IDLE = 1;
		else
			cur_state.CPU_IS_IDLE = 0;
		if ((processor_event_thrown == NO_PROCESSOR_EVENT_THROWN || processor_event_thrown == PROCESSOR_IDLE_THROWN)
		    && !cur_state.CPU_IS_IDLE){
			// for now only one CPU is supported 
			// dethrottle CPU immediately if CPU is above CPU_BELOW_DYNAMIC usage limit
			// and a low or no processor event has been thrown
			executeEvent("processor.busy");
			msecs_cpu_idle = 0;
			processor_event_thrown = PROCESSOR_BUSY_THROWN;
		}
		else if (cur_state.CPU_IS_IDLE && 
			 (processor_event_thrown == NO_PROCESSOR_EVENT_THROWN || processor_event_thrown == PROCESSOR_BUSY_THROWN)) {
			 // count up idle time and send a
			 // throttle CPU when time limit ( config->current_scheme->CPU_IDLE_TIMEOUT variable)
			 // is reached
			 pDebug(DBG_DEBUG, "cpu_idle_time: %lu; timeout: %lu; low event already thrown : %d"
				", poll_Intervall: %lu; CPU_IS_IDLE: %d\n" ,
				msecs_cpu_idle, config->current_scheme->CPU_IDLE_TIMEOUT * 1000, processor_event_thrown, 
				config->current_scheme->POLL_INTERVAL, cur_state.CPU_IS_IDLE);
			 msecs_cpu_idle += config->current_scheme->POLL_INTERVAL;
			 if (msecs_cpu_idle > config->current_scheme->CPU_IDLE_TIMEOUT * 1000){
				 executeEvent("processor.idle");
				 msecs_cpu_idle = 0;
				 processor_event_thrown = PROCESSOR_IDLE_THROWN;
			 }
		}			 
	}
	/*******************************************************/
	return 1;
}

int PM_Interface::updateACState(){
	int changed;
	int old_ac_state;

	old_ac_state = cur_state.AC_STATE;
	cur_state.AC_STATE = (AC_ADAPTER_STATE)getACAdapterStatus();
	/* there are new error codes returned 
	   (e.g. NO_DEVICE_ERROR and NO_MODULE_ERROR) */
	if (cur_state.AC_STATE != AC_ONLINE && cur_state.AC_STATE != AC_OFFLINE){
		pDebug(DBG_INFO, "Could not detect AC status, assume AC_ONLINE");
		cur_state.AC_STATE = AC_ONLINE;
	}
	if (old_ac_state != cur_state.AC_STATE){
		changed = 1;
		if (activateScheme(cur_state.AC_STATE) > 0){
			executeEvent("daemon.scheme.change");
		}
	}
	else
		changed = 0;
	return changed;
}



int PM_Interface::updateTemperatureState(){

	int x;
	int state;
	/* is also THERM_MODE but C (library) forbids the declaration */
	int temp_state = OK;

	for (x = 0; config->current_scheme->thermal_zones[x].present && (state = getThermalZoneState(x)) > 0 && 
		     x < MAX_THERMAL_ZONES; x++){
		config->current_scheme->thermal_zones[x].state = state;
		temp_state = max(state, temp_state);
	}
	/* here we need more proper error handling -> destroy allocated objects
	   probably the thermal module has been unloaded ... */
	if (config->current_scheme->thermal_zones[x].present){
		pDebug (DBG_DIAG, "Error during temperature evaluation, has the thermal "
			"module been unloaded? Then restart the daemon.\n");
		return -1;
	}
	if (temp_state > cur_state.TEMP_STATE){
		cur_state.TEMP_STATE = (THERM_MODE)temp_state;
		return 1;
	}
	else if (temp_state < cur_state.TEMP_STATE){
		cur_state.TEMP_STATE = (THERM_MODE)temp_state;
		return 2;
	}
	else
		return 0;
}

int PM_Interface::activateScheme(string scheme_name){

	int ret;
	if (!config->current_scheme->SCHEME_NAME.compare(scheme_name)){
		pDebug(DBG_DIAG, "Scheme already set");
		return 0;
	}
	ret = config->setActiveScheme(scheme_name);
	if (ret <= 0)
		return ret;

	pDebug(DBG_DIAG, "Successfully switched scheme to: %s",
	       scheme_name.c_str());
	activateSettings();
	return 1;
}

int PM_Interface::activateScheme(AC_ADAPTER_STATE ac_state){

	int ret = 0;
	switch (ac_state){
	case AC_ONLINE:
		ret = activateScheme(config->getACScheme());
		break;
	case AC_OFFLINE:
		ret = activateScheme(config->getBatteryScheme());
		break;
	case AC_UNKNOWN:
		// which one should we use -> I'd say better use AC
		// we could even test for battery level, or is this
		// too overdosed?
		ret = activateScheme(config->getACScheme());
		break;
	}
	return ret;
}

void PM_Interface::activateSettings(){
	
	/***** experimental: Switch kernel/userspace control during runtime
	       We should be able to just override old objects as we do not use new
	       any more **********************************************************/
	if (config->current_scheme->CPUFREQ_CONTROL != CPUFreq_Interface::getControlMode()){
		pDebug(DBG_DIAG, "Change cpufreq control to: %s\n",
		       ((config->current_scheme->CPUFREQ_CONTROL == CPUFREQ_USERSPACE) ?
			"userspace" : "kernel"));
		CPUFreq_Interface::createCPUFreqObject(config->current_scheme->CPUFREQ_CONTROL, 
						       config->current_scheme->POLL_INTERVAL * 1000);
	}
	/*************************************************************************/
	
	/* set new CPU freq mode */
	if (!CPUFreq_Interface::setModes(config->current_scheme->CPUFREQUENCY) < 0){
		pDebug(DBG_WARN, "Could not set CPUFreq mode\n");
	}

	switch (config->current_scheme->DISPLAY_BRIGHTNESS) {
	case -1:   // ignore
		break;
	case -2:   // low
		brightness->Min();
		break;
	case -3:   // medium
		brightness->Med();
		break;
	case -4:   // high
		brightness->Max();
		break;
	default:   // level
		brightness->Set(config->current_scheme->DISPLAY_BRIGHTNESS);
		break;
	}
}


int PM_Interface::throttle(int percent){

	int ret = -1;
	if (throttling_supported && config->current_scheme->ALLOW_THROTTLING && cur_state.TEMP_STATE < PASSIVE){
		if (percent == -1)
			percent = config->current_scheme->MAX_CPU_THROTTLING;
		if (setThrottlingPercent(percent) <= 0)
			throttling_supported = 0;
		else{
			ret = 1;
			pDebug(DBG_INFO, "CPU(s) throttled to %d percent", percent);
		}
	}
	return ret;
}


int PM_Interface::dethrottle(){

	int ret = -1;
	if (throttling_supported && config->current_scheme->ALLOW_THROTTLING && cur_state.TEMP_STATE < PASSIVE){
		if (setThrottlingPercent(0) <= 0)
			throttling_supported = 0;
		else{
			pDebug(DBG_INFO, "CPU(s) dethrottled");
			ret = 1;
		}
	}
	return ret;
}

void PM_Interface::resume(){
	 pDebug (DBG_DIAG, "Daemon reinitialises system states, resume finished");
	 requested_sleep_state = PSD_NO_SLEEP_REQUEST;
	 // reset frequency which may have changed without noticing it
	 CPUFreq_Interface::reinitSpeeds();
	 checkACStateChanges();
	 checkBatteryStateChanges();
}


void PM_Interface::checkEventTimeouts() {
	// go through all events and check if one of them timed out
	std::list<Event>::iterator it;
	for ( it = eM->Events.begin(); it != eM->Events.end(); it++) {
		
		if (it->timeElapsed() > it->getTimeout()) {
			pDebug(DBG_WARN, "Timeout for event %s with id %d reached, "
			       "removing it.\n",
			       it->name.c_str(), it->ID());
			eM->Events.erase(it);
			/* deleting the object may result in invokation
			   of a non existing event (it->timeElapsed() above)
			   -> start reiterating from the beginning of the list
			*/
			it = eM->Events.begin();
		}
	}
}

void PM_Interface::checkScriptReturn(int id, int returnValue, string comment) {
			
	// get the program name
	string prog = eM->cur_prog(id);
	
	// if a program matching an event was found
	if (! prog.empty()) {
		pDebug(DBG_INFO, "SCRIPT returned:, "
		       "Name: %s, "
		       "Return value: %d, "
		       "Comment: %s ", prog.c_str(), returnValue, comment.c_str());
	}
	else {
		pDebug(DBG_DIAG, "Unknown SCRIPT returned with return value %d, text: %s", returnValue, comment.c_str());
		// allow popup windows also from unknown processes?
		// if (returnValue !=2)
		return;
	}

	// enable the buttons when we come back from restore
	if (prog.find("restore_after_s") == 0){
		ignore_buttons(false);
	}

	// evaluate the return code
	switch (returnValue) {
	// script succeeded
	case 0: {
		// set the event to the next script
		int i = eM->moveOn(id);
		
		// no scripts left, so finished successfully
		if ( !i ) {
			const string name = eM->name(id);
			pDebug(DBG_INFO, "SCRIPT Event %s finished "
			       " successfully", name.c_str());
			// notify clients here
			eM->remove(id);
		} else {
			eM->event(id)->execute_program(this,"");
		}
		break;
	}
	// script requests a graphical notification from the daemon
	case 2:
		x_notification(comment);
		break;
	// script requests a xlock/screensaver from the daemon
	case 3:
		screenlock(comment);
		break;
	case 4:
		DBus_Server::emitSignal( "Progress", comment.c_str() );
		break;
	// script has failed
	case 1:
		const string event_name = eM->name(id);
		pDebug(DBG_INFO, "SCRIPT %s failed to execute", event_name.c_str());

		// if the suspend/standby event has failed --> enable buttons
		if  ((event_name.find("global.suspend") == 0)
		  || (event_name.find("global.standby") == 0)) {
			ignore_buttons(false);
		}

		// notify clients here
		// x_notification(comment);
		// remove the event from the list of pending events
		eM->remove(id);
		break;
	}
}


void PM_Interface::sleep_wrapper(string event) {
	static bool before_sleep = true;
	static struct timeval cur_time;
	static struct timezone dummy;

	if ( before_sleep ) {
		if ( event == "suspend2disk" ) {
			if (config->current_scheme->SUSPEND2DISK_DELAY > 0)
				sleep(config->current_scheme->SUSPEND2DISK_DELAY);
		} else if ( event == "suspend2ram" ){
			if (config->current_scheme->SUSPEND2RAM_DELAY > 0)
				sleep(config->current_scheme->SUSPEND2RAM_DELAY);
		} else if ( event == "standby" ){
			if (config->current_scheme->STANDBY_DELAY > 0)
				sleep(config->current_scheme->STANDBY_DELAY);
		}
		ignore_buttons(true);

		// we used to sync here, but we already sync in userspace and the
		// kernel syncs again if necessary => this only slows things down
// 		sync();
			
		pDebug(DBG_INFO, "Set machine into suspend2disk/ suspend2ram mode\n");
		
		before_sleep = false;
		gettimeofday(&cur_time, &dummy);
	} else {
		// the time where we went to sleep in milliseconds
		unsigned long sleeptime = cur_time.tv_sec * 1000 + cur_time.tv_usec / 1000;

		gettimeofday( &cur_time, &dummy );

		// the time were we are back in milliseconds
		unsigned long wakeuptime = cur_time.tv_sec * 1000 + cur_time.tv_usec / 1000;

		// substract the time we slept from all events
		eM->adjustTimeouts( wakeuptime - sleeptime );
		
		pDebug(DBG_INFO, "Back from %s inside sleep_wrapper\n", event.c_str() );

		resume();
		before_sleep = true;
		// throw a resume event
                executeEvent("global.resume." + event);
	}
}


int PM_Interface::executeScript(string script, string param){
	int ret = -1;
	int pid;
	int status;
	char *argv[3];
	argv [0] = new char[script.size() + 1];
	argv [1] = new char[param.size() + 1];
	strcpy (argv [0], script.c_str());
	strcpy (argv [1], param.c_str());
	argv [2] = NULL;

	string full_path_to_script = this->INT_SCRIPT_PATH + script;

	pid = fork();
	if (!pid){
		signal(SIGHUP, SIG_DFL);
		signal(SIGTERM, SIG_DFL);
		signal(SIGINT, SIG_DFL);
		signal(SIGQUIT, SIG_DFL);
		signal(SIGPIPE, SIG_DFL);
		signal(SIGCHLD, SIG_DFL);
		// close all fds
		int fdlimit = sysconf(_SC_OPEN_MAX);
		
		for (int fd = 0; fd < fdlimit; fd++)
			close(fd);


		execv(full_path_to_script.c_str(), argv);
		// we should never get here.
		pDebug (DBG_ERR, "could not execv the sleep script '%s'. Argument: '%s', error: '%s'", full_path_to_script.c_str(), argv[1], strerror(errno));
		_exit(127);
	}
	else if (pid == -1){
		pDebug(DBG_WARN, "Could not fork for sleep script");
	}
	else{
		// if the script is the sleep script, wait for its return
		if ( script == this->SLEEP_SCRIPT ){
			if (waitpid(pid, &status, 0) < 0){
				if (errno == ECHILD){
					pDebug (DBG_DIAG, "Sleep script already exited.");
				}
				else{
					pDebug (DBG_ERR, "Sleep script failed - %s", strerror(errno));
				}
			}
			else if (WIFEXITED(status)) {
				if (WEXITSTATUS(status)){
					pDebug(DBG_WARN, "Program to trigger sleep (%s) failed and exited with status %d\n",
					       full_path_to_script.c_str(), WEXITSTATUS(status));
					ret = -1;
				}
				else{
					/* everything is alright, go on executing ... */
					pDebug(DBG_INFO, "Program %s to trigger sleep executed successfully\n",
					       full_path_to_script.c_str());
					ret = 0;
				}
			}
			else if (WIFSIGNALED(status)) {
				pDebug(DBG_DIAG, "Program %s exited on signal %d\n",
				       full_path_to_script.c_str(), WTERMSIG(status));
				ret = -1;
			}
			
		}
		// no need to wait for script to return
	}
	return ret;
}

int PM_Interface::screenlock(string param) {

	DBus_Server::emitSignal( "Screenlock", param.c_str() );

	if ( !cC->checkClientCapabilities( CAPABILITY_SCREENLOCK ) ) {
		executeScript("do_screen_saver", param);
	}
	return 1;
}

int PM_Interface::x_notification(string param) {
	DBus_Server::emitSignal( "Notification", param.c_str() );

	if ( !cC->checkClientCapabilities( CAPABILITY_NOTIFICATIONS ) ){
		executeScript("do_x_notification", param);
	}
	return 1;
}

void PM_Interface::ignore_buttons(bool ignore){
	pDebug(DBG_DIAG,"ignore_buttons is set to: %d" ,ignore );
	sleep_triggered = ignore;
}


DBUS_ERROR PM_Interface::handleDBusRequest(DBusMessage *msg,
					   DBusMessage *reply,
					   POWERSAVE_MESSAGE_TYPE type){
	
	
	string msg_member = dbus_message_get_member(msg);
	string sender = dbus_message_get_sender(msg);
	uint32_t serial = dbus_message_get_serial (msg);
	
	/** @ToDo: Debug only, throw out when working... */
	pDebug (DBG_INFO, "Received msg: '%s'", msg_member.c_str());
	pDebug (DBG_DEBUG, "Obj.path '%s' serial '%d' from '%s'",
		dbus_message_get_path(msg), serial, sender.c_str());
	
	switch (type){
	case MANAGER_MESSAGE:
		return handleManagerMessage( msg );
		break;
	case REQUEST_MESSAGE:
		return handleRequestMessage( msg,
					     reply );
		break;
	case ACTION_MESSAGE:	       
		return handleActionMessage( msg );
		break;
	default:
		/* Should never happen -> get rid of message? */
		pDebug (DBG_ERR, "Received invalid message");
		return REPLY_INVALID_INTERFACE;
	}

	return REPLY_GENERAL_ERROR;		
}


DBUS_ERROR PM_Interface::handleManagerMessage( DBusMessage *msg ) {

	string msg_member = dbus_message_get_member(msg);
	string sender = dbus_message_get_sender(msg);

	char *dummy = "";

	pDebug (DBG_DIAG, "Handling manager request");
	if (msg_member == "Connect"){
		int capabilities;
		/* add client. No 'request string/modes' check so far */
		if (!dbusGetMessageInteger(msg, &capabilities, 0)){
			cC->addClient( sender.c_str(), capabilities );
			return REPLY_SUCCESS;
		}
		else
			return REPLY_INVALID_PARAM;
	}
	else if (msg_member == "Disconnect"){
		cC->checkClientRemoval( sender.c_str() );
		return REPLY_SUCCESS;
	}
	else if (msg_member == "EventReturn"){
		int event_id, request;

		// get the event id
		if ( dbusGetMessageInteger( msg, &event_id, 0 ) < 0 ) {
			pDebug(DBG_WARN, "Script returned but we were unable to get the event id");
			return REPLY_INVALID_PARAM;
		}

		// get the so called mode ( notify, progress, true, false )
		if ( !dbusGetMessageInteger( msg, &request, 1 ) < 0 ) {
			pDebug(DBG_WARN, "Script returned but we were unable to get the request");
			return REPLY_INVALID_PARAM;
		}

		// get the message
		if ( dbusGetMessageString( msg, &dummy, 0 ) < 0 ){
			pDebug(DBG_INFO,"Script returned but we were unable to get the message."
			       "Continueing anyway...");
			dummy = "";
		}

		pDebug(DBG_DEBUG, "Script returned: event id: %d, request: %d, message: %s",
		       event_id, request, dummy);

		checkScriptReturn( event_id, request, dummy);
		
		return REPLY_SUCCESS;
	}
	else
		return REPLY_INVALID_METHOD;

	return REPLY_GENERAL_ERROR;
}


DBUS_ERROR PM_Interface::handleRequestMessage( DBusMessage *msg,
					       DBusMessage *reply ) {

	string msg_member = dbus_message_get_member(msg);
	char *dummy = "";

	/** @ToDo Also reply REPLY_HW_NOT_SUPPORTED for suspend*/
	if (msg_member == "AllowedSuspendToDisk" || msg_member == "suspend_to_disk_allowed"){
		if (config->current_scheme->USER_SUSPEND2DISK_DISABLED)
			return REPLY_DISABLED;
		else
			return REPLY_SUCCESS;
	}
	else if (msg_member == "AllowedSuspendToRam" || msg_member == "suspend_to_ram_allowed"){
		if (config->current_scheme->USER_SUSPEND2RAM_DISABLED)
			return REPLY_DISABLED;
		else
			return REPLY_SUCCESS;
	}
	else if (msg_member == "AllowedStandby" || msg_member == "standby_allowed"){
		if (config->current_scheme->USER_STANDBY_DISABLED)
			return REPLY_DISABLED;
		else
			return REPLY_SUCCESS;
	}
	else if (msg_member == "BatteryChargingState"){
		switch (bg.charging_state){
		case CHARG_STATE_UNKNOWN:
			return REPLY_HW_NOT_SUPPORTED;
		case CHARG_STATE_CHARGING:
			dummy = "charging";
			break;
		case CHARG_STATE_DISCHARGING:
			dummy = "discharging";
			break;
		case (CHARG_STATE_CHARGING | CHARG_STATE_DISCHARGING):
			dummy = "charging discharging";
			break;
		default:
			return REPLY_INVALID_METHOD;
		}
		dbus_message_append_args(reply, 
					 DBUS_TYPE_STRING, &dummy, 
					 DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	}
	else if (msg_member == "BatteryRemainingTime"){
		if (bg.remMins < 0)
			return REPLY_HW_NOT_SUPPORTED;
		else{
			dbus_message_append_args(reply,
						 DBUS_TYPE_INT32, &bg.remMins,
						 DBUS_TYPE_INVALID);
			return REPLY_SUCCESS;
		}
	}
	else if (msg_member == "BatteryRemainingPercent"){
		if (bg.remPerc < 0)
			return REPLY_HW_NOT_SUPPORTED;
		else{
			dbus_message_append_args(reply,
						 DBUS_TYPE_INT32, &bg.remPerc,
						 DBUS_TYPE_INVALID);
			return REPLY_SUCCESS;
		}
	}
	else if (msg_member == "BatteryRemainingChargingTime"){
		if (bg.rem_charging_mins < 0)
			return REPLY_HW_NOT_SUPPORTED;
		else{
			dbus_message_append_args(reply,
						 DBUS_TYPE_INT32, &bg.rem_charging_mins,
						 DBUS_TYPE_INVALID);
			return REPLY_SUCCESS;
		}
	}
	else if (msg_member == "BatteryState"){
		switch(cur_state.BAT_STATE){
		case BAT_NONE_STATE:
			return REPLY_HW_NOT_SUPPORTED;
			break;
		case BAT_NORM_STATE:
			dummy = "normal";
			break;
		case BAT_WARN_STATE:
			dummy = "warning";
			break;
		case BAT_LOW_STATE:
			dummy = "low";
			break;
		case BAT_CRIT_STATE:
			dummy = "critical";
			break;
		default:
			return REPLY_GENERAL_ERROR;
			break;
		}
		dbus_message_append_args(reply,
					 DBUS_TYPE_STRING, &dummy,
					 DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	}
	else if (msg_member == "CpufreqPolicy"){
		const char *policy;
		if (!CPUFreq_Interface::isSupported())
			return REPLY_HW_NOT_SUPPORTED;

		switch (CPUFreq_Interface::getMode()){
		case _PERFORMANCE:
			policy = "performance";
			break;
		case _POWERSAVE:
			policy = "powersave";
			break;
		case _DYNAMIC:
			policy = "dynamic";
			break;
		default:
			return REPLY_HW_NOT_SUPPORTED;
		}
		dbus_message_append_args(reply,
					 DBUS_TYPE_STRING, &policy,
					 DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	}
	else if (msg_member == "AcPower"){
		switch(cur_state.AC_STATE){
		case AC_ONLINE:
			dummy = "on";
			break;
		case AC_OFFLINE:
			dummy = "off";
			break;
		case AC_UNKNOWN:
			dummy = "unknown";
			break;
		default:
			return REPLY_HW_NOT_SUPPORTED;
		}
		dbus_message_append_args(reply,
					 DBUS_TYPE_STRING, &dummy,
					 DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	}
	else if (msg_member == "SchemesGet"){
		int active   = -1;
		int battery  = -1;
		int ac_power = -1;
		SchemeConfig *scheme;
		const char* scheme_name;

		for (int x = 0; x < MAX_SCHEMES; x++){
			scheme = config->getSchemeByNumber(x);
			if (scheme == NULL)
				break;
			/* only set \n if there will be another string to append */
			scheme_name = scheme->SCHEME_NAME.c_str();
			dbus_message_append_args(reply,
						 DBUS_TYPE_STRING, &scheme_name,
						 DBUS_TYPE_INVALID);
			if (scheme == config->current_scheme)
				active = x;
			if (scheme == config->Battery_scheme)
				battery = x;
			if (scheme == config->AC_scheme)
				ac_power = x;
		}
		/*
		  pDebug (DBG_DIAG, "Active: %d, battery: %d, ac_power: %d",
		  battery, active, ac_power);
		*/

		/* order of appended INT32 for request "schemes":
		 * 0: no. of currently active   scheme
		 * 1: no. of           battery  scheme
		 * 2: no. of           ac_power scheme
		 * Use the values of these appended INT32 to determine which
		 * appended strings are active/battery/ac_power scheme
		 */
		if (active == -1 || battery == -1 || ac_power == -1){
			pDebug (DBG_ERR, "Could not find active/battery/ac_power scheme %d/%d/%d",
				active, battery, ac_power);
			return REPLY_GENERAL_ERROR;
		}
		else{
			dbus_message_append_args(reply,
						 DBUS_TYPE_INT32, &active,
						 DBUS_TYPE_INT32, &battery,
						 DBUS_TYPE_INT32, &ac_power,
						 DBUS_TYPE_INVALID);
		}

		return REPLY_SUCCESS;
	}
	else if (msg_member == "BrightnessGet"){
		int brt = brightness->Get();
			
		if ( brt < 0 ) {
			return REPLY_HW_NOT_SUPPORTED;
		}

		printf("brt: %d\n", brt);
		dbus_message_append_args(reply,
					 DBUS_TYPE_INT32,
					 &brt,
					 DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
			
	}
	else if (msg_member == "BrightnessLevelsGet"){
		int brt = brightness->GetLevels();
			
		if ( brt < 0 ) {
			return REPLY_HW_NOT_SUPPORTED;
		}

		printf("brt: %d\n", brt);
		dbus_message_append_args(reply,
					 DBUS_TYPE_INT32,
					 &brt,
					 DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
			
	}
	else if (msg_member == "supp_sleep_states"){

	}
	else
		return REPLY_INVALID_METHOD;

	return REPLY_GENERAL_ERROR;
}


DBUS_ERROR PM_Interface::handleActionMessage( DBusMessage *msg ) {

	string msg_member = dbus_message_get_member(msg);
	char *dummy = "";
	
	if (msg_member == "SuspendToDisk" || msg_member == "suspend_to_disk"){
		if (config->current_scheme->USER_SUSPEND2DISK_DISABLED){
			pDebug(DBG_DIAG, "Suspend to disk is disabled in config.\n");
			return REPLY_DISABLED;
		}
		else{
			pDebug(DBG_INFO, "Client has requested suspend2disk\n");
			requested_sleep_state = PSD_SUSPEND2DISK;
			/* used for workaround for apm to remembemer triggered sleep state (not needed at
			   this point any more)*****/
				
			// executing event now triggers suspend2disk, using do_suspend_to_disk variable
			executeEvent("global.suspend2disk");
			return REPLY_SUCCESS;
		}
	}
	else if(msg_member == "SuspendToRam" || msg_member == "suspend_to_ram"){
		if (config->current_scheme->USER_SUSPEND2RAM_DISABLED){
			pDebug(DBG_DIAG, "Suspend2ram is disabled in config.\n");
			return REPLY_DISABLED;
		}
		else if (!(supported_sleeping_states & APM_SUSPEND) &&
			 !(supported_sleeping_states & ACPI_S3)) {
			pDebug(DBG_DIAG, "Suspend2ram is not supported");
			return REPLY_HW_NOT_SUPPORTED;
		}
		else{
			pDebug(DBG_INFO, "Client has requested suspend2ram\n");
			requested_sleep_state = PSD_SUSPEND2RAM;
			// executing event now triggers suspend2ram, using do_suspend2ram variable!
			executeEvent("global.suspend2ram");
			//suspend2ram();
			return REPLY_SUCCESS;
		}
	}
	else if(msg_member == "Standby" || msg_member == "standby"){
		if (config->current_scheme->USER_STANDBY_DISABLED){
			pDebug(DBG_DIAG, "Standby is disabled in config.\n");
			return REPLY_DISABLED;
		}
		else if (!(supported_sleeping_states & APM_STANDBY) &&
			 !(supported_sleeping_states & ACPI_S1)) {
			pDebug(DBG_DIAG, "Standby is not supported");
			return REPLY_HW_NOT_SUPPORTED;
		}
		else{
			pDebug(DBG_INFO, "Client has requested standby\n");
			/* workaround for apm to remembemer triggered sleep state ***************
			   but defined for whole PM_Interface to avoid dynamic cast */
			requested_sleep_state = PSD_STANDBY;
			/* workaround for apm to remembemer triggered sleep state ***************/
			// executing event now triggers suspend2ram, using do_suspend_to_ram variable!
			executeEvent("global.standby");
			//suspend2ram();
			return REPLY_SUCCESS;
		}
	}
	else if(msg_member == "CpufreqPerformance"){
		if (CPUFreq_Interface::isSupported()){
			if (CPUFreq_Interface::getMode() != _PERFORMANCE){
				if (CPUFreq_Interface::setModes(_PERFORMANCE) < 0){
					pDebug(DBG_ERR, "Could not set performance mode\n");
					return REPLY_HW_NOT_SUPPORTED;
				}
				// do not throw dynamic processor low/high events
				// when on max performance!
				processor_event_thrown = PROCESSOR_EVENTS_DISABLED;
				msecs_cpu_idle = 0;
				executeEvent("processor.performance");

				return REPLY_SUCCESS;
			}
			else
				return REPLY_ALREADY_SET;
		}
		else
			return REPLY_HW_NOT_SUPPORTED;
	}
	else if(msg_member == "CpufreqPowersave"){
		if (CPUFreq_Interface::isSupported()){
			if (CPUFreq_Interface::getMode() != _POWERSAVE){
				if (CPUFreq_Interface::setModes(_POWERSAVE) < 0){
					pDebug(DBG_ERR, "Could not set powersave mode\n");
					return REPLY_HW_NOT_SUPPORTED;
				}
				// this only concerns dynamic events
				processor_event_thrown = NO_PROCESSOR_EVENT_THROWN;
				msecs_cpu_idle = 0;
				executeEvent("processor.powersave");

				return REPLY_SUCCESS;
			}
			else
				return REPLY_ALREADY_SET;
		}
		else
			return REPLY_HW_NOT_SUPPORTED;
	}
	else if(msg_member == "CpufreqDynamic"){
		if (CPUFreq_Interface::isSupported()){
			if (CPUFreq_Interface::getMode() != _DYNAMIC){
				if (CPUFreq_Interface::setModes(_DYNAMIC) < 0){
					pDebug(DBG_ERR, "Could not set dynamic mode\n");
					return REPLY_HW_NOT_SUPPORTED;
				}
				// this only concerns dynamic events
				processor_event_thrown = NO_PROCESSOR_EVENT_THROWN;
				msecs_cpu_idle = 0;
				executeEvent("processor.dynamic");

				return REPLY_SUCCESS;
			}
			else
				return REPLY_ALREADY_SET;
		}
		else
			return REPLY_HW_NOT_SUPPORTED;
	}
	else if(msg_member == "Ping"){
		return REPLY_SUCCESS;
	}
	else if(msg_member == "SchemesSet"){
		if (!dbusGetMessageString(msg, &dummy, 0)){
			int r = activateScheme(dummy);
			if (r == 0){
				return REPLY_ALREADY_SET;
			}
			else if (r < 0){
				pDebug (DBG_DIAG, "Scheme '%s' does not exist", dummy);
				return REPLY_INVALID_PARAM;
			}
			else{
				pDebug(DBG_DIAG, "Scheme '%s' successfully"
				       " received and activated.\n", dummy);
					
				executeEvent("daemon.scheme.change");
				return REPLY_SUCCESS;
			}
		}
		else
			return REPLY_INVALID_PARAM;
	}
	else if(msg_member == "BrightnessSet"){
		int value;
		if ( !dbusGetMessageInteger( msg, &value, 0 ) ){
			brightness->Set( value );
			return REPLY_SUCCESS;
		}
		else {
			return REPLY_GENERAL_ERROR;				
		}
	}
	else if(msg_member == "BrightnessMin"){
		brightness->Min();
		return REPLY_SUCCESS;
	}
	else if(msg_member == "BrightnessMed"){
		brightness->Med();
		return REPLY_SUCCESS;
	}
	else if(msg_member == "BrightnessMax"){
		brightness->Max();
		return REPLY_SUCCESS;
	}
	else if(msg_member == "BrightnessUp"){
		brightness->Up();
		return REPLY_SUCCESS;
	}
	else if(msg_member == "BrightnessDown"){
		brightness->Down();
		return REPLY_SUCCESS;
	}
	else{
		return REPLY_INVALID_METHOD;
	}

	return REPLY_GENERAL_ERROR;
}

