#ifndef POWERSAVE_POWERSAVED_H
#define POWERSAVE_POWERSAVED_H


#include <sys/un.h>

#include <string>
#include <iostream>

#include <errno.h>
#include "config_pm.h"
#include "powerlib.h"
//#include "cpufreq.h"

class EventManagement;
#include "clientConnection.h"
#include "brightness.h"


#include "dbus_server.h"

using namespace std;

class CPUFreq_Interface;

// Declarations:

#ifdef CPUFREQ_MEASURE
#define MEASURE_LOG_FILE "/tmp/cpufreq.log"
#endif

#define POWERSAVED_PID_FILE_PATH "/var/run/powersaved.pid"

//#define MAX_FILE_PATH 512
#define MAX_EVENT_ELEM_LENGTH 64
#define CONFIGURE_FILE "/etc/powersave.conf"
// MAX_PARAM_LENGTH also includes paths, and these can be long
// especially in test installations
#define MAX_PARAM_LENGTH 128

/** @brief used for different states of cpu, battery and temperature!
 *
 * Do not change order as it is used to evaluate whether a higher or lower
 * battery state is reached
 */
enum BATTERY_STATES {BAT_CRIT_STATE, BAT_LOW_STATE, BAT_WARN_STATE, BAT_NORM_STATE, BAT_NONE_STATE};
/** CPU States concern about throwing a processor.busy and processor.idle event
 * and to adjust CPU speed if CPU freq policy is set to dynamic.
 * CPU_DYNAMIC is needed for throwing processor.idle event(throttling in proxy). To have an
 * CPU frequency independent variable so that setting unsetting does not build up
 * on constant CPU usage ...
 */
enum CPU_STATES {CPU_LOW, CPU_HIGH, CPU_DYNAMIC_BELOW, CPU_DYNAMIC_ABOVE};

typedef struct PM_STATE{
	BATTERY_STATES BAT_STATE;
	// THERM_MODE comes from library libpower.h
	THERM_MODE TEMP_STATE;
	int CPU_IS_IDLE;
	AC_ADAPTER_STATE AC_STATE;
PM_STATE() {
     BAT_STATE = BAT_NORM_STATE;
     TEMP_STATE = OK;
     CPU_IS_IDLE = 0;
     AC_STATE = AC_UNKNOWN;
}

}PM_STATE;

/** sleep states requested by user, only used for APM standby/suspend
   workarounds. This should go somehow to APM_Interface later */
enum PSD_SLEEP_STATE_REQUESTS {PSD_NO_SLEEP_REQUEST=0, PSD_SUSPEND2DISK, PSD_SUSPEND2RAM, PSD_STANDBY};


// signal handlers
void usr1_handler_do_suspend_to_disk(int);
void usr2_handler_do_suspend_to_ram(int);
/** @brief // HUP signal handling ->re-read configs */
void hup_handler(int);
/** @brief terminate prog properly */
void exit_handler(int);

// helper functions
/** @brief helper function to damonize program in main function
 *
 * this is from: http://lib.ru/UNIXFAQ/unixprogrfaq.txt
 */
int daemonize();
std::string trim(const std::string& str);

/** used to evaluate whether to throw a processor.idle or processor.busy event
 * PROCESSOR_EVENTS_DISABLED is used when in high speed/performance or low speed /powersave cpu mode
 */
enum PROCESSOR_EVENTS_THROWN {PROCESSOR_BUSY_THROWN, PROCESSOR_IDLE_THROWN,
			      NO_PROCESSOR_EVENT_THROWN, PROCESSOR_EVENTS_DISABLED};

/** @brief global object, on this the whole app runs */
class PM_Interface{
 private:

	/* script path and name that gets executed for triggering 
	   sleep mode */
	static string INT_SCRIPT_PATH;

	// hold battery info for client requests
	BatteryGeneral bg;
	
	// this is the value of the highest fd value in the system
	int max_FD_value;
	
	// value of 1 indicates that processor.low event
	// has been thrown, value of 0 indicates, that it
	// has not thrown till now
	PROCESSOR_EVENTS_THROWN processor_event_thrown;
	// timout when processor.low event should be thrown
	// if it stays under the limit for this time
	unsigned long msecs_cpu_idle;
	
	// sum up time temperature was not polled
	unsigned msecs_last_temperature_read;

	/* this is to remember pid of daemon for children
	   if forked multiple times -> father may be dead */
	pid_t daemon_pid;

	/** @brief checks all pending events whether there timeouts are reached */
	void checkEventTimeouts();


	/** @brief handles a message on the manager interface
	 * @param msg the dbus message received
	 * @return DBUS_ERROR
	 */
	DBUS_ERROR handleManagerMessage( DBusMessage *msg );

	/** @brief handles a message on the request interface
	 * @param msg the dbus message received
	 * @param reply the dbus reply to send back
	 * @return DBUS_ERROR
	 */
	DBUS_ERROR handleRequestMessage( DBusMessage *msg,
					 DBusMessage *reply );

	/** @brief handles a message on the request interface
	 * @param msg the dbus message received
	 * @return DBUS_ERROR
	 */
	DBUS_ERROR handleActionMessage( DBusMessage *msg );

 protected:
	// sums up time since battery was last polled. not private, because it needs to be reset on ACPI event from outside.
	unsigned msecs_last_battery_read;

	int cooling_mode_supported;
	int throttling_supported;
	int thermal_trip_points_supported;
	
	PM_STATE cur_state;
	int hwEvent_fd;
	int poll_battery;
	int64_t supported_sleeping_states;

	/* used for workaround for apm to remembemer triggered sleep state *****
	   for ACPI to discard button events during resume */
	static PSD_SLEEP_STATE_REQUESTS requested_sleep_state;
	static int sleep_triggered;
	/* used for workaround ******************************************* *****/

 public:
	static string SLEEP_SCRIPT;

	/** @brief checks if a returned script belongs to a 
	   pending event and executes the appropriate actions
	*/
	void checkScriptReturn(int id, int returnValue, string comment);

	// general config object holds all config variables
	GeneralConfig *config;

	// object to manage client connections which will be informed of specific events
	clientConnection *cC;
	// object to manage running events
	EventManagement *eM;

	DBus_Server dbus_server;

	Brightness *brightness;

	// these two variables will be set through signals(usr1/usr2 signals) from child
	// that executes proxy before suspend
	// they are needed to be sure that proxy has terminated and prepared system for
	// suspend_to_ram/suspend_to_disk

	PM_Interface(GeneralConfig *config_obj);
	virtual ~PM_Interface();

	// have to be acpi/apm specific because resume in apm comes asynchronous
	// through /dev/apm_bios, acpi I can just trigger a resume event after
	// recovering from printf("3/4") to /proc/acpi/sleep
	
	/** @brief internal action which is executed directly before and after
	 * do_suspend_* should go into this function
	*/
	void sleep_wrapper(string);

	/** @brief used to execute scripts in INT_SCRIPT_PATH */
	int executeScript(string script, string param);

	/** @brief if a client is present which requested screenlock events,
	 * send event,
	 * otherwise execute the internal script do_screen_saver
	*/
	int screenlock(string param);

	/** @brief if a client is present which requested notification events,
	 * send event, otherwise execute the internal script do_x_notification
	 */
	int x_notification(string param);

	/** @brief sets if we ignore button events */
	void ignore_buttons(bool);

	virtual int suspend_to_ram() = 0;

	virtual int standby() = 0;

	virtual int executeEvent(const string eventName, const string acpi_event_line = "");

	/* these functions are for thread handling ******************/
	virtual int start();

	virtual int openHWEventFD() = 0;

	virtual int openAcpidSocket() = 0;

	virtual int handleHWEventRequest() = 0;

	virtual void checkTemperatureStateChanges();

        /** @brief Checks for Battery State changes (when AC adapter is not plugged in)
	    and throws appropriate event
	  *
	  * Be careful in ACPI you have to check whether the battery alarm also needs
	  * to be set
	  */
	virtual void checkBatteryStateChanges();

	/** @brief Checks for AC state changes and throws appropraite events
	 *
	 * if AC changes to offline, also battery state changes will be checked
	 * @return integer with state of AC 
	 * @retval 0 if AC state has not changed
	 * @retval 1 if AC state has changed
	 * @retval -1 if an error occur
	 */
	virtual int checkACStateChanges();

	/** @brief update temperature state
	 *
	 * @return integer with state of temperature
	 * @retval -1 on failure
	 * @retval 0 when state did not change
	 * @retval 1 when a more critical temperature level is reached
	 * @retval 2 when a less critical battery level is reached
	 */
	virtual int updateTemperatureState();

	/** @brief update the battery state
	 *
	 * @return integer with current battery state
	 * @retval -1 on failure
	 * @retval 0 when state did not change
	 * @retval 1 when a lower battery level is reached
	 * @retval 2 when a higher battery level is reached
	 */
	virtual int updateBatteryState();

	/** @brief updates ACState
	 *
	 * @return integer with result of check for changes
	 * @retval 0 if AC state did not change
	 * @retval 1 else 
	 */
	virtual int updateACState();

	virtual int updateCPUState();

	/** @brief switches settings to a possibly new scheme
	    and activates them
	    @return integer with result whether the scheme could (not) be set
	            or is already set.
	    @retval -1 if scheme does not exist
	    @retval  0 if scheme is already set
	    @retval  1 if scheme has been successfully set
	 */
	virtual int activateScheme(string scheme_name);
	/** @brief same as above, just use it if you want to switch the scheme
	    according to the AC state
	*/
	virtual int activateScheme(AC_ADAPTER_STATE ac_state);

	/** @brief Activate scheme settings 
	    Inheriting functions in APM_Interface and ACPI_Interface
	    set them specifically for APM/ACPI
	    
	    Add additional code that is needed after activating a new scheme
	    or after re-reading configurations here (e.g. brightness?)

	    Be aware that first the inhereting functions in the ACPI/APM_Interface
	    layer are called
	 */
	virtual void activateSettings();

	virtual int rereadConfig();

	/** @brief throttles the cpu
	 * 
	 * Only throttle if temperature is in unknown, ok or active state.
	 * Otherwise kernel may have throttled ...
	 *
	 * @return integer with result of throttle cpu
	 * @retval 1 on success
	 * @retval -1 if throttling failed
	 */
	virtual int throttle(int percent = -1);

	/** @brief dethrottles the cpu
	 *
	 * Only dethrottle if temperature is in unknown, ok or active state.
	 * Otherwise kernel may have throttled ...
	 *
	 * @return integer with result of dethrottle cpu 
	 * @retval 1 on success
	 * @retval -1 if throttling failed
	 */
	virtual int dethrottle();

	virtual void resume();
	
	//	class DBus_Message;
	DBUS_ERROR handleDBusRequest(DBusMessage *req, DBusMessage *reply, POWERSAVE_MESSAGE_TYPE type);
};

/** @brief global object, on this, the whole app runs
 *         must be declared extern to let DBus functions
 *         be able to access the handleDBusRequest function
 */
extern PM_Interface *pm;


#endif
