#include "config.h"
#include <stdio.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>

#include "powerlib.h"
#include "powerlib_local.h"

#define ACPI_BATTERY_DIR  "/proc/acpi/battery"
#define ACPI_BATTERY_INFO_FILE  "/proc/acpi/battery/info"
#define ACPI_BATTERY_STATE_FILE "/proc/acpi/battery/state"

#define ACPI_AC_DIR "/proc/acpi/ac_adapter"

/* needed to evaluate remaining minutes of battery power
   if present rate is not available (for broken ACPI systems) */
static capacity_snapshot rem_capacity[10] = {{0, 0}, };
static int current_snapshot = 0;


int getBatteriesInfo(BatteryGeneral *bat){
	int test;

	if (bat == NULL){
		errno = EINVAL;
		return -2;
	}
	bat->remPerc = -1;
	bat->remMins = -1;
	bat->charging_state=CHARG_STATE_UNKNOWN;
	bat->rem_charging_mins = -1;

	test = checkACPI();
	if (test == ACPI){
		if (get_ACPI_General_Battery_Info(bat) < 0){
			pDebug(DBG_INFO, "Could not get battery info for ACPI subsystem\n");
			return UNKNOWN;
		}
		else {
			return 1;
		}
	}
	else if (test == APM){
		if (get_APM_General_Battery_Info(bat) < 0){
			pDebug(DBG_INFO, "Could not get battery info for APM subsystem\n");
			return UNKNOWN;
		}
		else {
			return 1;
		}
	}
	else{
		return UNKNOWN;
	}
}


/* only for ACPI!!!*/

#ifndef HAVE_LIBHAL
int getBatteryDetailedInfo(const int no, BatteryDetailed *bd) {
	
	DIR *dir;
	struct dirent *dirent;
	char file[MAX_FILE_PATH + 1] = "";
        char dirent_name[MAX_FILE_PATH + 1] = "";
	char value[MAX_LINE_SIZE + 1] = "";
	char value1[MAX_LINE_SIZE + 1] = "";
	char unit[10] = "";
	int x;
	int res;

	FILE *rdFile;
	int dir_count = 0;

	if (bd == NULL){
		errno = EINVAL;
		return -3;
	}

	if (checkACPI() != ACPI){
		return NO_ACPI_ERROR;
	}
	
	if (no > MAX_BATTERIES-1 || no < 0){
		return -2;
	}
	/* check whether module has been loaded and device is available */
	res = check_ACPI_dir(ACPI_BATTERY_DIR);
	if (res < 0)
		/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return res;

	dir = opendir(ACPI_BATTERY_DIR);

	/* jump no of times in getting a new valid directory*/
	/* expect that dirent will always give back the same order of files(directories)*/
	while (1){
		dirent = readdir(dir);
		if (dirent == NULL){
			closedir(dir);
			return -2;
		}
		if (dirent->d_name[0] == '.')
			continue; /* skip dotfiles */
		if (dir_count == no)
			break;
		dir_count++;
	}
	
        strncpy(dirent_name, dirent->d_name, MAX_FILE_PATH);
        /* Close directly, we do not need dir anymore below.  */
        closedir(dir);

	/* get state file in battery directory*/
	res = snprintf(file, MAX_FILE_PATH, "%s/%s/state", ACPI_BATTERY_DIR, dirent_name);
	if (res >= MAX_FILE_PATH) {
	  pDebug (DBG_DIAG, "string too short, should be %d\n", res);
	  return -1;
	}
	
	rdFile = fopen(file, "r");
	if (!rdFile){
		pDebug(DBG_DIAG, "Could not open file %s for reading out battery information. err: %s\n",
		       file, strerror(errno));
		return -2;
	}
	pDebug(DBG_INFO, "File %s open to read out battery.\n", file);

	/*present*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0){
		bd->present = UNKNOWN;
	}
	else {
		if (!strcmp(value, "yes"))
			bd->present = PRESENT_YES;
		/* if not present initialise everything with unknown and return!*/
		else if (!strcmp(value, "no")){
			bd->present = PRESENT_NO;
			bd->capacity_state = UNKNOWN;
			bd->charging_state = CHARG_STATE_UNKNOWN;
			bd->design_capacity_low = UNKNOWN;
			bd->design_voltage = UNKNOWN;
			bd->battery_technology = UNKNOWN;
			bd->last_full_capacity = UNKNOWN;
			bd->design_capacity = UNKNOWN;
			bd->present_voltage = UNKNOWN;
			bd->power_unit = UNKNOWN;
			bd->remaining_capacity = UNKNOWN;
			bd->present_rate = UNKNOWN;
			bd->capacity_granularity_2 = UNKNOWN;
			bd->capacity_granularity_1 = UNKNOWN;
			bd->design_capacity_warning = UNKNOWN;
			strcpy(bd->OEM_info, "UNKNOWN");
			strcpy(bd->serial_number, "UNKNOWN");
			strcpy(bd->model_number, "UNKNOWN");
			strcpy (bd->battery_type, "UNKNOWN");
			bd->alarm_limit = UNKNOWN;
			pDebug(DBG_INFO, "Battery %s not present!\n", dirent_name);
			fclose(rdFile);
			return 0;
		}
		else{
			pDebug(DBG_DIAG, "Syntax in file: %s changed! Please report this line "
			       "(including Function name and line) to trenn@suse.de.\n", file);
			bd->present = UNKNOWN;
		}
	}
	
	/* capacity state*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof(value1)) < 0){
		pDebug(DBG_DIAG, "Could not read out capacity state of battery from %s.\n", file);
		 bd->capacity_state = UNKNOWN;
	}
	else {
		if (!strcmp(value, "ok"))
			bd->capacity_state = STATE_OK;
		else if (!strcmp(value, "critical"))
			bd->capacity_state = STATE_CRIT;
		else{
			pDebug(DBG_DIAG, "Syntax in file: %s changed! "
			       "Please report this line (including Function name and line)"
			       " to trenn@suse.de.\n", file);
			bd->capacity_state = UNKNOWN;
		}		
	}
	
	/* charging_state*/
        if (getColonValue(rdFile, value, sizeof(value), value1, sizeof(value1)) < 0)
		bd->charging_state = CHARG_STATE_UNKNOWN;
	else {
		if (!strcmp(value, "unknown"))
			bd->charging_state = CHARG_STATE_UNKNOWN;
		else if (!strcmp(value, "discharging"))
			bd->charging_state = CHARG_STATE_DISCHARGING;
		else if (!strcmp(value, "charging"))
			bd->charging_state = CHARG_STATE_CHARGING;
		else if (!strcmp(value, "charging/discharging"))
			bd->charging_state = CHARG_STATE_CHARG_DISCHARG;
		else if (!strcmp(value, "No charging/discharging"))
			bd->charging_state = CHARG_STATE_DISCHARGING;
		else 
			bd->charging_state = CHARG_STATE_UNKNOWN;
	}
	
	/* present_rate*/
        if (getColonValue(rdFile, value, sizeof(value), value1, sizeof (value1)) < 0)
		bd->present_rate = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->present_rate = UNKNOWN;
		else {
			bd->present_rate = (int)strtol(value, NULL, 10);
			//must have a value above 0!
			if (bd->present_rate == 0){
				bd->present_rate = UNKNOWN;
			}
		}
	}
	
	/* remaining_capacity*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->remaining_capacity = UNKNOWN;
	else {
		x = sscanf(value, "%s%s", value1, unit);
		if (*value == '0')
			bd->remaining_capacity = 0;
		else if (x == 0 || x == EOF || *value == '\0' || !strcmp(value, "unknown"))
			bd->remaining_capacity = UNKNOWN;
		else {
			bd->remaining_capacity = (int)strtol(value, NULL, 10);
			if (!bd->remaining_capacity){
				bd->remaining_capacity = UNKNOWN;
			}
			/* get power unit*/
			else {
				if ('A' == toupper(unit[1])){
					bd->power_unit = AMP;
				}
				else if ('W' == toupper(unit[1])){
					bd->power_unit = WATT_H;
				}
				else {
					bd->power_unit = UNKNOWN;
				}
			}		
		}
	}

	/* present_voltage*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->present_voltage = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->present_voltage = UNKNOWN;
		else {
			bd->present_voltage = (int)strtol(value, NULL, 10);
			if (!bd->present_voltage){
				bd->present_voltage = UNKNOWN;
			}
		}
	}
	
	fclose(rdFile);
	
	/* get info file in battery directory*/
	snprintf(file, MAX_FILE_PATH, "%s/%s/info", ACPI_BATTERY_DIR, dirent_name);
	rdFile = fopen(file, "r");
	if (!rdFile){
		pDebug(DBG_DIAG, "Could not open file %s"
		       " for reading out battery information. err: %s\n", file, strerror(errno));
		return -2;
	}
	pDebug(DBG_INFO, "File %s open to read out battery.\n", file);
	
	/* skip first line (we already have present ...)*/
	if (!fgets(value, sizeof(value), rdFile)) {
		pDebug(DBG_DIAG, "Could not get all lines of file %s ", file);
		fclose (rdFile);
		return -1;
	}

	
	/* design_capacity*/
        if (getColonValue(rdFile, value, sizeof(value), value1, sizeof (value1)) < 0)
		bd->design_capacity = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->design_capacity = UNKNOWN;
		else {
			bd->design_capacity = (int)strtol(value, NULL, 10);
			if (!bd->design_capacity){
				bd->design_capacity = UNKNOWN;
			}
		}
	}

	/* last_full_capacity*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->last_full_capacity = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->last_full_capacity = UNKNOWN;
		else {
			bd->last_full_capacity = (int)strtol(value, NULL, 10);
			if (!bd->last_full_capacity){
				bd->last_full_capacity = UNKNOWN;
			}
		}
	}

	/* battery_technology*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->battery_technology = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->battery_technology = UNKNOWN;
		else if (!strcmp(value, "rechargeable"))
			bd->battery_technology = TECH_RECHARGEABLE;
		else if (!strcmp(value, "non-rechargeable"))
			bd->battery_technology = TECH_NON_RECHARGEABLE;
		else {
			bd->battery_technology = UNKNOWN;
		}
	}

	/* design_voltage*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->design_voltage = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->design_voltage = UNKNOWN;
		else {
			bd->design_voltage = (int)strtol(value, NULL, 10);
			if (!bd->design_voltage){
				bd->design_voltage = UNKNOWN;
			}
		}
	}

	/* design_capacity_warning*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->design_capacity_warning = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->design_capacity_warning = UNKNOWN;
		else {
			bd->design_capacity_warning = (int)strtol(value, NULL, 10);
			if (!bd->design_capacity_warning){
				bd->design_capacity_warning = UNKNOWN;
			}
		}
	}

	/* design_capacity_low*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->design_capacity_low = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->design_capacity_low = UNKNOWN;
		else {
			bd->design_capacity_low = (int)strtol(value, NULL, 10);
			if (!bd->design_capacity_low){
				bd->design_capacity_low = UNKNOWN;
			}
		}
	}

	/* capacity_granularity_1*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->capacity_granularity_1 = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->capacity_granularity_1 = UNKNOWN;
		else {
			bd->capacity_granularity_1 = (int)strtol(value, NULL, 10);
			if (!bd->capacity_granularity_1){
				bd->capacity_granularity_1 = UNKNOWN;
			}
		}
	}

	/* capacity_granularity_2*/
        if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->capacity_granularity_2 = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->capacity_granularity_2 = UNKNOWN;
		else {
			bd->capacity_granularity_2 = (int)strtol(value, NULL, 10);
			if (!bd->capacity_granularity_2){
				bd->capacity_granularity_2 = UNKNOWN;
			}
		}
	}

	/* model_number*/
        if (getColonValue(rdFile, bd->model_number, sizeof(bd->model_number), value1, sizeof (value1)) < 0)
		strcpy(bd->model_number, "UNKNOWN");
		
	/* serial_number*/
	if (getColonValue(rdFile, bd->serial_number, sizeof(bd->serial_number), value1, sizeof (value1)) < 0)
		strcpy(bd->serial_number, "UNKNOWN");

	/* battery_type*/
	if (getColonValue(rdFile, bd->battery_type, sizeof (bd->battery_type), value1, sizeof (value1)) < 0)
		strcpy (bd->battery_type, "UNKNOWN");

	/* OEM_info*/
	if (getColonValue(rdFile, bd->OEM_info, sizeof (bd->OEM_info), value1, sizeof (value1)) < 0)
		strcpy(bd->OEM_info, "UNKNOWN");


	fclose(rdFile);

	/* get state file in battery directory*/
	snprintf(file, MAX_FILE_PATH, "%s/%s/alarm", ACPI_BATTERY_DIR, dirent_name);
	rdFile = fopen(file, "r");
	if (!rdFile)
		bd->alarm_limit = UNKNOWN;
	else {
		if (*value == '\0' || !strcmp(value, "unknown"))
			bd->alarm_limit = UNKNOWN;
		else {
			bd->alarm_limit = (int)strtol(value, NULL, 10);
			if (!bd->alarm_limit){
				bd->alarm_limit = UNKNOWN;
			}
		}
	}


	pDebug(DBG_INFO, "File %s open to read out battery.\n", file);

	/* alarm_info*/
	if (getColonValue(rdFile, value, sizeof (value), value1, sizeof (value1)) < 0)
		bd->alarm_limit = UNKNOWN;

	/* finish up*/
	fclose(rdFile);
	pDebug(DBG_INFO, "Battery status and info read out.\n");
        if ( bd->remaining_capacity > bd->last_full_capacity ) {
            pDebug(DBG_INFO, "remaining capacity is whack (%d), setting to last full (%d).\n",
                   bd->remaining_capacity, bd->last_full_capacity );
            bd->remaining_capacity = bd->last_full_capacity;
        }
	return 0;
}
#endif

/** 
 * Calculate general battery struct elements for ACPI
 * (remPerc, remMins, charging_state)
 * remPerc/remMins could be -1 if it could currently not be
 * calculated (could often happen shortly after unplugging AC power),
 * or on bad ACPI implemented machines.
 *  
 * @return integer with status of get batteryinfos
 * @retval 1 if machine has batteries plugged in
 * @retval 0 if machine supports batteries, but none is plugged in
 * @retval < 0 if there will never be any batteries(workstations)
 *         could e.g. be NO_DEVICE_ERROR or NO_MODULE_ERROR
*/
int get_ACPI_General_Battery_Info(BatteryGeneral *bat){

	int x, present=0;
	BatteryDetailed bd;
	int remaining_capacity, last_full_capacity, present_rate;
	present_rate = 0;
	remaining_capacity = 0;
	last_full_capacity = 0;
	int cmp = 0;
	float rem_battery_sum = 0;
	int divisor = 0;

	/* check whether module has been loaded and device is available */
	present = check_ACPI_dir(ACPI_BATTERY_DIR);
	if (present < 0)
		/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return present;

	/* iterate existing batteries*/
	for (x = 0; (getBatteryDetailedInfo(x, &bd) == 0) && x < MAX_BATTERIES; x++){
		if (bd.present == PRESENT_YES)  {
			present = 1;
			if (bd.last_full_capacity > 0){
				/* last remaining capacity might be greater than last full capacity */
				last_full_capacity += (bd.remaining_capacity > bd.last_full_capacity) ? bd.remaining_capacity : bd.last_full_capacity;
				/* on buggy batteries remaining capacity might be greater than last_full/design capacity */
				remaining_capacity += bd.remaining_capacity;
				present_rate += bd.present_rate;
			}
			/* keep charging info over multiple batteries */
			bat->charging_state=(bat->charging_state | bd.charging_state);
		}
	}
	if (!present){
		pDebug(DBG_INFO, "No batteries available.\n");
		return 0;
	}
	// cannot calc anything
	if (remaining_capacity <= 0){
		return 1;
	}
	else{

		/* calc remMins ***********************************/
		if (getACAdapterStatus() == AC_OFFLINE){

			// calc remMins from present_rate and remaining_capacity
			if (present_rate > 0){
				bat->remMins =  (int)roundf(((float)(remaining_capacity)/present_rate)*60.0);
			}

			// calc remMins depending on used capacity (stored from
			// previous remaining capacities
			else{
				/* previous and current remaining capacity are equal.
				   This happens if measurements are done too often.
				   Use the previous previous measurement then */
				if (rem_capacity[current_snapshot].remaining_capacity == remaining_capacity){
					pDebug(DBG_DIAG, "Battery capacity is equal to last measurement,"
					       "the battery should not be polled that often");
				}
				else{
					current_snapshot++;
					if (current_snapshot > 9)
						current_snapshot = 0;
					rem_capacity[current_snapshot].measure_point = time(NULL);
					rem_capacity[current_snapshot].remaining_capacity = remaining_capacity;
					for (int x = current_snapshot - 1, k = 0; k < 9; x--, k++){
						if (x < 0)
							cmp = x + 10;
						else
							cmp = x;
						/* check whether measurement is valid */
						if (rem_capacity[cmp].measure_point == 0
						    || rem_capacity[current_snapshot].measure_point < rem_capacity[cmp].measure_point
						    || rem_capacity[current_snapshot].remaining_capacity > rem_capacity[cmp].remaining_capacity)
							break;
						/* cmp. with last measure, with the one before until all ten
						   (if available) are cmp. with current rem. capacity.
						   Sum them up and divide later */
						rem_battery_sum += (float) rem_capacity[current_snapshot].remaining_capacity 
							* (rem_capacity[current_snapshot].measure_point -
							   rem_capacity[cmp].measure_point)
							/ ((rem_capacity[cmp].remaining_capacity - 
							    rem_capacity[current_snapshot].remaining_capacity) * 60);
						divisor++;
					}
					/* only calc when we have more than one measurement */
					if (divisor > 1){
						bat->remMins = rem_battery_sum / divisor;
					}
				}
			}
		}
		/* calc remMins ***********************************/

		/* calc remaining charging time until battery is fully charged again
		   if AC online ********/

		else if (last_full_capacity > 0){
			if (present_rate > 0){
				bat->rem_charging_mins = (int)roundf(((float)(last_full_capacity-remaining_capacity)
								      /present_rate)*60.0);
			}
			/* estimate remaing charging time from increasing remaining capacity,
			   as it is also done for rem minutes (see above)
			*/
			else{
				/* previous and current remaining capacity are equal.
				   This happens if measurements are done too often.
				   Use the previous previous measurement then */
				if (rem_capacity[current_snapshot].remaining_capacity == remaining_capacity){
					pDebug(DBG_DIAG, "Battery capacity is equal to last measurement,"
					       "the battery should not be polled that often");
				}
				else{
					current_snapshot++;
					if (current_snapshot > 9)
						current_snapshot = 0;
					rem_capacity[current_snapshot].measure_point = time(NULL);
					rem_capacity[current_snapshot].remaining_capacity = remaining_capacity;
					for (int x = current_snapshot - 1, k = 0; k < 9; x--, k++){
						if (x < 0)
							cmp = x + 10;
						else
							cmp = x;
						/* check whether measurement is valid */
						if (rem_capacity[cmp].measure_point == 0
						    || rem_capacity[current_snapshot].measure_point < rem_capacity[cmp].measure_point
						    || rem_capacity[current_snapshot].remaining_capacity > rem_capacity[cmp].remaining_capacity)
							break;
						/* cmp. with last measure, with the one before until all ten
						   (if available) are cmp. with current rem. capacity.
						   Sum them up and divide later */
						rem_battery_sum += (float) (rem_capacity[current_snapshot].remaining_capacity-last_full_capacity)
							* (rem_capacity[current_snapshot].measure_point -
							   rem_capacity[cmp].measure_point)
							/ ((rem_capacity[current_snapshot].remaining_capacity -
							    rem_capacity[cmp].remaining_capacity)
							   * 60);
						divisor++;
					}
					/* only calc when we have more than one measurement */
					if (divisor > 1){
						bat->remMins = rem_battery_sum / divisor;
					}
				}
			}
		}
		/* calc remaining chargin time if AC online *******/

		/* calc remPerc ***********************************/
		if (last_full_capacity > 0){
		bat->remPerc = (int)roundf(((float)remaining_capacity/(float)last_full_capacity)*100.0);
		if (bat->remPerc > 150 || bat->remPerc < 0){
			pDebug(DBG_INFO, "Wrong calculated battery remaining "
			       "Percent capacity value: %d, set to 100", bat->remPerc);
			bat->remPerc = 100;
		}
		else if (bat->remPerc > 100)
			bat->remPerc = 100;
		}
		/* calc remPerc ***********************************/
	}
	return 1;
}

int get_APM_General_Battery_Info(BatteryGeneral *bg){
	struct apm_info apmInfo;
	int t;

	t = read_apm_info(&apmInfo);
	if (t == -2){
		pDebug(DBG_DIAG, "Old apm style not suppported !\n");
	}
	else if (t < 0){
		pDebug(DBG_DIAG, "apm not suppported, this must not happen,"
		       " apm supported already tested!\n");
		return -1;
	}
	bg->remPerc = apmInfo.battery_percentage;
	if (apmInfo.battery_status == 5){
		bg->charging_state = CHARG_STATE_CHARGING;
		/*		bg->capacity_state = STATE_OK;*/
	}
	else 
		bg->charging_state = CHARG_STATE_DISCHARGING;

	/* there is no warning state in APM, only low and critical !*/
	/* but I do not need it, because I define it myself! */
	/* inside the daemon ...

	if (apmInfo.battery_status == 3) {
		bg->capacity_state = STATE_LOW;
	}
	else if (apmInfo.battery_status == 4) {
		bg->capacity_state = STATE_CRIT;
	}
	*/
	/* no remaining time available, probably on AC power*/
	if (apmInfo.battery_time == -1)
		bg->remMins = UNKNOWN;
	/* remaining time in minutes */
	else if (apmInfo.using_minutes){
		bg->remMins = apmInfo.battery_time;
	}
	/* remaining time in seconds */
	else{
		bg->remMins = apmInfo.battery_time / 60;
	}
	/* no battery flag set ? */
	if (apmInfo.battery_flags & 0x80)
		return -1;
	else
		return 1;
}

#ifndef HAVE_LIBHAL
int getACAdapterStatus(void){
	/* differ between several kernel implementations (where is the proc entry) here*/

 	char value[MAX_LINE_SIZE+1] = "";
 	char temp[MAX_LINE_SIZE+1] = "";
	char file[MAX_FILE_PATH+1] = "";
 	char file1[MAX_FILE_PATH+1] = "";

	FILE *rdFile;
	struct apm_info apmInfo;
	int ret;
	int test;
	
	test = checkACPI();
	/* --------- ACPI ------------- //*/
		
	if (test == ACPI){
		/* check whether module has been loaded and device is available */
		ret = check_ACPI_dir(ACPI_AC_DIR);
		if (ret < 0)
			/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return ret;

		if (getDirEntry(0, file1, sizeof (file1), ACPI_AC_DIR) < 0){
			pDebug(DBG_INFO, "Could not determine file where ac"
			       "status is written in directory: %s! Is this a workstation?\n", ACPI_AC_DIR);
			return NO_DEVICE_ERROR;
		}
		/* possible to overwrite file ?!?*/
		snprintf(file, MAX_FILE_PATH, "%s/%s/state", ACPI_AC_DIR, file1);
		pDebug(DBG_INFO, "AC state file: %s\n", file);
		rdFile = fopen(file, "r");
		if (!rdFile){
			return AC_UNKNOWN;
		}
		
		if (getColonValue(rdFile, value, sizeof (value), temp, sizeof (temp)) < 0){
			pDebug(DBG_DIAG, "Could not read out ac state from %s!\n", file);
			ret = AC_UNKNOWN;
		}
		else {
			if (!strcmp(value, "on-line")){
				/* reset remaining battery capacity measurements */
				memset(rem_capacity, 0, sizeof(capacity_snapshot)*10);
				current_snapshot = 0;
				ret = AC_ONLINE;
			}
			else if (!strcmp(value, "off-line"))
				ret = AC_OFFLINE;
			else{
				pDebug(DBG_DIAG, "Syntax in file: %s changed!"
				       " Value returned: -%s-!\n", ACPI_AC_DIR, value);
				ret = AC_UNKNOWN;
			}
		}
		fclose(rdFile);		
	}
	/*------------- APM ----------------//*/

	else if (test == APM){
		test = read_apm_info(&apmInfo);
		if (test == -2){
			pDebug(DBG_DIAG, "Old apm style not suppported !\n");
		}
		else if (test < 0){
			pDebug(DBG_DIAG, "apm not suppported, this must not happen,"
			       " apm supported already tested!\n");
			return AC_UNKNOWN;
		}
		if (apmInfo.ac_line_status == 0)
			return AC_OFFLINE;
		/*		else if (apmInfo.ac_line_status == 2)
				return AC_BACK_LINE !?!?*/ /* ToDo: what is back line for a state ?!?*/
		else
			return AC_ONLINE;
	}
	else{
		pDebug(DBG_DIAG, "Neither APM nor ACPI support found!\n");		
		return AC_UNKNOWN;
	}
	pDebug(DBG_INFO, "AC status and info read out !\n");
	return ret;
}
#endif

/* only for ACPI!!!
return x < 0 on error
returns 0 if alarm has already sub-ceded(alarm will still be set)
returns positive value if alarm has been set successfully
*/

int setBatteryAlarm(int percent){
	int x, max_present_bats;
	BatteryDetailed *bd[MAX_BATTERIES];
	/* divide_factor: factor on which all batteries are above the wanted percent (statistically)*/
	int remaining_capacity_sum, last_full_capacity_sum, rem_Perc_all;
	float divide_factor = 0.0;
	remaining_capacity_sum = 0;
	last_full_capacity_sum = 0;
	int alarm_value = 0;
	char file[MAX_FILE_PATH+1] = "";
	char dev_name[MAX_FILE_PATH+1] = "";
	char verify[MAX_LINE_SIZE+1] = "";
	char temp[MAX_LINE_SIZE+1] = "";
	int ret = 1;
	FILE *fp;

	if (checkACPI() != ACPI){
		return NO_ACPI_ERROR;
	}
	
	/* check whether module has been loaded and device is available */
	ret = check_ACPI_dir(ACPI_BATTERY_DIR);
	if (ret < 0)
		/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return ret;

	if (percent <= 0 || percent >100){
		pDebug(DBG_DIAG, "Wrong value for set battery alarm in percent: %d", percent);
		return -1;
	}

	for (x = 0; x < MAX_BATTERIES; x++){
		bd[x] = (BatteryDetailed*)malloc(sizeof(BatteryDetailed));
	}
	/* iterate present battery devices. Don't iterate on existing devices: */
	/* Imagine BAT0 -> not present, BAT1 -> present*/
	for (x = 0; (getBatteryDetailedInfo(x, bd[x]) == 0) && x < MAX_BATTERIES; x++){
		/* I could also get general bat info again, but this would */
		/*result in reading out battery state one more time ...*/
		
		if (bd[x]->present == PRESENT_YES)  {
			if (bd[x]->remaining_capacity > 0){
				if (bd[x]->last_full_capacity > 0){
					last_full_capacity_sum += bd[x]->last_full_capacity;
					remaining_capacity_sum += bd[x]->remaining_capacity;
				}
			}
		}
	}
	max_present_bats = x;
	/* calculate remaining percent of the capacity of all sumerized batteries ***/
	if (last_full_capacity_sum > 0 && remaining_capacity_sum > 0){
		rem_Perc_all = 
			(int)roundf(((float)remaining_capacity_sum/(float)last_full_capacity_sum)*100.0);
		if (rem_Perc_all > 140 || rem_Perc_all < 0){
			pDebug(DBG_DIAG, "Wrong calculated battery remaining Percent"
			       " capacity value: %d percent left, set to 100 percent\n", rem_Perc_all);
			rem_Perc_all = 100;
		}
		else if (rem_Perc_all > 100)
			rem_Perc_all = 100;
	}
	else {
		pDebug(DBG_INFO, "Could not read out last_full "
		       "capacity and remaining_capacity of batteries!\n");
		goto ERROR;
	}
	/*****************************************************************************/
	if (percent > 0)
		divide_factor = (float)rem_Perc_all/percent;
	pDebug(DBG_INFO, "Adjust alarms: rem_Perc of all batteries: %d percent,"
	       " wanted alarm at: %d percent, divide factor: %f.\n", 
	       rem_Perc_all, percent, divide_factor);
	for (x = 0; x < max_present_bats && x < MAX_BATTERIES; x++){
		if (bd[x]->present != PRESENT_YES)
			continue;
		if (divide_factor < 1){
			pDebug(DBG_DIAG, "Divide factor: %f, request to set battery alarm"
			       " over current percent remaining capacity!\n", divide_factor);
			goto ERROR;
		}
		alarm_value = (int)roundf(bd[x]->remaining_capacity / divide_factor);
		pDebug(DBG_DIAG, "\nBattery %d:\n Remaining capacity: %d\n"
		       "last full capacity: %d \nAlarm will be set to %d !\n",
		       x, bd[x]->remaining_capacity, bd[x]->last_full_capacity, alarm_value);
		
		/*******  write alarm for each battery ************************/
		if (getDirEntry(x, dev_name, sizeof (dev_name), ACPI_BATTERY_DIR) < 0){
			pDebug(DBG_DIAG, "Could not write alarm value for battery no.: %d", x);
			goto ERROR;
		}
		strncpy (file, ACPI_BATTERY_DIR, MAX_FILE_PATH);
		if (strlen(file) + strlen(dev_name), strlen("/alarm" + 1) > MAX_FILE_PATH){
			pDebug(DBG_DIAG, "Path of file %s/alarm too long,"
			       " please increase MAX_FILE_PATH variable!", file);
			goto ERROR;
		}
		else {
			strcat(file, "/");	       
			strcat(file, dev_name);	       
			strcat(file, "/alarm");
			if (_write_line(file, "%d\n", alarm_value) < 0){
				pDebug(DBG_INFO, "Could not write alarm value!");			
				goto ERROR;
			}
			else {
				/* don't set return value has already been set to 0 or initialised with 1*/
				if ((fp = fopen(file, "r")) == NULL){
					pDebug(DBG_DIAG, "Could not read alarm settings made!\n");
				}

				getColonValue(fp, verify, sizeof (verify), temp, sizeof (temp));
				if (atoi(verify) != alarm_value){
					pDebug(DBG_DIAG, "failure. wrote: %d, read: %s\n", alarm_value, verify);
					fclose(fp);
					goto ERROR;
				}
				fclose(fp);
				pDebug(DBG_DEBUG, "Alarm value: %d successfully written to %s!", alarm_value, file);
			}
		}
		/*******  write alarm for each battery ************************/
	}
	for (x = 0; x < MAX_BATTERIES; x++){
		if (bd[x] != NULL)
			free(bd[x]);
	}
	return ret;

 ERROR: 
	for (x = 0; x < MAX_BATTERIES; x++){
		if (bd[x] != NULL)
			free(bd[x]);
	}
	return UNKNOWN;
	
}
