#include "config.h"

#ifdef HAVE_LIBHAL

#include "hal.h"
#include <errno.h>
#include <stdlib.h>

#define ACPI_BATTERY_DIR  "/proc/acpi/battery"
#define ACPI_AC_DIR "/proc/acpi/ac_adapter"

extern DBusError dbus_error;
extern LibHalContext *hal_ctx;

int getBatteryDetailedInfo(const int no, BatteryDetailed *bd) {
	char value[MAX_LINE_SIZE + 1] = "";

	if ( ! initHal() ){
		/* error message already thrown in initHal() */
		return -1;
	}

	if (bd == NULL){
		errno = EINVAL;
		return -3;
	}

	if (checkACPI() != ACPI){
		return NO_ACPI_ERROR;
	}

	if (no > MAX_BATTERIES-1 || no < 0){
		return -2;
	}

	/* check whether module has been loaded and device is available */
	int res = check_ACPI_dir(ACPI_BATTERY_DIR);
	if (res < 0)
		/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return res;

	char **batteries;
	int numBatteries;
	batteries = libhal_find_device_by_capability( hal_ctx,
						      "battery",
						      &numBatteries,
						      &dbus_error);

	if ( dbus_error_is_set( &dbus_error) ) {
		pDebug(DBG_ERR, "Could not get list of batteries in system: %s",
		       dbus_error.message );
		freeHal();
		dbus_error_free( &dbus_error );
		libhal_free_string_array( batteries );
		return 0;
	}

	/* if there are no batteries */
	if ( numBatteries == 0 ) {
		libhal_free_string_array( batteries );
		return NO_DEVICE_ERROR;
	}

	/* if battery does not exist */
	if ( no > numBatteries-1 || no < 0 ){
		libhal_free_string_array( batteries );
		return -2;
	}

	int present = hal_get_property_bool( batteries[no],
					     "battery.present" );

	if ( present < 0 )
		bd->present = UNKNOWN;
	/* if not present initialise everything with unknown and return!*/
	else {
		if ( present ) {
			bd->present = PRESENT_YES;
		}
		else {
			bd->present = PRESENT_NO;
			bd->capacity_state = UNKNOWN;
			bd->charging_state = CHARG_STATE_UNKNOWN;
			bd->design_capacity_low = UNKNOWN;
			bd->design_voltage = UNKNOWN;
			bd->battery_technology = UNKNOWN;
			bd->last_full_capacity = UNKNOWN;
			bd->design_capacity = UNKNOWN;
			bd->present_voltage = UNKNOWN;
			bd->power_unit = UNKNOWN;
			bd->remaining_capacity = UNKNOWN;
			bd->present_rate = UNKNOWN;
			bd->capacity_granularity_2 = UNKNOWN;
			bd->capacity_granularity_1 = UNKNOWN;
			bd->design_capacity_warning = UNKNOWN;
			strcpy(bd->OEM_info, "UNKNOWN");
			strcpy(bd->serial_number, "UNKNOWN");
			strcpy(bd->model_number, "UNKNOWN");
			strcpy (bd->battery_type, "UNKNOWN");
			bd->alarm_limit = UNKNOWN;
			pDebug(DBG_INFO, "Battery %s not present!\n", batteries[no]);
			libhal_free_string_array( batteries );
			return 0;
		}
	}

	/* capacity state */
	if ( hal_get_property_string( batteries[no],
				      "battery.charge_level.capacity_state",
				      value ) ) {
		if (!strcmp(value, "ok"))
			bd->capacity_state = STATE_OK;
		else if (!strcmp(value, "critical"))
			bd->capacity_state = STATE_CRIT;
		else{
			bd->capacity_state = UNKNOWN;
		}		
	}

	/* --- get charging state --- */
	bd->charging_state = hal_get_property_bool( batteries[no],
						    "battery.rechargeable.is_charging" );

	if ( bd->charging_state == 1 )
		bd->charging_state = CHARG_STATE_CHARGING;
	else if ( bd->charging_state == 0 )
		bd->charging_state = CHARG_STATE_DISCHARGING;
	else
		bd->charging_state = CHARG_STATE_UNKNOWN;



	/* present_rate*/
	bd->present_rate = hal_get_property_int( batteries[no],
						 "battery.charge_level.rate" );
	if (bd->present_rate <= 0 ) {
		bd->present_rate = UNKNOWN;		
	}

	/* --- get remaining capacity --- */
	bd->remaining_capacity = hal_get_property_int( batteries[no],
						       "battery.charge_level.current" );
	if ( bd->remaining_capacity < 0 ){
		bd->remaining_capacity = UNKNOWN;
	}

	/* get power unit*/
	hal_get_property_string( batteries[no],
				 "battery.charge_level.unit",
				 value );
	if ( !strcmp( value, "mWh") ) {
		bd->power_unit = WATT_H;
	}
	else if ( ! strcmp( value, "mA") ) {
		bd->power_unit = AMP;
	}
	else {
		bd->power_unit = UNKNOWN;
	}

	/* present_voltage */
	bd->present_voltage = hal_get_property_int( batteries[0],
						    "battery.charge_voltage.current" );
	if ( bd->present_voltage < 0 ) {
		bd->present_voltage = UNKNOWN;
	}

	/* design_capacity*/
	bd->design_capacity = hal_get_property_int( batteries[no],
						    "battery.charge_level.design" );
	if ( bd->design_capacity < 0 ){
		bd->design_capacity = UNKNOWN;
	}
	

	/* last_full_capacity*/
	bd->last_full_capacity = hal_get_property_int( batteries[no],
						       "battery.charge_level.last_full" );
	if ( bd->design_capacity < 0 ){
		bd->last_full_capacity = UNKNOWN;
	}

	/* battery_technology*/
	bd->battery_technology = hal_get_property_bool( batteries[no],
						       "battery.is_rechargeable" );
	if ( bd->battery_technology ) {
		bd->battery_technology = TECH_RECHARGEABLE;
	}
	else if ( ! bd->battery_technology ) {
		bd->battery_technology = TECH_NON_RECHARGEABLE;
	}
	else {
		bd->battery_technology = UNKNOWN;
	}

	/* desing voltage */
	bd->design_voltage = hal_get_property_int( batteries[no],
						   "battery.voltage.design" );
	if ( bd->design_voltage < 0 ) {
		bd->design_voltage = UNKNOWN;
	}
	
	/* design_capacity_warning */
	bd->design_capacity_warning = hal_get_property_int( batteries[no],
							    "battery.charge_level.warning" );
	if ( bd->design_capacity_warning < 0 ) {
		bd->design_capacity_warning = UNKNOWN;
	}

	/* design_capacity_low */
	bd->design_capacity_low = hal_get_property_int( batteries[no],
							"battery.charge_level.low" );
	if ( bd->design_capacity_low < 0 ) {
		bd->design_capacity_low = UNKNOWN;
	}

	/* capacity_granularity_1 */
	bd->capacity_granularity_1 = hal_get_property_int( batteries[no],
							   "battery.charge_level.granularity_1" );
	if ( bd->capacity_granularity_1 < 0 ) {
		bd->capacity_granularity_1 = UNKNOWN;
	}

	/* capacity_granularity_1 */
	bd->capacity_granularity_2 = hal_get_property_int( batteries[no],
							   "battery.charge_level.granularity_2"
								     );
	if ( bd->capacity_granularity_2 <=0 ) {
		bd->capacity_granularity_2 = UNKNOWN;
	}

	/* alarm limit */
	bd->alarm_limit= hal_get_property_int( batteries[no],
					       "battery.alarm.design" );
	if ( bd->alarm_limit <=0 ) {
		bd->alarm_limit = UNKNOWN;
	}


	if ( ! hal_get_property_string( batteries[no],
					"battery.model",
					bd->model_number ) ) {
		strcpy( bd->model_number, "UNKNOWN" );
	}

	if ( ! hal_get_property_string( batteries[no],
					"battery.serial",
					bd->serial_number ) ) {
		strcpy(bd->serial_number, "UNKNOWN");
	}
		

	if ( ! hal_get_property_string( batteries[no],
					"battery.technology",
					bd->battery_type ) ) {
		strcpy(bd->battery_type, "UNKNOWN");
	}

	if ( ! hal_get_property_string( batteries[no],
					"battery.vendor",
					bd->OEM_info ) ) {
		strcpy(bd->OEM_info, "UNKNOWN");
	}

	libhal_free_string_array( batteries );
 
	return 0;
}


int getACAdapterStatus(void){
	int test;
	struct apm_info apmInfo;
	int ret;

	if ( ! initHal() ){
		/* error message already thrown in initHal() */
		return -1;
	}
	
	char **ac_adapter;
	char *ac_device;
	int temp;
	int num;

	test = checkACPI();
	/* --------- ACPI ------------- //*/
		
	if (test == ACPI){
		/* TODO: Only check for module here, because device
		 * availability is checked for above */
		ret = check_ACPI_dir(ACPI_AC_DIR);
		if (ret < 0)
			/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return ret;

		// currently only ACPI property is fetched through hal
		ac_adapter = libhal_find_device_by_capability( hal_ctx,
							       "ac_adapter",
							       &num,
							       &dbus_error);
		if (ac_adapter != NULL){
			ac_device = strdup(*ac_adapter);
			for (temp = 0; ac_adapter[temp]; temp++)
				free (ac_adapter[temp]);
			free(ac_adapter);
		}
		else
			return NO_DEVICE_ERROR;
		
		if ( dbus_error_is_set( &dbus_error)){
			pDebug(DBG_ERR, "Could not get ac_adapter device: %s",
			       dbus_error.message );
			dbus_error_free( &dbus_error );
			return NO_DEVICE_ERROR;
		}

		int ac_state = hal_get_property_bool( ac_device,
						      "ac_adapter.present" );
		free (ac_device);
		if ( ac_state ) {
			return AC_ONLINE;
		}
		else if ( ! ac_state ) {
			return AC_OFFLINE;
		}
		else {
			return AC_UNKNOWN;
		}
	}
	else if (test == APM){
		test = read_apm_info(&apmInfo);
		if (test == -2){
			pDebug(DBG_DIAG, "Old apm style not suppported !\n");
		}
		else if (test < 0){
			pDebug(DBG_DIAG, "apm not suppported, this must not happen,"
			       " apm supported already tested!\n");
			return AC_UNKNOWN;
		}
		if (apmInfo.ac_line_status == 0)
			return AC_OFFLINE;
		/*		else if (apmInfo.ac_line_status == 2)
				return AC_BACK_LINE !?!?*/ /* ToDo: what is back line for a state ?!?*/
		else
			return AC_ONLINE;
	}
	else{
		pDebug(DBG_DIAG, "Neither APM nor ACPI support found!\n");		
		return AC_UNKNOWN;
	}
	pDebug(DBG_INFO, "AC status and info read out !\n");
	return ret;
}

#endif /* HAVE_LIBHAL */
