/** @defgroup examples_dbus_testclient Example client or testclient with dbus
 * @include dbus_testclient.c
 */

#include "../libpower/powersave_dbus.h"
#include "../libpower/powersave_clientsocket.h"

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

void usage();
DBusHandlerResult filter_function (DBusConnection *, DBusMessage *, void *);

GMainLoop *gmain;

int main (int argc, char *argv[]) {
	if (argv[1]) {
		if (strcmp(argv[1],"-h")==0 || strcmp(argv[1],"--help")==0 ) {
			usage();
			exit(EXIT_SUCCESS);
		}
	}

	dbus_int32_t capabilities;
	DBusError error;
	DBusConnection *connection;

	if (!argv[1]) {
		printf( "No capabilities given. Using default "
			"(means get everything).\n" );
		/* the capacbilities consist of everthing powersave is
		 * able to give us */
		capabilities =  CAPABILITY_NOTIFICATIONS
			| CAPABILITY_SCREENLOCK;
	}
	else {
		capabilities = strtol( argv[1], NULL, 10 );
	}

	/* set up a new main loop with glib */
	gmain = g_main_loop_new (NULL, FALSE);

	dbus_error_init( &error );
	
	/* get connection from dbus */
	connection = dbus_bus_get( DBUS_BUS_SYSTEM, &error );

	if ( dbus_error_is_set( &error ) ) {
		printf("Error: %s\n", error.message);
		dbus_error_free( &error );
		exit(EXIT_FAILURE);
	}

	/* set up the dbus connection with the glib main loop */
	dbus_connection_setup_with_g_main( connection, NULL);

	/* add the filter function which should be executed on events on the bus */
	if ( ! dbus_connection_add_filter( connection, filter_function, gmain, NULL) ) {
		printf("Error: Not enough memory to add filter to dbus connection\n");
		exit(EXIT_FAILURE);
	}
	
	/* add a match rule to catch all signals going through th bus with
	 * powersave manager interface */
	dbus_bus_add_match( connection,
			    "type='signal',"
			    "interface='com.novell.powersave.manager',"
			    "path='/com/novell/powersave',", NULL);

	/* connect to the daemon ( from powersave_dbus )*/
	if ( establishConnection( capabilities ) != REPLY_SUCCESS ) {
		printf( "Powersave daemon did not accept connection request." );
		exit(EXIT_FAILURE);
	}

	/* run the main loop */
	g_main_loop_run( gmain );
	
	exit(EXIT_SUCCESS);
}


DBusHandlerResult filter_function( DBusConnection *connection,
				   DBusMessage *message,
				   void *data) {

	DBusError error;
	dbus_error_init( &error );

	if ( dbus_message_get_type( message ) != DBUS_MESSAGE_TYPE_SIGNAL ) {
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}

	/* the argument received */
	char *value;
	
	/* get the name of the signal */
	const char *signal = dbus_message_get_member( message );
	
	if ( dbusGetMessageString( message, &value, 0 ) < 0 ) {
		printf( " Warning: Received signal %s, but no string argument\n",
			error.message );
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	
	/* our name is... */
	if ( ! strcmp( signal, "NameAcquired" ) ) {
		printf( "DBus connection acquired: %s\n"
			"Waiting for signals from powersave daemon...\n\n",
			value );
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	/* acpi event received */
	else if ( ! strcmp( signal, "AcpiEvent" ) ) {
		printf( "Received acpi event. Value: %s\n", value );
	}
	/* powersave event received */
	else if ( ! strcmp( signal, "PowersaveEvent" ) ) {
		printf( "Received powersave event. Value: %s\n", value );
		
		/* daemon shut down, we do the same... */
		if ( ! strcmp( value, "daemon.terminate" ) ) {
			g_main_loop_quit( gmain );
		}
	}
	/* progress information received */
	else if ( ! strcmp( signal, "Progress" ) ) {
		printf( "Received progress information. Value: %s\n", value );
	}
	/* notification received */
	else if ( ! strcmp( signal, "Notification" ) ) {
		printf( "Received notification. Value: %s\n", value );
	}
	/* screenlock request received */
	else if ( ! strcmp( signal, "Screenlock" ) ) {
		printf( "Received screenlock request. Value: %s\n", value );
	}
	else {
		printf( "Received unknown signal %s. Value: %s\n",
			signal, value );
	}
	
	return 0;
}


void usage(){
	printf("\nTest client for connetion to powersaved and\n"
	       "getting informed about acpi events, powersave events,\n"
	       "progress information or notifications.\n"
	       "This client establishes the connection with DBus.\n\n"
	       "Usage: ./dbus-testclient <capabilities>\n"
	       "Capabilities are a sum of CAPABILITIES from powersave_clientsocket.h\n"
	       "All capabilities can be added up. They are simple integers.\n\n"
	       "Example 1: ./dbus_testclient 1\n"
	       "Means: Get all acpi events from /proc/acpi\n\n"
	       "Example 2: ./dbus_testclient 2\n"
	       "Means: Get all powersave events \n\n" 
	       "Example 3: ./dbus_testclient 4\n"
	       "Means: Get all notifications\n\n"
	       "Example 4: ./dbus_testclient 8\n"
	       "Means: Get screenlock requests.\n\n"
	       "Example 5: ./dbus_testclient 16\n"
	       "Means: Get progress information.\n\n"
	       "Example 6: ./dbus_testclient 31\n"
	       "Means: Get everything\n\n" );
}
